
#ifndef TMC_ROUTINES_H
#define TMC_ROUTINES_H

   

#include <stdio.h>
#include <fstream.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "QTMC_idfs_utils.h"
#include "qdutil_copies.h"
#include "idfs_retrieval.h"
#include "write_data_files.h"
#include "truncated_moments_corrector.h"   
#include "Idfs.h"
#include "Qdos.h"
#include "Xrefs.h"    

#define DEBUG_LOG
    
#define N_TMC_ERROR 12 
    typedef enum TMC_ERROR_E{
      TMC_OK=0,     //SUCCESS!!!
	    TMC_JOIN_ERR, //joining error
	    TMC_IDFS_RET_ERR, //idfs retrieval error
	    TMC_IS_NIL_ERR, //is nil error
	    TMC_INPUT_ERR, //error in the inputs!
	    TMC_SET_LIN_ERR,  //error setting lineages
	    TMC_GET_CAL_ERR,  //error getting calibration info
	    TMC_SEP_ERR, //error separating overlap data (i.e. L1L2 -> L1+L2, etc)
	    TMC_MAIN_CALC_ERR, //abort in the main calculation (i.e. main solver loop)
	    TMC_CREATE_OUTPUT_ERR, //error creating output objects
	    TMC_CLEAN_ERR, //error cleaning outputs
	    TMC_ERR    // generic/unknown error
    } TMC_ERROR;
    
//simple function to print msg to stdout    
void DebugCout (const char* msg);
     
//main TMC routine which takes CL or GUI inputs and performs EVERYTHING! returns TMC_OK if successful
TMC_ERROR TMC(int start_year, int start_doy, int start_hour, int start_min, int start_sec, double start_nano,
		       int end_year, int end_doy, int end_hour, int end_min, int end_sec, double end_nano,
		       int space_craft,
		       double lower_guess_limit, double upper_guess_limit, double guess_loop_res,
		       double phi_offset, double sc_potential, bool force,
		       bool B, bool L, bool H, bool T,
		       int output_format, string *Lcalname, string *Hcalname, string *dest, bool logging, bool clobber);
		       
//prints an informative message given one of the enum TMC errors		      
void print_error_message(TMC_ERROR e);

//returns |Q| from the 3-vector Q          		      
double vector_mag(double Q[3]);

TMC_ERROR TMC_HISTORY( int start_year, int start_doy, int start_hour, int start_min, int start_sec, double start_nano,
              	       int end_year, int end_doy, int end_hour, int end_min, int end_sec, double end_nano,
	                     int spacecraft) ;

#endif //TMC_ROUTINES_H
    
