/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
**
** If you wish to add, delete or rename functions or slots use
** Qt Designer which will update this file, preserving your code. Create an
** init() function in place of a constructor, and a destroy() function in
** place of a destructor.
*****************************************************************************/

#ifndef TMCFORM_UI_H
#define TMCFORM_UI_H

#include <qvalidator.h>
#include <qcheckbox.h>
#include <qprogressbar.h>
#include <qmessagebox.h>
#include <qfiledialog.h>
#include <qmetaobject.h> 
#include "write_data_files.h"
#include "tmc_routines.h"

QString stars =  "\n*************************************************\n";
QString idfs_retrieval_error = "Error retrieving data!";
QString is_nil_error = "Error: sequence is nil!"; 
QString join_error = "Error during join method!";
QString generic_err = "Error correcting moments!";
QString L1 = "/CL1_LEEA.cal";
QString L2 = "/CL2_LEEA.cal";
QString L3 = "/CL3_LEEA.cal";
QString L4 = "/CL4_LEEA.cal";
QString H1 = "/CL1_HEEA.cal";
QString H2 = "/CL2_HEEA.cal";
QString H3 = "/CL3_HEEA.cal";
QString H4 = "/CL4_HEEA.cal";

   

void TMCForm::CopyTime()
{
    int bYear = bYearspinBox->value();
    int bDoy = bDoyspinBox->value();
    int eDoy = bDoy;
    int eYear = bYear;
    QTime bTime = btimeEdit->time();
    QTime eTime = bTime;
    eYearspinBox->setValue(eYear);
    eDoyspinBox->setValue(eDoy);
    etimeEdit->setTime(eTime);   
}


void TMCForm::Solve()
{
 
    string Lname = GetLCalName();
    string Hname = GetHCalName();
    string destination = DestlineEdit->text();
    TMC_ERROR retval;
 
    retval = TMC( GetSYear(), GetSDoy(), GetSTimeHour(), GetSTimeMin(), GetSTimeSec(), 0.,
	  GetEYear(), GetEDoy(), GetETimeHour(), GetETimeMin(), GetETimeSec(), 0.,
	  GetSpaceCraft(),
	  GetLowerGuess(), GetUpperGuess(), GetGuessRes(),
	  GetPotentialOffset(), GetSCPotential(), ForcePotential(),
	  GetB(), GetL(), GetH(), GetT(),
	  GetOutputFormat(), &Lname , &Hname, &destination, checkBox2->isChecked(), clobberBox->isChecked());
    
 if( retval==TMC_OK)
 { 
	QString m = "Calculation complete! TMC returned successfully!";
	QMessageBox::about( 0, "QTMC success", m);
	return;
 }
 else
 {
      
       QMessageBox::about( 0, "QTMC error", "Error correcting moments: check console"); 
       cout << "\n*******************************************************************\n";
 } 
  
    
}

void TMCForm::Help()
{
QMessageBox::about( 0, "QTMC about", QString("QTMC v1.0.6: Truncated moments corrector for PEACE data\n\nAuthors:\nJim Geach, Steve Schwartz Astronomy Unit, Queen Mary, University of London\nV. Genot CESR, Toulouse, France.\n\nEnter the data interval (either in YYYY-DDD or YYYY-MM-DD format) to a 1 sec accuracy.\n'>' copies the start time to the end time.\n\nSet the parameters:\n\nLower Guess (%): lower limit of guessing loop\nUpper Guess (%): upper limit of guessing loop\nRes. (%): scan-step of guessing loop\n\nDelta phi (V): how much greater the lower energy limit is than the spacecraft potential\nfor the B moments to be used\n\nSpacecraft: the ID of the spacecraft (1-4)\n\nOutput format: choose between CEF and QFT formats\n\nThe calibration line-edit box is filled with the path of the calibration files,\nthe data is output to the directory in the data destination field.\n\nThe four radio buttons allow the user to construct various combinations\nof summed moments, e.g. BH1H2T\n\nOverride the EFW U_probe_sc by checking the box. If U_probe_sc is not available,\nthe SC potential is used instead.\n\nHit 'Solve' to run the solver.\nSee detailed documentation for a more thorough description. April 2004"));
}

bool TMCForm::GetB()
{
    return BradioButton->state();
}

bool TMCForm::GetL()
{
    return LradioButton1->state();
}

bool TMCForm::GetH()
{
    return HradioButton2->state();
}

bool TMCForm::GetT()
{
    return TradioButton4->state();
}



int TMCForm::GetSYear()
{
    return bYearspinBox->value();
}

int TMCForm::GetSDoy()
{
    return bDoyspinBox->value();
}
 
int TMCForm::GetSTimeHour()
{
    QTime time;
    time = btimeEdit->time();
    return time.hour();
}
 
int TMCForm::GetSTimeMin()
{
    
    QTime time;
    time = btimeEdit->time();
    return time.minute();
}
 
int TMCForm::GetSTimeSec()
{
   QTime time;
    time = btimeEdit->time();
    return time.second();
 }

//
int TMCForm::GetEYear()
{
    return eYearspinBox->value();
}

int TMCForm::GetEDoy()
{
    return eDoyspinBox->value();
}
 
int TMCForm::GetETimeHour()
{
    QTime time;
    time = etimeEdit->time();
    return time.hour();
}
 
int TMCForm::GetETimeMin()
{
    
    QTime time;
    time = etimeEdit->time();
    return time.minute();
}
 
int TMCForm::GetETimeSec()
{
   QTime time;
    time = etimeEdit->time();
    return time.second();
 }


double TMCForm::GetPotentialOffset()
{
    return PhiOffsetlineEdit->text().toDouble();
}

double TMCForm::GetSCPotential()
{
    return SCpotnlLineEdit->text().toDouble();
}

int TMCForm::GetSpaceCraft()
{
  return SCspinBox->value();
}

 
int TMCForm::GetLowerGuess()
{
    return LowerspinBox->value();
}

int TMCForm::GetUpperGuess()
{
    return UpperspinBox->value();
}

int TMCForm::GetGuessRes()
{
    return ResspinBox->value();
}

int TMCForm::GetOutputFormat()
{
    return outputcomboBox->currentItem();
}

char * TMCForm::GetLCalName()
{
    int i;
    char * name = new char[256];
    QString longname;
    i = GetSpaceCraft();
    longname = CalibrationfileslineEdit->text();
    
    if(longname[longname.length()-1]!='/')
		longname+="/";
    
    if(i==1)	
    {longname.append(L1);}
    else if(i==2)
    {longname.append(L2);}
    else if(i==3)
    {longname.append(L3);}   
    else
    {longname.append(L4);}
    
    strcpy(name, longname);
 //   cout << name << endl;
    return name;
}

char * TMCForm::GetHCalName()
{
    int i;
    char * name = new char[256];
    QString longname;
    i = GetSpaceCraft();
    longname = CalibrationfileslineEdit->text();
    
     if(longname[longname.length()-1]!='/')
		longname+="/";
    if(i==1)	
    {longname.append(H1);}
    else if(i==2)
    {longname.append(H2);}
    else if(i==3)
    {longname.append(H3);}   
    else
    {longname.append(H4);}
      
    strcpy(name,longname);
    //cout << name << endl;
    return name;
}

void TMCForm::SetLcal(const char * name)
{    
    QString Lname = name;
  
    CalibrationfileslineEdit->setText(Lname);
}

void TMCForm::SetDest(const char * name)
{    
    QString Dest = name;
  
    DestlineEdit->setText(Dest);
}


void TMCForm::StartDateChange()
{
    int *doy = new int;
    QDate *sdate = new QDate;
    *sdate = StartdateEdit->date();
    
    QanDateToYearDoy(sdate->year(), sdate->month(), sdate->day(), doy);
    
    if((sdate->year() >= 1950) && (sdate->year()<=2100))
    { 
    bYearspinBox->setValue(sdate->year());
    
    bDoyspinBox->setValue(*doy);}  
      
}

void TMCForm::EndDateChange()
{
    int *doy = new int;
    QDate *edate = new QDate;
    *edate = EnddateEdit->date();
    
    QanDateToYearDoy(edate->year(), edate->month(), edate->day(), doy);
    
    if((edate->year() >= 1950) && (edate->year()<=2100))
    {eYearspinBox->setValue(edate->year());
    eDoyspinBox->setValue(*doy);}  
}

void TMCForm::StartYearChange()
{
   int *month = new int;
   int *day = new int;
   int year, doy;
   
   year = TMCForm::GetSYear();
   doy = TMCForm::GetSDoy();
   
   QanYearDoyToDate(year, doy, month, day);
   
   QDate date(year, *month, *day);
   
   StartdateEdit->setDate(date); 
    
}

void TMCForm::EndYearChange()
{
   int *month = new int;
   int *day = new int;
   int year, doy;
   
   year = TMCForm::GetEYear();
   doy = TMCForm::GetEDoy();
   
   QanYearDoyToDate(year, doy, month, day);
   
   QDate date(year, *month, *day);
   
   EnddateEdit->setDate(date); 
    
}

void TMCForm::StartDoyChange()
{
   int *month = new int;
   int *day = new int;
   int year, doy;
   
   year = TMCForm::GetSYear();
   doy = TMCForm::GetSDoy();
   
   QanYearDoyToDate(year, doy, month, day);
   
   QDate date(year, *month, *day);
   
   StartdateEdit->setDate(date); 
    
}

void TMCForm::EndDoyChange()
{
   int *month = new int;
   int *day = new int;
   int year, doy;
   
   year = TMCForm::GetEYear();
   doy = TMCForm::GetEDoy();
   
   QanYearDoyToDate(year, doy, month, day);
   
   QDate date(year, *month, *day);
   
   EnddateEdit->setDate(date); 
    
}

void TMCForm::Destination()
{
    QString s = QFileDialog::getExistingDirectory();
    DestlineEdit->setText(s);
}

void TMCForm::Destination2()
{
    QString s = QFileDialog::getExistingDirectory();
    CalibrationfileslineEdit->setText(s);
}



void TMCForm::closeEvent( QCloseEvent * )
 {
        tmcExit();
 }



void TMCForm::tmcExit()
 {
         QApplication::exit( 0 );
 }




bool TMCForm::ForcePotential()
{

return checkBox1->isChecked();

}


#endif



















