#include "Idfs.h"

void assign_fillval(QdObject_var obj, double val)
{
     char * fillval;   
     sprintf(fillval, "%g", val);
    
     QuSetTxtAttr("FILLVAL", fillval, obj);
     cout << "set xref" << endl;
}

//return a Dim(3) matrix sequence given the scalar sequence components x, y, z
QdRMatrixSeq_var components_to_vector(QdRScalarSeq_var x, QdRScalarSeq_var y, QdRScalarSeq_var z)
{

 //set up obj sequences for joining
 QdObjectSeq_var joined = new QdObjectSeq(3);
 QdObjectSeq_var toJoin = new QdObjectSeq(3);
 QdObjectSeq_var joinOnto = new QdObjectSeq(3);

 //going to join the inputs together
 toJoin[0] = x;
 toJoin[1] = y;
 toJoin[2] = z;

 int i;
 int sizearr[3];
 int size;
 int largeindex=0;

 //get the size of each sequence --> they should already be the same
 sizearr[0] = x->sequence_size();
 sizearr[1] = y->sequence_size();
 sizearr[2] = z->sequence_size();
 size=0;

 //do some checking to see which is largest seq 
 for(i=0; i<3; i++)
 { 
	  if(sizearr[i] > size)
		size = sizearr[i];
 }
	
 for(i=0; i<3; i++)
 {
   if(sizearr[i]==size)	
 	 largeindex=i;
 }

 //join onto the largest sequence 
 if(largeindex==0){ joinOnto[0] = x; joinOnto[1] = x; joinOnto[2] = x; }
 else if(largeindex==1){ joinOnto[0] = y; joinOnto[1] = y; joinOnto[2] = y; }
 else{ joinOnto[0] = z; joinOnto[1] = z; joinOnto[2] = z; }
	
 //perform join, return null on error
 if(join_components(toJoin, joinOnto, joined)!=0)	
 return NULL;

 //create vector the size of joined sequences and Dim(3) 
 QdRMatrixSeq_var vector = new QdRMatrixSeq(joined[0]->sequence_size(), Dim(3));

 //set up joined x, y, z 
 QdRScalarSeq_var jx = new QdRScalarSeq(joined[0]->sequence_size());
 QdRScalarSeq_var jy = new QdRScalarSeq(joined[1]->sequence_size());
 QdRScalarSeq_var jz = new QdRScalarSeq(joined[2]->sequence_size()); 

 //narrow down from joined object seq 
 jx = QdRScalarSeq_var::narrow(joined[0]);
 jy = QdRScalarSeq_var::narrow(joined[1]);
 jz = QdRScalarSeq_var::narrow(joined[2]); 

//scan through sequence assigning vector[i][x, y, z] with joine x, y, z
for(i=0; i<joined[0]->sequence_size(); i++)
{
  vector[i][0] = (double)jx[i];
  vector[i][1] = (double)jy[i];
  vector[i][2] = (double)jz[i];
}

 //get tags from x 
 set_timetags( vector, get_timetags( jx ) );

 //set the fill value to the usual
 QdObject_var fillobj = new QdRScalar(-3.4e38);

 try
 {
	    if(vector->xref_exists("FILLVAL"))
	          vector->delete_xref("FILLVAL"); 
	 
    	vector->set_xref("FILLVAL", fillobj);
 }
 catch(Exception &e){return NULL;} 
	
 //clean the matrix sequence  
 QdRMatrixSeq_var cvector = QdRMatrixSeq_var::narrow(QuCleanObjectMatrix(vector)); 

 //output the result 
 return cvector;
        
}

//routine to retrieve data described by start, end, lineage, etc.
//returns a clean (if clean==true) scalar seq var
QdRScalarSeq_var ret_obj(IDFS_TIME start, IDFS_TIME end, IDFS_LINEAGE lin, SDDAS_SHORT sensor, SDDAS_CHAR type, bool clean)
{

    //create our idfs data structure
    struct idf_data *EXP_DATA;

    //perform all the necessary idfs initialization - see IDFS docs for
    //more info
    CfgInit(); 
  
    SDDAS_ULONG data_key;
    SDDAS_SHORT ret_val;

    char more_data = 1;

    dbInitialize();
    init_idfs();

    IDFS_TIME ctime;
    IDFS_TIME dtime;
    
    to_idfs_time(0, 0, 0, 0, 0, 0, &ctime);

    void *idf_data_ptr;

    strcpy(lin.exten,"");

    ret_val = create_idf_data_structure(&idf_data_ptr);

    if(ret_val != ALL_OKAY){printf("\nError %d from create_idf_data_structure routine.\n", ret_val); return (NULL);}

    //promote the data described by the interval and lineage
    ret_val = PROMOTE(start, end, lin, &data_key);
   
   //do again to guard against known problem of having
   //to promote twice 
    ret_val = PROMOTE(start, end, lin, &data_key);

    EXP_DATA = (struct idf_data *) idf_data_ptr;

    //initialize idfs file, note that failures will result
    //in a null ptr
    ret_val = init_idf_file(start, end, data_key, EXP_DATA, lin.vsn);
    if(ret_val != ALL_OKAY){return (NULL);}
    
    int i = 0;

    //our data value
    double *val = new double;
    	    
    //more idfs initialization
    int month, day, hp, mp, sp;
    long np;
    int count=0;
    StringRep S;
    SDDAS_SHORT rcode;
    SDDAS_LONG lfill;
    SDDAS_CHAR chkfill = 0;
    SDDAS_USHORT vsn2;
    SDDAS_SHORT ridfs_rcode;
    
    //check info about fill values
    ridfs_rcode = read_idf(data_key, "", vsn2, &chkfill, _FILL_FLG, 0, 0, 1);

    if(chkfill==1)
      ridfs_rcode = read_idf(data_key, "", vsn2, &lfill, _FILL, 0, 0, 1);
	  else
	    lfill = 0;

    //this dummy stuff is for our first pass when we determine
    //how many data points there are -- so we can create a scalar
    //sequence of the correct size
    QdTime_var dummy_t;
    QdTimeSeq_var dummy_time_line = new QdTimeSeq(3);
    int k=0;
    //read data into a dummy
    double *dummyval = new double;
    
while(more_data) //scan while there's data, count each pass
{
	 
   if(RETURN_DATA(dummyval, &data_key, start, end, sensor, EXP_DATA, idf_data_ptr, &lin, more_data, &ret_val, &dtime, type, chkfill, lfill)==-1)
    break;
   
    //get correct time (including leap years, etc)
    to_hms_time(&dtime, &hp, &mp, &sp, &np);
   	QanYearDoyToDate( dtime.year, dtime.doy, &month, &day );
	  dummy_t = new QdTime((int)dtime.year, month, day, (double)hp, (double)mp, (double)sp+((double)np/1.e9) );
    dummy_time_line[k] = *dummy_t;  

   //this logic protects against timetags that are equal	
   if(k==0)
   {
	  if(NEXT_FILE(ret_val, &data_key, lin, idf_data_ptr, &end, more_data) != 1)
    break;
    k++; 
    count++;	 
   }
   else if(k>0 && (*dummy_t != dummy_time_line[k-1])  )
   {
	   if(NEXT_FILE(ret_val, &data_key, lin, idf_data_ptr, &end, more_data) != 1)
	   break;
	   k++;
	   count++;
   }
   else 
   {
    	if(NEXT_FILE(ret_val, &data_key, lin, idf_data_ptr, &end, more_data) != 1)
	    break;
	    k++;   
   }
   
    if(k==3)
	      k=1;  

}
    //now we know how many points, get ready for our proper pass--
    //this time is the same as above, but we read data into the ith element
    //of a sequence

    QdTime_var t;
    QdTimeSeq_var time_line = new QdTimeSeq(count);
    QdRScalarSeq_var object = new QdRScalarSeq(count);
    

    //reset
    rcode = reset_experiment_info(data_key,lin.exten,lin.vsn,start.year, 
		                              start.doy,start.sec,start.nano,end.year, 
				                          end.doy, end.sec, end.nano);
				  
    //position file
    rcode = file_pos(data_key,lin.exten,lin.vsn,idf_data_ptr, 
                     start.year,start.doy,start.sec,start.nano, 
            		     end.year, end.doy, end.sec,end.nano);

    //prepare for fill values -- standard value
    QdObject_var fillobj = new QdRScalar(-3.4e38);
    
while(more_data)
{
   if(RETURN_DATA(val, &data_key, start, end, sensor, EXP_DATA, idf_data_ptr, &lin, more_data, &ret_val, &ctime, type, chkfill, lfill)==-1)
     break;

  to_hms_time(&ctime, &hp, &mp, &sp, &np);
	QanYearDoyToDate( ctime.year, ctime.doy, &month, &day );
  t = new QdTime((int)ctime.year, month, day, (double)hp, (double)mp, (double)sp+((double)np/1.e9) );

  //same as before -- uncomment the cout lines to print out
  //what is being retrieved -- useful for finding bugs
  
  if(i==0)
   {
	    time_line[i] = *t;
	    object[i] = *val;
	  //  cout << S(t) << '\t' << (double)object[i] << endl;
	       if(NEXT_FILE(ret_val, &data_key, lin, idf_data_ptr, &end, more_data) != 1)
	           break;

	    i++;

   }
   else if((i>0) && (*t != time_line[i-1]) && (i<count)  )
   {

	  //  cout << "*t != time_line[i-1]\n";
	    time_line[i] = *t;
	    object[i] = *val;
	  //  cout << S(t) << '\t' << (double)object[i] << endl;
	       if(NEXT_FILE(ret_val, &data_key, lin, idf_data_ptr, &end, more_data) != 1)
	           break;

	    i++;
   }
   else
   {

     //  cout << S(t) << '\t' << (double)object[i] << endl;
	      if(NEXT_FILE(ret_val, &data_key, lin, idf_data_ptr, &end, more_data) != 1)
	        break;
  }



}
    //now to set the tags
    set_timetags( object, time_line );

    //free up
    free_experiment_info();

    //if desired, perform the clean operation
    if(clean==true)
    {
     if((int)chkfill==1)
     {
      try
      {
		      if(object->xref_exists("FILLVAL"))
		       object->delete_xref("FILLVAL");

		      object->set_xref("FILLVAL", fillobj);
	    }
	    catch(Exception &e){return NULL;}
	   }

    QdRScalarSeq_var rObj = QuCleanObject(object);

    //check for non conformities
    QdRScalarSeq_var rcObj = QdRScalarSeq_var::narrow(idfs_non_conformity(rObj));
    //return clean object
    return rcObj;
    }
    else //check for non-conformities and return object
      return QdRScalarSeq_var::narrow(idfs_non_conformity(object));

}

//return an object which is free from non-conformities
QdObject_var idfs_non_conformity(QdObject_var in)
{

//get input sequence size
int size = in->sequence_size();
int y1, y2, d1, d2, h1, h2, m1, m2;
double s1, s2;
int newsize =0;
double dspace;

    //create mask, all true to start
    SequenceMask mask(size, TRUE);

    //make a copy of the input
    QdRScalarSeq_var scalarin = QdRScalarSeq_var::narrow(in);

    //set up some Time variables
    Time timeY;
    Time timeZ;
    Time initial;
    Time tmp;

    //create a time sequence of the correct size
    QdTimeSeq_var timeseq = new QdTimeSeq(size);
    //set the tags from input
    timeseq = get_timetags(in);
    //get the spacing of tags
    if ( !QuIsTSRegular( timeseq , &dspace ) );

    QdTimeSeq_var line = QdTimeSeq_var::narrow(timeseq);

    for(int i=0;i<size-1;i++)
    {
      //time at point A
      timeY = line[i];
      //time at point A+spacing
	    timeZ = line[i+1];
      //initial tag
	    initial = line[0];

      //split times into constituent parts
    	timeY.split_ydoy(y1, d1, h1, m1, s1 );
    	timeZ.split_ydoy(y2, d2, h2, m2, s2 );

    //if next space is before initial time, or previous tag
    if( (timeZ<initial) || (timeZ < timeY)  )
	 	{

      //then mark that point as false
			mask[i+1]=FALSE;

      //check subsequent points to see if a clump of tags
      //are where they shouldn't be
			for(int j=1;j<(size-1-i);j++)
			{
				tmp = line[i+j+1];
				if( (tmp < timeY) )
				{mask[i+j+1]=FALSE; i+=j;}
				else{break;}
     	}//break at end of bad clump


		}
		else //otherwise we can use that point
		{
			mask[i+1]=TRUE; newsize++; //keep a track of the number of good points

		}

    }
            //new output object
    	      QdObject_var out;

            //scalar sequence of correct size
            QdRScalarSeq_var outscalar = new QdRScalarSeq(newsize);

      //apply mask and get subsequence from the input series
      QdTimeSeq_var tt_out = new QdTimeSeq();
	    timeseq->get_subsequence(*tt_out, mask);
      scalarin->get_subsequence(*outscalar, mask);
      set_timetags(outscalar, tt_out);
      out = outscalar;

      //return the de-non-conformitied object
      return out;

}


//routine to read a calibration file and return a scalar
//sequence of gain correction factors
QdRScalarSeq_var read_n_rtn_cal_file(string calname)
{
    int count;
    int i=0;

    //open file calname -- the calibration file
    ifstream calfile(calname.c_str());

    //report if can't open file
    if(!calfile.is_open())
    {cout << "\nError locating calibration file!\n"; return NULL;}
    
    //count how many gain corrections are in file
    while(!calfile.eof())
    {
	    calfile.ignore(256, '\n');
	    i++;
    }
    
    count = i-1; //because included header line in calfile
    
    calfile.close();

    //create a scalar seq of size count   
    QdRScalarSeq_var cal = new QdRScalarSeq(count);
    QdTimeSeq_var tl = new QdTimeSeq(count);
    
    //the times come from the file, and are read into buffer
    //there is also some other data which we don't need
    //we read this into a buffer and don't use it
    string t;
    string buffer;
    double cal_d;

    //reset i
    i = 0;

    //open again
    ifstream calin(calname.c_str());

    //ignore the first line -- header line
    calin.ignore(256, '\n');
    
    while(i<count)
    {
	    //read time
	    calin >> t;
	    //read junk
	    calin >> buffer;
      //read factor
	    calin >> cal_d;
	    //read more junk
    	calin >> buffer;

      //assign factor to sequence
	    cal[i] = cal_d;
      //and time
	    tl[i] = t;
	    //move on
	    i++;
	 }

    //set the timetags
    set_timetags(cal, tl);

    //return the object
    return cal;
}

//this function may be redundant -- will look into it
//but it joins a[i] onto b[i] and creates aj[i]
int join_components(QdObjectSeq_var a, QdObjectSeq_var b, QdObjectSeq_var aj)
{
    
    int triage_cnt = 0;
    int i = 0;

    //time line of the time series to be joined 
    QdObjectSeq_var ttIn = new QdObjectSeq(3);   
    if ( ttIn.is_nil() )
    	return 1; //error
        
    //set of time series to be joined
    QdObjectSeq_var toJoin = new QdObjectSeq(3);   
    if ( toJoin.is_nil() )
    	return 1;
        
    //join method
    QdObjectSeq_var joinMethod = new QdObjectSeq(3);   
    if ( joinMethod.is_nil() )
    	return 1;
        
    //gap value if the time series is not regular     
    QdObjectSeq_var gapValue = new QdObjectSeq(3); 
    if ( gapValue.is_nil() )
    	return 1;
        
    //gap handler if the time series is not regular
    QdObjectSeq_var gapHandler = new QdObjectSeq(3);   
    if ( gapHandler.is_nil() )
    	return 1;
        
    //tags to join onto
    QdTimeSeq_var join_pts = get_timetags( (*b)[i] );
    if (join_pts.is_nil())
    	return 1;
        
    //a copy
    QdTimeSeq_var ttJoined = new QdTimeSeq( *join_pts );
    
    double dSpacing  = 0.;
    
    QdTimeSeq_var toJoinTags;  // for the sake of initialization
    
    
    //remove gaps
    triage_cnt = 3;

    //linear interp
    for(i=0; i<3; i++)
    {
    	toJoin[i]     = ( (a)[i] ); 
	    ttIn[i]       = get_timetags( (a)[i] );
	    aj[i]        = toJoin[i]->create();
	    toJoinTags    = get_timetags( (a)[i] );                                                                                                                     
	    joinMethod[i] = new QdString(LINEAR);                
	    if ( !QuIsTSRegular( toJoinTags , &dSpacing ) );      					 		    
	    gapValue[i]   = new QdTime(Seconds(1.1*dSpacing));  
    }  		                               

    //hand the results to the multijoin function												     
    try 
    {
	    multijoin( ttIn, toJoin, ttJoined, aj, joinMethod, gapValue, gapHandler, triage_cnt);
    }
    catch( Exception &e ) 
    {
	    e.print_msg();
	    return 1;
    }

    //copy cross-refs
    for(i=0; i<3; i++)
    {
    	(aj[i])->copy_xrefs_from( toJoin[i] );
	    set_timetags( aj[i], ttJoined );
    }

    //return all ok
    return 0;
}















