#include "idfs_retrieval.h"

/* updated July 2007 to make tbl_oper an SDDAS_LONG - sjs */

//function to promote data between start and end described by lineage
SDDAS_SHORT PROMOTE(IDFS_TIME start, IDFS_TIME end, IDFS_LINEAGE lineage, SDDAS_ULONG *data_key)
{

//set up necessary idfs stuff
SDDAS_USHORT version;
SDDAS_SHORT  ret_val;
void *promote_struct;
get_version_number(&version); 
lineage.vsn = version;

//get the data key for the lineage
ret_val = get_data_key(lineage.project, 
                       lineage.mission, 
		                   lineage.experiment, 
		                   lineage.instrument, 
		                   lineage.virtual_inst, 
		                   data_key);
if(ret_val != ALL_OKAY)
	{printf("\nError %d from get_data_key routine.\n", ret_val); return(-1);}

//create promotion struct
promote_struct = pro_InitPromote(start.year, 
                                 start.doy, 
                        				 1000*start.sec,  //promote sec are msecs!!!!
				                         end.year, 
				                         end.doy, 
                        				 1000*end.sec, 
				                         sFalse);

pro_AddSource(promote_struct, *data_key);

//do we need to promote?
if(pro_NeedToPromote(promote_struct) == sTrue)
{//yes, so do
 pro_PromoteAll(promote_struct);
}

//free up
pro_FreePromote(promote_struct);

//return error code
return ret_val;

}

//simple function which assigns the structure IDFS_LINEAGE with predefined
//strings (see tmc_data.cpp)
void ASSIGN_LINEAGE(IDFS_LINEAGE *lin, const char * array[5])
{
  
strcpy(lin->project,     array[0]);
strcpy(lin->mission,     array[1]);
strcpy(lin->experiment,  array[2]);
strcpy(lin->instrument,  array[3]);
strcpy(lin->virtual_inst,array[4]);
strcpy(lin->exten, "");

}

//large routine to put data at a given time into val -- data
//described by lineage, interval, data key, etc
int RETURN_DATA(double *val, SDDAS_ULONG *data_key, IDFS_TIME start, IDFS_TIME end, SDDAS_SHORT sensor, struct idf_data *exp_data, void * idfptr,IDFS_LINEAGE *lin, char more_data, SDDAS_SHORT * ret_val, IDFS_TIME *ctime, SDDAS_CHAR type, SDDAS_CHAR chkfill, SDDAS_LONG lfill)
{

//again, set up idfs stuff
SDDAS_CHAR  num_tbls         = 0;
SDDAS_CHAR  *tbls_to_apply   = NULL;
SDDAS_LONG *tbl_oper        = NULL;
SDDAS_CHAR full_swp = 1, fwd = 1;
SDDAS_LONG ret_time_sec;
SDDAS_LONG ret_time_nano;
SDDAS_SHORT rcode;

//make sure we have enough mem
SDDAS_FLOAT *conv_data = (SDDAS_FLOAT *)malloc(sizeof(SDDAS_FLOAT));

   

  //read data record
  *ret_val = read_drec(*data_key, lin->exten, lin->vsn, idfptr, sensor, fwd, full_swp);
	if(*ret_val < 0){printf("\nError %d from read_drec routine.\n", *ret_val); return(-1);}
	
  //all this checks we are in range
  if(*ret_val == ALL_OKAY || exp_data->filled_data)
	{
		ret_time_sec = (exp_data->bmilli + (exp_data->bnano / 1000000)) / 1000;
		ret_time_nano= (exp_data->bmilli % 1000) * 1000000 + exp_data->bnano;

		
if((exp_data->byear > end.year) || 
  ((exp_data->byear == end.year) && (exp_data->bday > end.doy)) ||
  ((exp_data->byear == end.year) && (exp_data->bday == end.doy) && (ret_time_sec > end.sec)) ||
  ((exp_data->byear == end.year) && (exp_data->bday == end.doy) && (ret_time_sec == end.sec) && (ret_time_nano > end.nano)))
		  {
		  	more_data = 0;
		   	return -1;
		  }
		
      //put the data into the correct units
      rcode = convert_to_units(*data_key, lin->exten, lin->vsn, idfptr, 
		                          sensor, type , 0, num_tbls, tbls_to_apply, 
					                     tbl_oper, conv_data, chkfill, lfill );
		 
		if (rcode != ALL_OKAY){ printf("\nError %d from convert_to_units.\n", rcode); return(-1); }

    //set current time
		ctime->year = exp_data->byear;
		ctime->doy = exp_data->bday;
		ctime->sec = ret_time_sec;
		ctime->nano = ret_time_nano;

    //assign data value at current time
		*val = (double)*conv_data;
    //return ok
    return 0;
	
	}

  //return error
	return -1;
	
}

//function which handles the transition between idfs files
SDDAS_SHORT NEXT_FILE(SDDAS_SHORT ret_val, SDDAS_ULONG *data_key, IDFS_LINEAGE lin, void * idfptr, IDFS_TIME *etime, char more_data)
{

SDDAS_LONG  new_start_sec, new_start_nsec;
SDDAS_SHORT rcode, new_year, new_day;

//check status of the last error code which indicate whether we need
//to go to the next file
if(ret_val == LOS_STATUS || ret_val == NEXT_FILE_STATUS)
	{
    //if yse then get the start time of the next file
    rcode = next_file_start_time(*data_key, lin.exten, lin.vsn, 0, &new_year, 
		                             &new_day, &new_start_sec, &new_start_nsec);

      	
		if(rcode != ALL_OKAY){printf("\n Error %d from next_file_start_time.\n", rcode); 		                                      
		    return (SDDAS_SHORT)(-1);}

    //perform the necessary reset
		rcode = reset_experiment_info(*data_key, lin.exten, lin.vsn, new_year, 
		                              new_day, new_start_sec, new_start_nsec, 
		                  			      etime->year, etime->doy, etime->sec, etime->nano);

	  //check if there is data
		if(rcode == NO_DATA){more_data = 0; return (SDDAS_SHORT)(-1);}
		else if(rcode != ALL_OKAY){printf("\n Error %d from reset_experiment_info. \n", rcode);                                                      
		    return (SDDAS_SHORT)(-1);}
		
		//position file and we're ready to go!
 		rcode = file_pos(*data_key, lin.exten, lin.vsn, idfptr, new_year, new_day, 
		                 new_start_sec, new_start_nsec, etime->year, etime->doy, etime->sec, etime->nano);

		if(rcode != ALL_OKAY){printf("\n Error %d from file_pos.\n", rcode); return (SDDAS_SHORT)(-1);}
	
	 }

//ok

return 1;

}


