#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>
#include <string.h>
#include "tmc_routines.h"

/*

main function for non-Qt use (i.e. just batch mode)

*/



void print_usage();
void time_scan(const char * str, int *y, int *d, int *h, int *m, int *s, int *n);
void sum_scan(const char * str, int *SC, bool *B, bool *L, bool *H, bool *T);
int get_int(const char * str);
double get_double(const char *str);
string get_string(const char* str);

int main( int argc, char ** argv )
{
    int sy = 2001 ;
    int sd = 1;
    int sh = 0;
    int sm = 0;
    int ss = 0;
    int sn = 0;

    int ey = sy;
    int ed = sd;
    int eh = sh;
    int em = sm;
    int es = ss;
    int en = sn;

    int sc = 1;
    bool b = 1;
    bool l = 1;
    bool h = 0;
    bool t = 1;

    double gl = -200.;
    double gu = 200.;
    double gr = 1.;
    double delta_phi = 0.0;
    double sc_potential = 7.0;

    string oefw;
    bool override_efw = 0;
    string location, destination;
    int output = 1;
    string cef_or_qft;
    string lcal;
    string hcal;
    bool logging = true;
    bool clobber = false;
    bool history = false;

    const char * cfgname = new char[256];
    if(argc==1)
    {

    print_usage();
    return 1;

    }
    else
    {

    for (int i = 1; i < argc; i++)
  	{

	/* Check for a switch (leading "-"). */
        if (argv[i][0] == '-') {

	    /* Use the next character to decide what to do. */

	    switch (argv[i][1]) {

		  case 'S':	time_scan(argv[i], &sy, &sd, &sh, &sm, &ss, &sn); //start time
								break;

		  case 'E':	time_scan(argv[i], &ey, &ed, &eh, &em, &es, &en); //end time
				        break;

		  case 'M':	sum_scan(argv[i], &sc, &b, &l, &h, &t); //sc & moment sum
			  				break;

		  case 'O':	cef_or_qft = get_string(argv[i]);
				if(cef_or_qft[1]=='e')
					output = 1;
				else if(cef_or_qft[1]=='d')
          output = 2;
        else
          output = 0;

				 break; //output


		  case 'C': location = get_string(argv[i]);
		    				 break;            //calibration file location


		  case 'D':	destination = get_string(argv[i]);
				         break;	     //output destination

		  case 'l':	gl = get_double(argv[i]);
		             break;	      //lower guess

		  case 'u':	gu = get_double(argv[i]);
				         break;	     //upper guess

	  	case 'r':       gr = get_double(argv[i]);
  	             break;       //guess loop res

		  case 'V':       sc_potential = get_double(argv[i]);
			           break;       //sc potential

		  case 'U':	 delta_phi = get_double(argv[i]);
			           break;	     //delta phi

	  	case 'o':       oefw = get_string(argv[i]);

				if(oefw[0]=='n')
					{override_efw = 0;}
				else
					override_efw = 1;

			           break;       //override efw?

      case '-':  switch(argv[i][2])
                 {
                     case  'r': logging = false; break;

                     case 'c': clobber = true; break;

                     case 'h': history = true; break;

                     default: cout << "Unknown flag" << endl; print_usage(); return 1;

                 }

                  break;

		default: cout << "Unknown flag" << endl; print_usage();
			return 1;


	    			}

				}
				else{ cout << "Unknown flag" << endl; print_usage(); return 1;}
		//	cout << "\r";
	}

	if(location[location.size()-1]!='/')
		location+="/";

	switch ( sc )
	{

	case 1:  lcal = location+"CL1_LEEA.cal";
		hcal = location+"CL1_HEEA.cal";
		break;

	case 2: lcal = location+"CL2_LEEA.cal";
		hcal = location+"CL2_HEEA.cal";
		break;

	case 3:  lcal = location+"CL3_LEEA.cal";
		hcal = location+"CL3_HEEA.cal";
		break;

	case 4: lcal = location+"CL4_LEEA.cal";
	 	hcal = location+"CL4_HEEA.cal";
		 break;

	default: break;

	}

	TMC_ERROR e;

  if(history==false)
  	{
      e = TMC(sy, sd, sh, sm, ss, sn, ey,ed,eh, em, es, en, sc, gl, gu, gr, delta_phi, sc_potential, override_efw, b, l, h, t, output, &lcal, &hcal, &destination, logging, clobber);
      	if(e!=TMC_OK)
      	{
	       cout << "\n*******************************************************************\n";
	      }
     }
	else
    e = TMC_HISTORY(sy, sd, sh, sm, ss, sn, ey, ed, eh, em, es, en, sc);



	return 1;

    }

}




string get_string(const char * str)
{

char buff[50];
string output;

for(int i=0; i<strlen(str); i++)
	buff[i] = str[i+2];

output = buff;

return output;

}

int get_int(const char * str)
{

char * buff;
int x;
int length = strlen(str);

for(x=0; x<length; x++)
 buff[x] = str[x+2];

 return atoi(buff);
}

double get_double(const char * str)
{

/*char * buff;
int x;
int length = strlen(str);

for(int i=0; i<length; i++)
 buff[i] = str[i+2];

 return atof(buff); */

      //char * buff;

int length = strlen(str);

char buff[length];



for(int i=2; i<=length; i++)
{
	buff[i-2] = str[i];
}

 return atof(buff);


}


void sum_scan(const char * str, int *SC, bool *B, bool *L, bool *H, bool *T)
{


*B = 0;
*L = 0;
*H = 0;
*T = 0;

char sc[2];
sc[0] = str[2];
sc[1] = '\0';

*SC = atoi(sc);

for(int i=0; i<strlen(str); i++)
{


if(str[i]=='B' || str[i]=='b')
	*B = 1;
if(str[i]=='L' || str[i]=='l')
	*L = 1;
if(str[i]=='H' || str[i]=='h')
	*H = 1;
if(str[i]=='T' || str[i]=='t')
	*T = 1;

}

return;

}

void time_scan(const char * str, int *y, int *d, int *h, int *m, int *s, int *n)
{

int length = strlen( str );
char year[] = "2001";
char doy[] = "001";
char month[] = "01";
char day[] = "01";
char hour[] = "00";
char min[] = "00";
char sec[] = "00";
char nano[] = "000";

int dd, mm;

int YEAR, DOY, HOUR, MIN, SEC, NANO;

if(length==23)
{

year[0]=str[2];  year[1]=str[3]; year[2]=str[4]; year[3]=str[5]; year[4]='\0';

doy[0]=str[7];   doy[1]=str[8];  doy[2]=str[9];

hour[0]=str[11]; hour[1]=str[12];

min[0]=str[14];  min[1]=str[15];

sec[0]=str[17];  sec[1]=str[18];

nano[0]=str[20]; nano[1]=str[21]; nano[2]=str[22];


YEAR = atoi(year);
DOY = atoi(doy);
HOUR = atoi(hour);
MIN = atoi(min);
SEC = atoi(sec);
NANO = atoi(nano);

*y = YEAR; *d = DOY; *h = HOUR; *m = MIN; *s = SEC; *n = NANO;

return;

}
else if(length==25)
{

year[0]=str[2];  year[1]=str[3]; year[2]=str[4]; year[3]=str[5];
month[0]=str[7];   month[1]=str[8];
day[0]=str[10]; day[1]=str[11];
hour[0]=str[13]; hour[1]=str[14];
min[0]=str[16];  min[1]=str[17];
sec[0]=str[19];  sec[1]=str[20];
nano[0]=str[22]; nano[1]=str[23]; nano[2]=str[24];

YEAR = atoi(year);
mm = atoi(month);
dd = atoi(day);
HOUR = atoi(hour);
MIN = atoi(min);
SEC = atoi(sec);
NANO = atoi(nano);

QanDateToYearDoy(YEAR, mm, dd, d);
*y = YEAR; *h = HOUR; *m = MIN; *s = SEC; *n = NANO;

return;

}
else
{

print_usage();
return;

}




}



void print_usage()
{
cout << "\nQTMC: truncated moments corrector. USAGE:\n\n";
    cout << "**************************************************************************\n";
    cout << "\nGUI mode: QTMC -C{/cal_file/location} -D{/data/destination}\n";
    cout << "Batch mode: QTMC -S{YYYY-DDDZHH:MM:SS.SSS} (or -S{YYYY-MM-DDZHH:MM:SS.SSS})" << endl;
    cout << "                 -E{YYYY-DDDZHH:MM:SS.SSS} (or -E{YYYY-MM-DDZHH:MM:SS.SSS})" << endl;
    cout << "                 -l{lower guess [-200]} -u{upper guess [200]} -r{guess resolution [1]}" << endl;
    cout << "                 -O{cef, qft or cdf [cef]} -o{override EFW yes/no [no]} -V{SC potential [7.0]}" << endl;
    cout << "                 -U{delta phi [0.0]} -M{SCBL1L2T, etc. [1]}" << endl;
    cout << "                 -C{/calibration location} -D{/destination} (--reticent will turn of logging --clobber will overwrite existing files)" << endl;
    cout << endl;
    cout << "**************************************************************************\n";

}
