//  CDFDataBaseDataSource.cc
//
//  Qdos - data interfacing
//
//  David Burgess
//  March 2000
//

#include <map>
#include "CDFDataBaseDataSource.h"
#include "CDFDataSource.h"

namespace QDOS {

void CDFDataBaseDataSource::close()
{
}

// +++++++++++++++++++++++++++++++++++++++++++++++ class CDFDataBaseDataSource
// :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- is_data_available
//
bool CDFDataBaseDataSource::is_data_available(const string& data_source_name,\
      const TimeInterval& tint)
{
  /*  Name n( data_source_name );
    if( n.num_components() > 1 )
    {
        if(n[0]==DATA)
        {
            Name n2(n);   
            n2.delete_component(0);
            return cdf_data_configmap->is_data_available(n2.srep(),tint);
        }            
        
    }
    else if( n[0] == METADATA )
    {    
        CDF_DCMMetaDataEntry* metadata_entry =\
            cdf_data_configmap->get_metadata_entry( data_source_name );
        return (metadata_entry!=0);
    }
    else 
    */
    return cdf_data_configmap->is_data_available(data_source_name,tint);  
}
// ++++++++++++++++++++++++++++++++++++++ class CDFDataBaseDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- get_data
//

DataBuffer_var CDFDataBaseDataSource::get_data(
     const string& data_name,
     const TimeInterval& time_interval )
{
  DataBuffer_var nil_object; // for nil return
  Name n( data_name );
  if( n.num_components() > 1 )
  {
    Name n2(n);
    n2.delete_component(0);
    if( n[0] == DATA )
    {
      return get_data_data( n2.srep(), time_interval );
    
    } else if( n[0] == METADATA )
    {
        return get_meta_data( n2.srep() );
    }
  }
  throw NoSuchDataSourceName(
    "CDFDataBaseDataSource::get_data: data name=" + data_name );
  return nil_object;  
}
// ++++++++++++++++++++++++++++++++++++++ class CDFDataBaseDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- get_data
//

DataBuffer_var CDFDataBaseDataSource::get_data( const string& data_name )
{
  DataBuffer_var nil_object; // for nil return
  Name n( data_name );
  if( n.num_components() > 1 )
  {
    Name n2(n);
    n2.delete_component(0);
    if( n[0] == DATA )
    {
      throw CDFDataBaseException(
            "CDFDataBaseDataSource::get_data: Time series data requested with no time interval");
      return nil_object;
    
    } else if( n[0] == METADATA )
    {
        return get_meta_data( n2.srep() );
    }
  }
  throw NoSuchDataSourceName(
    "CDFDataBaseDataSource::get_data: data name=" + data_name );
  return nil_object;  
}
// ++++++++++++++++++++++++++++++++++++++ class CDFDataBaseDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- get_names
DataNamesTable_var CDFDataBaseDataSource::get_names(void)
{
    if( data_names.is_nil() ){
        get_data_names();
        get_meta_names();
    }
    return data_names;
}       
// ++++++++++++++++++++++++++++++++++++++ class CDFDataBaseDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------- get_data_names
DataNamesTable_var CDFDataBaseDataSource::get_data_names(void)
{
  DataNamesTable_var data_data_names;
  if(!data_names.is_nil())
  {
     try{
       data_data_names=data_names->resolve_context(DATA);
     }
     catch(Exception& e)
     {;}
  }
  else 
     data_names=new DataNamesTable;
      
  if( data_data_names.is_nil() )
  { 
     data_data_names=new DataNamesTable;
     
    // get names of ts data sources
    CDF_DCMDataSourceTable_var data_table=\
          cdf_data_configmap->get_data_source_table();
    
    CDF_DCMDataSourceTable::recursive_iterator itr;
    for( itr=data_table->recursive_begin();
      itr!=data_table->recursive_end();itr++)
        if(itr.type()==N_OBJECT)
          data_data_names->bindx( itr.name(), "" );
    
    data_names->bind_context(DATA,data_data_names);
  }
  
  return data_data_names;
}
// ++++++++++++++++++++++++++++++++++++++ class CDFDataBaseDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------- get_meta_names
DataNamesTable_var CDFDataBaseDataSource::get_meta_names(const string& data_name)
{
  DataNamesTable_var meta_names;
  if(!data_names.is_nil())
  {
     try{
       meta_names=data_names->resolve_context(METADATA);
     }
     catch(Exception& e)
     {;}
  }
  else 
     data_names=new DataNamesTable;
      
  CDF_DCMMetaDataTable_var metadata_table = \
    cdf_data_configmap->get_meta_data_table();
  if(metadata_table==0)
    throw CDFDataBaseException("[get_meta_data]:No MetaData Table");
  
  if( meta_names.is_nil() )
    meta_names=new DataNamesTable; 
           
  // Get data source names 
  DataNamesTable_var data_data_names=get_data_names();
  TimeInterval tint;
  get_time_range(tint);
  vector<string> data_files;
  vector<string> data_root_names;
  // Get unique list of first file for each data source
  for(DataNamesTable::recursive_iterator itr=\
   data_data_names->recursive_begin();\
   itr!=data_data_names->recursive_end();itr++)
  {
    if(itr.type()!=N_OBJECT)
      continue;
    // only requested data source if supplied
    if(data_name!="")
    {
      if(itr.name()!=data_name)
        continue;
    }
    // ignore data sources for which meta names already obtained 
    try{
      (void)meta_names->resolve_context(itr.name());  
      continue;
    }
    catch(NameNotFound& e)
    {;}
    DataNamesTable_var fnames=get_datafile_names(itr.name(),tint);
    if(!fnames.is_nil())
    {
      vector<string> fnames_list;
      fnames->list_names_r(fnames_list,"/");
      int i;
      for(i=0;i<fnames_list.size();i++)
      {
        try{
          CDFFileDataSource cdf_ts(fnames_list[i]);
          break;
        }
        catch(Exception& e)
        {;}
      }
      if(i!=fnames_list.size())
      {
        vector<string>::iterator sitr=\
          find(data_files.begin(),data_files.end(),fnames_list[i]);
        if(sitr==data_files.end())
        {
          data_files.push_back(fnames_list[i]);
          Name n(itr.name());
          n.delete_component(n.num_components()-1);
          data_root_names.push_back(n.srep());
        }
      }
    }
  }
  // Get file and data meta data 
  vector<string> root_names(2);
  root_names[0]=VARIABLE_ATTRIBUTES;
  root_names[1]=GLOBAL_ATTRIBUTES;
  for(int i=0;i<data_files.size();i++)
  {
    try{
      int offset=(data_files[i].substr(0,1)=="/")?1:0;
      DataNamesTable_var file_data_names;
      CDFFileDataSource cdf_file(data_files[i]);
      file_data_names=cdf_file.get_data_names();
      if(!file_data_names.is_nil())
      {
        DataNamesTable_var file_meta_names;
        for(int j=0;j<root_names.size();j++)
        {
          try{
            file_meta_names=file_data_names->resolve_context(root_names[j]);
          }
          catch(NameNotFound& e)
          {;}
          if(!file_meta_names.is_nil())
          {
            for(DataNamesTable::recursive_iterator ritr=\
              file_meta_names->recursive_begin();\
              ritr!=file_meta_names->recursive_end();ritr++)
            {
              if(ritr.type()==N_OBJECT)
              {
                string meta_data_name;
                if(j==1)
                   meta_data_name+=data_files[i].substr(offset) +"/"+\
                     ritr.name();
                else if(j==0)
                {
                  Name attr_name(ritr.name());
                  NameComponent& last_comp=attr_name[\
                    attr_name.num_components()-1];
                  last_comp=cdf_file.get_reg_attr_name(last_comp.get_id());
                  meta_data_name+=data_root_names[i]+"/"+\
                    attr_name.srep();
                }
                try
                {
                  (void)metadata_table->resolve(meta_data_name);
                } // try
                catch(NameNotFound& not_found)
                {
                  // create a meta data entry
                  CDF_DCMMetaDataEntry *meta_entry=new \
                    CDF_DCMMetaDataEntry(meta_data_name.c_str(),\
                    DCM_CDF_FILE,data_files[i].c_str(),\
                    (root_names[j]+"/"+ritr.name()).c_str());
                  try{
                    metadata_table->bindx(meta_data_name,meta_entry);
                  }
                  catch(NameAlreadyBound& e)
                  {
                    cout << meta_data_name << " already bound" << endl;
                    delete meta_entry;
                  }
                } // catch
              } // if              
             } // for
           } // if ++
         } // for j
      } // if
      // need to define this in CDFDataConfigurationMap.h
      CDF_DCMMetaDataTable_var file_meta_entries;
      try{
        string meta_data_root=data_files[i].substr(offset);
        file_meta_entries=\
          metadata_table->resolve_context(meta_data_root);
        CDF_DCMMetaDataTable::recursive_iterator itr;
        for( itr=file_meta_entries->recursive_begin();\
          itr!=file_meta_entries->recursive_end();itr++)
        {
          if(itr.type()!=N_OBJECT)
             continue;
          string name=meta_data_root+"/"+itr.name();
          try{
            data_names->bindx(string(METADATA)+"/"+name,"");
            meta_names->bindx(name,"");
          }
          catch(NameAlreadyBound& e)
          {;}
         }
      }
      catch(NameNotFound& e)
      {;}
    }
    catch(Exception& e)
    { e.print_msg();throw e;}
  } // for
  // get list of cdf files
  vector<string> database_files_list;
  DataNamesTable_var database_files=get_datafile_names();
  database_files->list_names_r(database_files_list,"/");
  for(int i=0;i<database_files_list.size();i++)
  {
    if(data_name!="")
    {
      if(database_files_list[i]!=data_name)
        continue;
    }
    vector<string>::iterator sitr=\
      find(data_files.begin(),data_files.end(),database_files_list[i]);
    if(sitr!=data_files.end())
      continue;
    int offset=(database_files_list[i].substr(0,1)=="/")?1:0;
    string meta_name_root=database_files_list[i].substr(offset);
    try{
      (void)meta_names->resolve_context(meta_name_root);
      continue;
    }
    catch(NameNotFound& e)
    {;}
    DataNamesTable_var file_data_names;
    try{
      CDFFileDataSource cdf_file(database_files_list[i]);
      file_data_names=cdf_file.get_data_names();
      DataNamesTable_var file_meta_names;
      if(!file_data_names.is_nil())
      {
        try{
          file_meta_names=file_data_names->resolve_context(root_names[1]);
        }
        catch(NameNotFound& e)
        {;}
      }
      if(!file_meta_names.is_nil())
      {
         for(DataNamesTable::recursive_iterator ritr=\
           file_meta_names->recursive_begin();\
           ritr!=file_meta_names->recursive_end();ritr++)
         {
           if(ritr.type()!=N_OBJECT)
             continue;
           Name attr_name(ritr.name());
           NameComponent& last_comp=attr_name[\
             attr_name.num_components()-1];
           last_comp=cdf_file.get_reg_attr_name(last_comp.get_id());
           string meta_data_name=meta_name_root+"/"+\
             attr_name.srep();
           try
           {
             (void)metadata_table->resolve(meta_data_name);
           } // try
           catch(NameNotFound& not_found)
           {
             CDF_DCMMetaDataEntry *meta_entry=new \
               CDF_DCMMetaDataEntry(meta_data_name.c_str(),\
               DCM_CDF_FILE,database_files_list[i].c_str(),\
               (root_names[1]+"/"+ritr.name()).c_str());
             try{
               metadata_table->bindx(meta_data_name,meta_entry);
             }
             catch(NameAlreadyBound& e)
             { delete meta_entry;}
           } // catch
         } // for
      } // if
      // need to define this in CDFDataConfigurationMap.h
      CDF_DCMMetaDataTable_var file_meta_entries;
      try{
        file_meta_entries=\
          metadata_table->resolve_context(meta_name_root);
        CDF_DCMMetaDataTable::recursive_iterator itr;
        for( itr=file_meta_entries->recursive_begin();\
          itr!=file_meta_entries->recursive_end();itr++)
        {
          if(itr.type()!=N_OBJECT)
             continue;
          string name=meta_name_root+"/"+itr.name();
          try{
            data_names->bindx(string(METADATA)+"/"+name,"");
            meta_names->bindx(name,"");
          }
          catch(NameAlreadyBound& e)
          {;}
        }
      }
      catch(NameNotFound& e)
      {;}
    } // try
    catch(Exception& e)
    { e.print_msg();}
  } // for
  for(DataNamesTable::recursive_iterator itr=\
   data_data_names->recursive_begin();\
   itr!=data_data_names->recursive_end();itr++)
  {
    if(itr.type()!=N_OBJECT)
      continue;
    if(data_name!="")
    {
      if(itr.name()!=data_name)
        continue;
    }
    CDF_DCMMetaDataTable_var data_meta_entries;
    try{
      string meta_name_root=itr.name();
      data_meta_entries=\
        metadata_table->resolve_context(meta_name_root);
      CDF_DCMMetaDataTable::recursive_iterator ritr;
      for( ritr=data_meta_entries->recursive_begin();\
        ritr!=data_meta_entries->recursive_end();ritr++)
      {
        if(ritr.type()!=N_OBJECT)
           continue;
        string name=meta_name_root+"/"+ritr.name();
        try{
          data_names->bindx(string(METADATA)+"/"+name,"");
          meta_names->bindx(name,"");
        }
        catch(NameAlreadyBound& e)
        {;}
      }
    }
    catch(NameNotFound& e)
    {;}
  }
  if(data_name!="")
  {
    DataNamesTable_var meta_names_sub;    
    int offset=(data_name.substr(0,1)=="/")?1:0;
    try{
      meta_names_sub=meta_names->resolve_context(data_name.substr(offset));
    }
    catch(NameNotFound& e)
    {;}
    return meta_names_sub;
  }
  else
    return meta_names;
}

// ++++++++++++++++++++++++++++++++++++++ class CDFDataBaseDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------- get_meta_data

DataBuffer_var
  CDFDataBaseDataSource::get_meta_data( const string& data_name )
{
  DataBuffer_var ret_dbuff_var; // return data buffer

  CDF_DCMMetaDataEntry* metadata_entry =
   cdf_data_configmap->get_metadata_entry( data_name );
  
  if( metadata_entry == 0 )
    throw NoSuchDataSourceName(
     "CDFDataBaseDataSource::get_meta_data " + data_name );

  #ifdef DIAG
  cout << "meta entry: " << endl;
  cout << "_data_source_name: " << metadata_entry->_data_source_name << endl;
  cout << "_file_type: " << metadata_entry->_file_type << endl;
  cout << "_file_name: " << metadata_entry->_file_name << endl;
  cout << "_file_data_source_name: " << metadata_entry->_file_data_source_name << endl;
  #endif

  if( metadata_entry->_file_type == "CDF"
      || metadata_entry->_file_type == "cdf" )
  {
     CDFFileDataSource cdf_file( metadata_entry->_file_name );
     ret_dbuff_var = cdf_file.get_data(
                       metadata_entry->_file_data_source_name );
  }
  return ret_dbuff_var;
}
// ++++++++++++++++++++++++++++++++++++++ class CDFDataBaseDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------- get_data_data
//
// get_data for ts data

DataBuffer_var CDFDataBaseDataSource::get_data_data(
     const string& data_name,
     const TimeInterval& time_interval,int entry )
{
// return buffer _var
  TSDataBuffer_var ret_tsdata_buffer_var = new TSDataBuffer; 

  CDF_DCMDataIntervalList data_interval_list;

  cdf_data_configmap->get_data_intervals(
    data_name, time_interval, data_interval_list );

// if no data intervals, then return nil buffer _var
  if( data_interval_list.size() == 0 )
    return ret_tsdata_buffer_var;

// time tags buffer assumed numeric scalar
  NumericVDataBuffer* ttags_nvdb_p = new NumericVDataBuffer;

  vector<DataIntervalInfo> data_intervals_info;

  check_data_type( data_interval_list, data_intervals_info );
  
  bool data_is_numeric=!( data_intervals_info[0].cdf_data_type == CDF_CHAR
      || data_intervals_info[0].cdf_data_type == CDF_UCHAR );
 
 // allocate buffer for numeric return data, and set dimensionality
 //   to that of first data interval

  DataBuffer_var data_vdb_p;
  if(data_is_numeric)
  {
      NumericVDataBuffer_var data_nvdb_p = new NumericVDataBuffer;
      data_nvdb_p->set_dims( data_intervals_info[0].data_dims );
      data_vdb_p=data_nvdb_p;
  }
  else
  {
      StringVDataBuffer_var data_svdb_p=new StringVDataBuffer;
      data_svdb_p->set_dims( data_intervals_info[0].data_dims );
      data_vdb_p=data_svdb_p;
  }
  
  // loop over entries in data interval list
  for( int i=0; i<data_interval_list.size(); i++ )
  {
    // if entry is -1 select all entries, otherwise just get required 
    // entry
    if(entry!=-1 && i!=entry)
      continue;
    {
      CDF_TSFileDataSource cdf_file(
         data_interval_list[i]._dcm_entry._file_name );
      
      cdf_file.set_numeric_ignoration_functor(
        data_interval_list[i]._dcm_entry._variable_name,
        numeric_ignoration_table[ data_name ] );

      cdf_file.set_time_variable(
        data_interval_list[i]._dcm_entry._variable_name,
	data_interval_list[i]._dcm_entry._time_tags_variable_name );

//      #ifdef DIAG
      cout << "about to get data for " 
        << data_interval_list[i]._dcm_entry._variable_name
        << " for interval: " 
        << data_interval_list[i]._time_interval
        << endl;
//      #endif

      DataBuffer_var data_buff = cdf_file.get_data_by_variable(
        data_interval_list[i]._dcm_entry._variable_name, 
	data_interval_list[i]._time_interval );

      TSDataBuffer_var tsdata_buff =
        TSDataBuffer_var::narrow( data_buff );
      if( tsdata_buff.is_nil() )
        throw CDFDataBaseException( "tsdata_buff unexpectedly nil" );

      NumericVDataBuffer_var ttags_buff =
        NumericVDataBuffer_var::narrow( tsdata_buff->ttags );
      if( ttags_buff.is_nil() )
        throw CDFDataBaseException( "Time tags numeric buffer _var unexpectedly nil" );
      
      ttags_nvdb_p->push_back( *ttags_buff );
 
      if(data_is_numeric)
      {
        NumericVDataBuffer_var ndata_buff =
          NumericVDataBuffer_var::narrow( tsdata_buff->data );
        if( ndata_buff.is_nil() )
          throw CDFDataBaseException( \
            "Data numeric buffer _var unexpectedly nil" );
      
        NumericVDataBuffer_var data_nvdb_p=\
          NumericVDataBuffer_var::narrow(data_vdb_p);
        if(data_nvdb_p.is_nil()) 
          throw CDFDataBaseException(\
             "Data numeric receiving buffer _var unexpectedly nil");      
        data_nvdb_p->push_back( *ndata_buff );
      }
      else
      {
        StringVDataBuffer_var sdata_buff =
          StringVDataBuffer_var::narrow( tsdata_buff->data );
        if( sdata_buff.is_nil() )
          throw CDFDataBaseException( \
            "Data string buffer _var unexpectedly nil" );
      
        StringVDataBuffer_var data_svdb_p=\
          StringVDataBuffer_var::narrow(data_vdb_p);
        if(data_svdb_p.is_nil()) 
          throw CDFDataBaseException(\
             "Data string receiving buffer _var unexpectedly nil");      
        data_svdb_p->push_back( *sdata_buff );
      }
      
    }

  } // end of loop over entries in data interval list

  ret_tsdata_buffer_var->data = data_vdb_p;
  ret_tsdata_buffer_var->ttags = ttags_nvdb_p;
  if(ttags_nvdb_p->size()!=0)
  {
    ret_tsdata_buffer_var->time_interval.set(
       Seconds((*ttags_nvdb_p)[0]) ,
      Seconds((*ttags_nvdb_p)[ttags_nvdb_p->size()-1] ) );
  }
  
  return ret_tsdata_buffer_var;

} // end CDFDataBaseDataSource::get_data



// ++++++++++++++++++++++++++++++++++++++ class CDFDataBaseDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------- check_data_type

void CDFDataBaseDataSource::check_data_type( 
      const CDF_DCMDataIntervalList& data_interval_list,
      vector<DataIntervalInfo>& info )
{
  info.resize( data_interval_list.size() );
  
  long first_cdf_data_type;
  vector<int> first_data_dims;
  bool first_type_is_string;
  
  for( int i=0; i<data_interval_list.size(); i++ )
  {
    {
      CDF_TSFileDataSource cdf_file(
         data_interval_list[i]._dcm_entry._file_name );
      
      cdf_file.select();
      
      CDFFileDataSource::CDFVarInfo var_info( 
        data_interval_list[i]._dcm_entry._variable_name );
     
      if( i == 0 )
      {
        first_cdf_data_type = var_info.data_type;
        for( int idim=0; idim<var_info.num_dims; idim++ )
          first_data_dims.push_back( var_info.dim_sizes[idim] );
        info[0].data_type_ok = true;
        info[0].cdf_data_type = first_cdf_data_type;
        info[0].data_dims = first_data_dims;
        first_type_is_string = 
          info[0].cdf_data_type == CDF_CHAR
          || info[0].cdf_data_type == CDF_UCHAR;
       } else
      {
        info[i].cdf_data_type = var_info.data_type;
        for( int idim=0; idim<var_info.num_dims; idim++ )
          info[i].data_dims.push_back( var_info.dim_sizes[idim] );
        bool type_is_string = 
          info[i].cdf_data_type == CDF_CHAR
          || info[i].cdf_data_type == CDF_UCHAR;

// data conformance test: (type same as first entry and equal dimensions
// same type means either both string like, or both numeric like

        info[i].data_type_ok =
             ( (type_is_string && first_type_is_string)
               || (!type_is_string && !first_type_is_string) )
             &&  info[i].data_dims == first_data_dims;       
      }
      #ifdef DIAG
      cout << "For variable: "
       << data_interval_list[i]._dcm_entry._variable_name
       << " got cdf type: " << var_info.var_info_srep() 
       << " data_type_ok: " << info[i].data_type_ok <<  endl;
      #endif
    }  
  } // end loop over data intervals

} // end CDFDataBaseDataSource::check_data_type


} // end namespace QDOS
