//  CDFDataSource.cc
//
//  Qdos2 - data interfacing
//
//  David Burgess
//  July 1999
//

#include <iostream>
#include <stdio.h>   // for sprintf.h

#include "CDFDataSource.h"
#include "DCMSpec.h"
#include "Factory.h"
#include "StringRep.h"

namespace QDOS {

// static variables
DataNamesTable_var initialise_attr_names_table()
{
  DataNamesTable_var names_table=new DataNamesTable;
  names_table->bindx("PTR/LABL_PTR_1","LABLAXIS");
  names_table->bindx("PTR/UNIT_PTR","UNITS");
  names_table->bindx("PTR/FORM_PTR","FORMAT");
  return names_table;
}

DataNamesTable_var CDFFileDataSource::attr_names_table;

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- get_data
DataBuffer_var CDFFileDataSource::get_data( const string& data_name )
{
  DataBuffer_var nil_object; // for nil return

  Name dname( data_name ); // split into components
  if(dname.num_components()==1)
  {
     if(cdf_variable_exists(data_name))
       return get_data_by_variable(data_name);
     else 
       return get_gattr_data(data_name);
  }
  else if( dname.num_components() == 2 )
  {
    if( dname[0] == "variables" )
      return get_data_by_variable( dname[1].get_id() );
    else if( dname[0] == GLOBAL_ATTRIBUTES )
      return get_gattr_data( dname[1].get_id() );
    else 
      return get_vattr_data(
        dname[0].get_id(), dname[1].get_id() );     
  } 
  else if( dname.num_components() == 3 )
  {
    if( dname[0] == VARIABLE_ATTRIBUTES )
      return get_vattr_data( dname[1].get_id(), dname[2].get_id() );
    else if( dname[0] == "cdf_data_type" )
    {
      if( dname[1] == "variables" )
        return get_var_cdf_data_type( dname[2].get_id() );
      else if( dname[1] == GLOBAL_ATTRIBUTES )
        return get_gattr_cdf_data_type( dname[2].get_id() );
    }
    else
      return nil_object; 
  }
  else if( dname.num_components() == 4 )
  {
     if( dname[0] == "cdf_data_type" )
     {
       if( dname[1] == VARIABLE_ATTRIBUTES )
         return get_vattr_cdf_data_type(
           dname[2].get_id(), dname[3].get_id() );     
     }
     else return nil_object;
  }
  else
    return nil_object; 

}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- get_data
DataBuffer_var CDFFileDataSource::get_data(
  const string& data_name, int rec_start, int num_rec )
{
  DataBuffer_var nil_object; // for nil return
  Name dname( data_name ); // split into components
  if(dname.num_components()==1)
  {
     if(cdf_variable_exists(data_name))
       return get_data_by_variable(data_name,rec_start,num_rec);
     else 
       return nil_object;
  }
  else if( dname.num_components() == 2 )
  {
    if( dname[0] == "variables" )
      return get_data_by_variable( dname[1].get_id(), rec_start, num_rec );
    else
      return nil_object;
  }
  else
    return nil_object;
  
// lots more to go here for other data name components


}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------------- open
void CDFFileDataSource::open( const string& cdf_file_name )
{
  CDFstatus status;

  if( open_status )
    close();

  open_status=false;

  file_name = cdf_file_name;

  if( (status = CDFlib(OPEN_, CDF_, file_name.c_str(), &cdf_id, NULL_))<CDF_OK)
   throw CDFFileBadOpen(status,"Failed to open CDF file: "+file_name );
  //if( (status=CDFopen( file_name.c_str(), &cdf_id )) < CDF_OK )
  //  throw CDFFileBadOpen(status,"Failed to open CDF file: "+file_name );

  select();
  open_status = true;

  status = CDFlib( GET_,
                     CDF_VERSION_, &cdf_version,
                     CDF_RELEASE_, &cdf_release,
                     CDF_INCREMENT_, &cdf_release_increment,
                     CDF_MAJORITY_, &cdf_majority,
                   NULL_ );
  if( status != CDF_OK )
      throw CDFFileBadRead(status,"Failed to GET_ info on CDF file" );

}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------- select_cdf_id
void CDFFileDataSource::select( void ) const
{
  CDFstatus status;
  if( (status=CDFlib( SELECT_, CDF_, cdf_id, CDF_zMODE_, zMODEon2, NULL_ ))
       != CDF_OK )
    throw CDFFileBadOpen(status,"Failed to select CDF file: " + file_name );
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------------ close
void CDFFileDataSource::close()
{
  if( open_status )
  {
    CDFclose( cdf_id );
    open_status = false;
  }
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------- get_data_names

DataNamesTable_var CDFFileDataSource::get_data_names(void)
{
// if the table has been already read in, then return it now
  if( !data_names_table.is_nil() )
    return data_names_table;

  create_data_names_table();

  return data_names_table;
}


// +++++++++++++++++++++++++++++ class CDFFileDataSource::CDFTypeInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------- num_bytes_for_cdf_data_type
int CDFFileDataSource::num_bytes_for_cdf_data_type( long data_type )
{
  int num_bytes=0;
  switch( data_type )
  {
  case CDF_CHAR: case CDF_UCHAR:
    num_bytes = 1;
    break;
  case CDF_BYTE: case CDF_INT1: case CDF_UINT1:
    num_bytes = 1;
    break;
  case CDF_INT2: case CDF_UINT2:
    num_bytes = 2;
    break;
  case CDF_INT4: case CDF_UINT4: case CDF_REAL4: case CDF_FLOAT:
    num_bytes = 4;
    break;
  case CDF_REAL8: case CDF_DOUBLE: case CDF_EPOCH:
    num_bytes = 8;
    break;
  }
  return num_bytes;
}
// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------- initialise_attr_names_table
string CDFFileDataSource::get_ptr_attr_name(const string& attr_name) const
{
  if(attr_names_table.is_nil())
    attr_names_table=initialise_attr_names_table();
  string ptr_attr_name(attr_name);
  try
  {
    ptr_attr_name=attr_names_table->resolve(attr_name);
  }
  catch(Exception& e)
  {
    ptr_attr_name+="_ptr";
  }
  select();
  CDFstatus s = CDFlib( CONFIRM_, ATTR_EXISTENCE_, ptr_attr_name.c_str(), NULL_ );
  if( s == CDF_OK )
    return ptr_attr_name;
  else
    return attr_name;
}

string CDFFileDataSource::get_reg_attr_name(const string& ptr_attr_name) const
{
  if(attr_names_table.is_nil())
    attr_names_table=initialise_attr_names_table();
  string attr_name(ptr_attr_name);
  try
  {
    attr_name=attr_names_table->resolve("PTR/"+ptr_attr_name);
  }
  catch(Exception& e)
  {
    int pos=ptr_attr_name.find("_PTR");
    if(pos==string::npos)
      pos=ptr_attr_name.find("_ptr");
    if(pos!=string::npos)
      attr_name=ptr_attr_name.substr(0,pos);
  }
  return attr_name;
}
// +++++++++++++++++++++++++++++ class CDFFileDataSource::CDFTypeInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------- calc_bytes_elt
void CDFFileDataSource::CDFTypeInfo::calc_bytes_elt( void )
{
  num_bytes_elt=0;
  switch( data_type )
  {
  case CDF_CHAR: case CDF_UCHAR:
    num_bytes_elt = num_elems;
    break;
  case CDF_BYTE: case CDF_INT1: case CDF_UINT1:
    num_bytes_elt = 1;
    break;
  case CDF_INT2: case CDF_UINT2:
    num_bytes_elt = 2;
    break;
  case CDF_INT4: case CDF_UINT4: case CDF_REAL4: case CDF_FLOAT:
    num_bytes_elt = 4;
    break;
  case CDF_REAL8: case CDF_DOUBLE: case CDF_EPOCH:
    num_bytes_elt = 8;
    break;
  }
}
// +++++++++++++++++++++++++++++ class CDFFileDataSource::CDFTypeInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------- data_type_srep
string CDFFileDataSource::CDFTypeInfo::data_type_srep(void)
{
  string s;
  switch( data_type )
  {
  case CDF_CHAR: s="CDF_CHAR"; break;
  case CDF_UCHAR: s="CDF_UCHAR"; break;
  case CDF_BYTE: s="CDF_BYTE"; break;
  case CDF_INT1: s="CDF_INT1"; break;
  case CDF_UINT1: s="CDF_UINT1"; break;
  case CDF_INT2: s="CDF_INT2"; break;
  case CDF_UINT2: s="CDF_UINT2"; break;
  case CDF_INT4: s="CDF_INT4"; break;
  case CDF_UINT4: s="CDF_UINT4"; break;
  case CDF_REAL4: s="CDF_REAL4"; break;
  case CDF_FLOAT: s="CDF_FLOAT"; break;
  case CDF_REAL8: s="CDF_REAL8"; break;
  case CDF_DOUBLE: s="CDF_DOUBLE"; break;
  case CDF_EPOCH: s="CDF_EPOCH"; break;
  }
  return s;
}
// ++++++++++++++++++++++++++++++ class CDFFileDataSource::CDFVarInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- CDFVarInfo
CDFFileDataSource::CDFVarInfo::CDFVarInfo( const string& var_name )
{
  CDFstatus s;

  s = CDFlib( SELECT_, zVAR_NAME_, var_name.c_str(), NULL_ );
  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to select CDF variable: " + var_name );

  name = var_name;

  s = CDFlib( GET_,
              zVAR_DATATYPE_, &data_type,
              zVAR_NUMELEMS_, &num_elems,
              zVAR_NUMDIMS_,  &num_dims,
              zVAR_RECVARY_,  &rec_vary,
              zVAR_MAXREC_,   &max_rec_num,
              zVAR_DIMSIZES_, dim_sizes,
              zVAR_DIMVARYS_, dim_varys,      NULL_ );

  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to get info on CDF variable: " + var_name );

// note: num_elts is number of elements (ie numbers) in a possibly
// multidimensional matrix variable
// num_elems is the cdf variable information, that is 1 for all types
// except CDF_CHAR and CDF_UCHAR, when it is number of characters

  calc_bytes_elt();

  num_elts=1;
  for(int i=0;i<num_dims;i++)
    num_elts*=dim_sizes[i];
  num_bytes_var = num_elts * num_bytes_elt;

//  cout << " CDFVarInfo: numelems=" << num_elems << " num_bytes_elt="
//       << num_bytes_elt << " num_elts=" << num_elts << endl;
//  cout << "           : num_bytes_var="<<num_bytes_var<<endl;
}

// ++++++++++++++++++++++++++++++ class CDFFileDataSource::CDFVarInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------------- var_info_srep
string CDFFileDataSource::CDFVarInfo::var_info_srep(void)
{
  static char s[200];

  sprintf(s,"%s/%i %i:[" , data_type_srep().c_str(), num_elems, num_dims );
  for(int i=0;i<(num_dims-1);i++)
    sprintf(s+strlen(s),"%i,", dim_sizes[i] );
  if(num_dims!=0)
    sprintf(s+strlen(s),"%i", dim_sizes[num_dims-1]);
  sprintf(s+strlen(s),"] ");

  if( rec_vary== VARY ) sprintf(s+strlen(s),"T/");
  else if( rec_vary== NOVARY ) sprintf(s+strlen(s),"F/");
  else sprintf(s+strlen(s),"?/");

  for(int i=0;i<num_dims;i++)
    if( dim_varys[i]== VARY ) sprintf(s+strlen(s),"T");
    else if( dim_varys[i]== NOVARY ) sprintf(s+strlen(s),"F");
    else sprintf(s+strlen(s),"?");

  return s;
}


// ++++++++++++++++++++++++++++ class CDFFileDataSource::CDFVarBuffer
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------- CDFVarBuffer
CDFFileDataSource::CDFVarBuffer::CDFVarBuffer( const CDFVarInfo& v, int size )
{
  p = malloc( v.num_bytes_var*size );
//  if( p == 0 )   throw NoMemory();
}

// ++++++++++++++++++++++++++++++++ class CDFDataSource::CDFGAttrBuffer
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------- CDFGAttrBuffer
CDFFileDataSource::CDFGAttrBuffer::CDFGAttrBuffer( const CDFGAttrInfo& g )
{
  p = malloc( g.num_elems * g.num_bytes_elt );
//  if( p == 0 )  throw NoMemory();
}
// ++++++++++++++++++++++++++++++++ class CDFDataSource::CDFGAttrBuffer
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------- CDFGAttrBuffer
CDFFileDataSource::CDFGAttrBuffer::CDFGAttrBuffer(
   long cdf_data_type, long num_elements  )
{
  p = malloc( num_elements * num_bytes_for_cdf_data_type(cdf_data_type) );
//  if( p == 0 )  throw NoMemory();
}

// ++++++++++++++++++++++++++++++++ class CDFDataSource::CDFVAttrBuffer
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------- CDFVAttrBuffer
CDFFileDataSource::CDFVAttrBuffer::CDFVAttrBuffer( const CDFVAttrInfo& a )
{
  p = malloc( a.num_elems * a.num_bytes_elt );
//  if( p == 0 )   throw NoMemory();
}

// ++++++++++++++++++++++++++++++ class CDFFileDataSource::CDFGAttrInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- CDFGAttrInfo
CDFFileDataSource::CDFGAttrInfo::CDFGAttrInfo(
  const CDFFileDataSource& cdf_file_ds, const string& gattr_name )
  : cdf( cdf_file_ds )
{
  CDFstatus s;

  cdf.select();

  _gattr_name = gattr_name;

  if( ! cdf.cdf_gattribute_exists( _gattr_name ) )
    throw CDFFileNoSuchAttribute(
      0, "No such global attribute: " + _gattr_name );


  s = CDFlib( SELECT_, ATTR_NAME_, _gattr_name.c_str(), NULL_ );
  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to select Global attribute: " + _gattr_name );

  s = CDFlib( GET_,
              ATTR_MAXgENTRY_,   &max_entry_num,
              ATTR_NUMgENTRIES_, &num_entries,
              ATTR_NUMBER_,      _gattr_name.c_str(), &attr_num,      NULL_ );

  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to get info on Global attribute: " + _gattr_name );

// find first existing entry, and from that entry set the
//  data type and num_elems (which will be assumed to hold for
//  all entries
  for ( int i_entry=0; i_entry < max_entry_num+1; i_entry++)
  {
    s = CDFlib( CONFIRM_, gENTRY_EXISTENCE_, i_entry, NULL_ );
    if( s == CDF_OK )
    {
      s = CDFlib(SELECT_, gENTRY_, i_entry, NULL_);
      if( s != CDF_OK )
         throw CDFFileBadRead(s,
           "Failed to select entry for global attribute: " + _gattr_name );

      s = CDFlib( GET_,
              gENTRY_DATATYPE_, &data_type,
              gENTRY_NUMELEMS_, &num_elems,      NULL_ );

      if( s != CDF_OK )
        throw CDFFileBadRead(s,"Failed to get info on G attr: " + _gattr_name );
// quit loop
      break;
    }
  }

// use data_type and num_elems to calc number of bytes per element
  calc_bytes_elt();
}
// ++++++++++++++++++++++++++++ class CDFFileDataSource::CDFGAttrInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------- gattr_info_srep
string CDFFileDataSource::CDFGAttrInfo::gattr_info_srep(void)
{
  static char s[200];

  sprintf(s,"GATTR: %s %i %s/%i" ,
    _gattr_name.c_str(),
    max_entry_num+1,
    data_type_srep().c_str(),
    num_elems );

  return s;
}

// ++++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------- get_gattr_cdf_data_type
DataBuffer_var CDFFileDataSource::get_gattr_cdf_data_type(
                 const string& gattr_name )
{
  DataBuffer_var data_buffer;   // return object var
  //select();
  //if( ! cdf_gattribute_exists( gattr_name ) )
  //  throw CDFFileNoSuchAttribute(
  //    0, "No such global attribute: " + gattr_name );
  CDFGAttrInfo gattr_info( *this, gattr_name );
  StringVDataBuffer* sdb_p = new StringVDataBuffer();
  data_buffer = sdb_p;
  sdb_p->push_back( gattr_info.gattr_info_srep() );

  return data_buffer;
}

// ++++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------- get_gattr_data
DataBuffer_var CDFFileDataSource::get_gattr_data( const string& gattr_name )
{
  CDFstatus s;
  DataBuffer_var data_buffer;   // return object var

  //select();

  //if( ! cdf_gattribute_exists( gattr_name ) )
  //  throw CDFFileNoSuchAttribute(
  //    0, "No such global attribute: " + gattr_name );

  CDFGAttrInfo gattr_info( *this, gattr_name );

// If number of existing entries is less than 1 (ie zero!)
// then return null string
// this occurs when a global attribute is created but never
//  written to

  if( gattr_info.num_entries < 1 )
  {
    StringVDataBuffer* sdb_p = new StringVDataBuffer();
    sdb_p->push_back( "" );
    data_buffer = sdb_p;
    return data_buffer;
  }

// if the global attribute is string ...
  if( gattr_info.data_type==CDF_CHAR || gattr_info.data_type==CDF_UCHAR )
  {
    string xstr;
    StringVDataBuffer* sdb_p = new StringVDataBuffer();
    data_buffer = sdb_p;

    for( long i_entry=0; i_entry <= gattr_info.max_entry_num; i_entry++ )
    {
      //  missing entries represented by ""
      if( !cdf_gattr_entry_exists( gattr_info.attr_num, i_entry ) )
      {
        sdb_p->push_back( "" );
        continue;
      }

// get data type of this entry
      long entry_data_type, entry_num_elems;
      get_gattr_data_type( gattr_info.attr_num, i_entry,
                           entry_data_type, entry_num_elems);

// mismatched data types
      if ( entry_data_type != gattr_info.data_type )
      {
        sdb_p->push_back( "" );
        continue;
      }

      CDFGAttrBuffer read_buffer( entry_data_type, entry_num_elems );

      s = CDFlib( SELECT_, ATTR_, gattr_info.attr_num,
                  SELECT_, gENTRY_, i_entry,
                  GET_, gENTRY_DATA_, read_buffer.p,
                  NULL_ );
      if( s != CDF_OK )
        throw CDFFileBadRead(s,"Failed to read entry for G attribute: " + gattr_info._gattr_name );

      sdb_p->push_back( string((const char*)read_buffer.p, entry_num_elems ) );

    } // end of loop over entries

  } else // other than strings
  {
    NumericVDataBuffer* ndb_p = new NumericVDataBuffer();
    data_buffer = ndb_p;
    if( gattr_info.num_elems > 1)
    {
      vector<int> dims;
      dims.push_back( gattr_info.num_elems );
      ndb_p->set_dims( dims );
    }

// byte buffer to be used by CDFlib, for non-char type this is fixed size
    CDFGAttrBuffer read_buffer( gattr_info );

    for( long i_entry=0; i_entry <= gattr_info.max_entry_num; i_entry++ )
    {
     //  missing entries are zero filled
      if( ! cdf_gattr_entry_exists( gattr_info.attr_num, i_entry ) )
      {
       for(int i=0;i<gattr_info.num_elems;i++) ndb_p->push_back(0.0);
       continue;
      }
// get data type of this entry
      long entry_data_type, entry_num_elems;
      get_gattr_data_type( gattr_info.attr_num, i_entry,
                           entry_data_type, entry_num_elems);

// mismatched data types
//  IF .. data type of entry is different from first existing
//    OR .. number of elements of entry is different

// NOTE: this could be improved: ie use MAX number of elements
//       and pad up to this with zero if entry has fewer elements

      if ( entry_data_type != gattr_info.data_type
           || entry_num_elems != gattr_info.num_elems )
      {
        for(int i=0;i<gattr_info.num_elems;i++) ndb_p->push_back(0.0);
        continue;
      }

      s = CDFlib( SELECT_, ATTR_, gattr_info.attr_num,
                  SELECT_, gENTRY_, i_entry,
                  GET_, gENTRY_DATA_, read_buffer.p,
                  NULL_ );
      if( s != CDF_OK )
        throw CDFFileBadRead(s,
          "Failed to read entry for G attribute: " + gattr_info._gattr_name );

// for numeric types use element by element static casting
//   and let ndb_p->push_back(double) do the conversion to double

      switch( gattr_info.data_type )
      {
        case CDF_BYTE:
        case CDF_INT1:
          for(int i=0;i<gattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<char*>(read_buffer.p) + i ) );
          break;
        case CDF_UINT1:
          for(int i=0;i<gattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<unsigned char*>(read_buffer.p) + i ) );
          break;
        case CDF_INT2:
          for(int i=0;i<gattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<short*>(read_buffer.p) + i ) );
          break;
        case CDF_UINT2:
          for(int i=0;i<gattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<unsigned short*>(read_buffer.p) + i ) );
          break;
        case CDF_INT4:
          for(int i=0;i<gattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<int*>(read_buffer.p) + i ) );
          break;
        case CDF_UINT4:
          for(int i=0;i<gattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<unsigned int*>(read_buffer.p) + i ) );
          break;
        case CDF_REAL4:
        case CDF_FLOAT:
          for(int i=0;i<gattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<float*>(read_buffer.p) + i ) );
          break;
        case CDF_REAL8:
        case CDF_DOUBLE:
        case CDF_EPOCH:
          for(int i=0;i<gattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<double*>(read_buffer.p) + i ) );
          break;
      }
    }  // loop over entries
  } // end if on data type

  return data_buffer;
}

// ++++++++++++++++++++++++++++++ class CDFFileDataSource::CDFVAttrInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- CDFVAttrInfo
CDFFileDataSource::CDFVAttrInfo::CDFVAttrInfo(
  const CDFFileDataSource& cdf_file_ds,
  const string& var_name, const string& vattr_name )
    : cdf( cdf_file_ds )
{
  CDFstatus s;

  cdf.select();

  _vattr_name=vattr_name;
  _var_name=var_name;

  if( ! cdf.cdf_vattribute_exists( _var_name, _vattr_name ) )
    throw CDFFileNoSuchAttribute(
      0, "CDFVAttrInfo: No such variable attribute: " + _var_name + " - " + _vattr_name );

  attr_num = cdf.get_cdf_attribute_number( _vattr_name );
  var_num = cdf.get_cdf_variable_number( _var_name );

  s = CDFlib( SELECT_, ATTR_NAME_, _vattr_name.c_str(),
                       zENTRY_NAME_, _var_name.c_str(), NULL_ );
  if( s < CDF_OK )
    throw CDFFileBadRead(s,"Failed to SELECT_, ATTR_NAME_and/or zENTRY_NAME_ "
    + _var_name +" " + _vattr_name );


  s = CDFlib( GET_,
              zENTRY_DATATYPE_,  &data_type,
              zENTRY_NUMELEMS_,  &num_elems,      NULL_ );

  if( s < CDF_OK )
      throw CDFFileBadRead(s,"Failed to get info on Variable attr: " +
         _var_name + " - " + _vattr_name );

// use data_type and num_elems ...
  calc_bytes_elt();

}
// ++++++++++++++++++++++++++++ class CDFFileDataSource::CDFVAttrInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------- vattr_info_srep
string CDFFileDataSource::CDFVAttrInfo::vattr_info_srep(void)
{
  static char s[200];

  sprintf(s,"VATTR: %s (%s) %s/%i" ,
   _vattr_name.c_str(),
   _var_name.c_str(),
   data_type_srep().c_str(),
   num_elems );

  return s;
}


// ++++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------- get_vattr_cdf_data_type
DataBuffer_var CDFFileDataSource::get_vattr_cdf_data_type(
                 const string& var_name, const string& vattr_name )
{
  DataBuffer_var data_buffer;   // return object var
  //select();
  //if( !cdf_vattribute_exists( var_name, vattr_name ) )
  //  throw CDFFileNoSuchAttribute(
  //    0, "get_vattr_data: No such variable attribute: "
  //     + var_name + " - " + vattr_name );
  CDFVAttrInfo vattr_info( *this, var_name, vattr_name );
  StringVDataBuffer* sdb_p = new StringVDataBuffer();
  data_buffer = sdb_p;
  sdb_p->push_back( vattr_info.vattr_info_srep() );

  return data_buffer;
}


// ++++++++++++++++++++++++++++++++ class CDFFileDataSource::CDFVAttrInfo
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- get_vattr_data
DataBuffer_var CDFFileDataSource::get_vattr_data(
  const string& var_name, const string& vattr_name )
{
  CDFstatus s;
  DataBuffer_var data_buffer;  // return object var

  //select();

  //if( !cdf_vattribute_exists( var_name, vattr_name ) )
  //  throw CDFFileNoSuchAttribute(
  //    0, "get_vattr_data: No such variable attribute: "
  //     + var_name + " - " + vattr_name );

// read  v attr info and select named attr in current CDF file.
  CDFVAttrInfo vattr_info( *this, var_name, vattr_name );

  if( vattr_info.data_type==CDF_CHAR || vattr_info.data_type==CDF_UCHAR )
  {

// byte buffer to be used by CDFlib
    CDFVAttrBuffer read_buffer( vattr_info );

    s = CDFlib( SELECT_, ATTR_NAME_, vattr_info._vattr_name.c_str(), NULL_ );
    if( s < CDF_OK )
      throw CDFFileBadRead(s,"Failed to SELECT_, ATTR_NAME_ "
                          + var_name +" - " + vattr_info._vattr_name );
    s = CDFlib( GET_, zENTRY_DATA_, read_buffer.p,
                NULL_ );
    if( s != CDF_OK )
      throw CDFFileBadRead(s,"Failed to read data for variable attribute: "
       + vattr_info._vattr_name );

    string sdata((const char*)read_buffer.p, vattr_info.num_elems);
    StringVDataBuffer* sdb_p = new StringVDataBuffer();
    sdb_p->push_back(sdata);
    // now do special processing here if vattr_name is a pointer type
    if(cdf_variable_exists(sdata))
    {
      DataPtrBuffer_var data_ptr_var=new DataPtrBuffer();
      data_ptr_var->data_ptr_name=sdb_p;
      data_ptr_var->data=get_data_by_variable(sdata);
      data_buffer=data_ptr_var;
    }
    else
      data_buffer = sdb_p;

  } else // other than strings
  {
    NumericVDataBuffer* ndb_p = new NumericVDataBuffer();
    data_buffer = ndb_p;

    if( vattr_info.num_elems > 1 )
    {
      vector<int> dims;
      dims.push_back( vattr_info.num_elems );
      ndb_p->set_dims( dims );
    }

    CDFVAttrBuffer read_buffer( vattr_info );

    s = CDFlib( SELECT_, ATTR_NAME_, vattr_info._vattr_name.c_str(), NULL_ );
    if( s < CDF_OK )
      throw CDFFileBadRead(s,"Failed to SELECT_, ATTR_NAME_ "
                          + var_name +" " + vattr_info._vattr_name );
    s = CDFlib( GET_, zENTRY_DATA_, read_buffer.p,
                  NULL_ );
    if( s != CDF_OK )
      throw CDFFileBadRead(s,"Failed to read data for variable attribute: "
                          + var_name +" " + vattr_info._vattr_name );

// for numeric types use element by element static casting
//   and let ndb_p->push_back(double) do the conversion to double

      switch( vattr_info.data_type )
      {
        case CDF_BYTE:
        case CDF_INT1:
          for(int i=0;i<vattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<char*>(read_buffer.p) + i ) );
          break;
        case CDF_UINT1:
          for(int i=0;i<vattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<unsigned char*>(read_buffer.p) + i ) );
          break;
        case CDF_INT2:
          for(int i=0;i<vattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<short*>(read_buffer.p) + i ) );
          break;
        case CDF_UINT2:
          for(int i=0;i<vattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<unsigned short*>(read_buffer.p) + i ) );
          break;
        case CDF_INT4:
          for(int i=0;i<vattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<int*>(read_buffer.p) + i ) );
          break;
        case CDF_UINT4:
          for(int i=0;i<vattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<unsigned int*>(read_buffer.p) + i ) );
          break;
        case CDF_REAL4:
        case CDF_FLOAT:
          for(int i=0;i<vattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<float*>(read_buffer.p) + i ) );
          break;
        case CDF_REAL8:
        case CDF_DOUBLE:
        case CDF_EPOCH:
          for(int i=0;i<vattr_info.num_elems;i++)
            ndb_p->push_back( *( static_cast<double*>(read_buffer.p) + i ) );
          break;
      }
  } // end if on data type

  return data_buffer;

}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------- get_var_cdf_data_type
DataBuffer_var
 CDFFileDataSource::get_var_cdf_data_type( const string& var_name )
{
  CDFstatus s;

  select();

  s = CDFlib( SELECT_, zVAR_NAME_, var_name.c_str(), NULL_ );
  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to select CDF variable: " + var_name );

// read var info for currently selected zVar and CDF file.
  CDFVarInfo v( var_name );

  StringVDataBuffer* data_buffer = new StringVDataBuffer();

// no need to set dims, since created as a zero dim buffer

  data_buffer->push_back( v.data_type_srep() );
  data_buffer->push_back( v.var_info_srep() );

  return data_buffer;
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------------------- get_var_data
DataBuffer_var
  CDFFileDataSource::get_data_by_variable(
    const string& var_name, int rec_start, int num_rec )
{
  CDFstatus s;

  select();

  if( ! cdf_variable_exists( var_name ) )
    throw CDFFileNoSuchVariable(
      0, "No such variable " + var_name  );


// read var info for currently selected zVar and CDF file.
  CDFVarInfo v( var_name );

// a negative rec_start is silently zeroed
  if( rec_start < 0 )
    rec_start == 0;

// a num_rec of -1 means up to max_rec_num
  if( num_rec == -1 )
    num_rec = v.max_rec_num - rec_start + 1;

// deal with case where no data can be returned because of
// of requested record range
// viz: return empty data buffer of correct type
//
// the following conditions also satisfy max_rec_num == -1
//    which happens when nothing has been written to variable

  if( rec_start > v.max_rec_num || num_rec <= 0 )
  {
    DataBuffer_var data_buffer;

    if( v.data_type==CDF_CHAR || v.data_type==CDF_UCHAR )
    {
      StringVDataBuffer* sdb_p = new StringVDataBuffer();
      vector<int> dims;
      for(int i=0;i<v.num_dims;i++)
        dims.push_back( v.dim_sizes[i] );
      sdb_p->set_dims( dims );

      if( cdf_majority == ROW_MAJOR ) sdb_p->set_majority( DBUFFER_ROW_MAJOR );
      else if( cdf_majority == COL_MAJOR ) sdb_p->set_majority( DBUFFER_COL_MAJOR );
      else sdb_p->set_majority( DBUFFER_NIL_MAJOR );

      data_buffer = sdb_p; // this is where _var takes over mem management
    } else
    {
      NumericVDataBuffer* ndb_p = new NumericVDataBuffer();
      vector<int> dims;
      for(int i=0;i<v.num_dims;i++)
        dims.push_back( v.dim_sizes[i] );
      ndb_p->set_dims( dims );

      if( cdf_majority == ROW_MAJOR ) ndb_p->set_majority( DBUFFER_ROW_MAJOR );
      else if( cdf_majority == COL_MAJOR ) ndb_p->set_majority( DBUFFER_COL_MAJOR );
      else ndb_p->set_majority( DBUFFER_NIL_MAJOR );

      data_buffer = ndb_p;
    }
    return data_buffer;
  }

// No need to:
// SELECT_, zVAR_DIMINDICES, since set to (0,0,...) on open of file
// SELECT_, zVAR_DIMCOUNTS,  since set to (dim_sizes) on open of file
// SELECT_, zVAR_DIMINTERVALS, since set to (1,1,..) on open of file
// SELECT_, zVAR_RECCOUNT_,  since set to 1 on open of file
// SELECT_, zVAR_RECINTERVAL,  since set to 1 on open of file

// return object var
  DataBuffer_var data_buffer;

  if( v.data_type==CDF_CHAR || v.data_type==CDF_UCHAR )
  {

// CDF strings from file are not null terminated, so we have
// to make a string object, and use assign, before we push_back
// the string on to the StringVDataBuffer.
// xstr is the temporary string that we use for this.
    string xstr;

    StringVDataBuffer* sdb_p = new StringVDataBuffer();
    data_buffer = sdb_p;
// set dims of data buffer
    vector<int> dims;
    for(int i=0;i<v.num_dims;i++)
      dims.push_back( v.dim_sizes[i] );
    sdb_p->set_dims( dims );
    if( cdf_majority == ROW_MAJOR ) sdb_p->set_majority( DBUFFER_ROW_MAJOR );
    else if( cdf_majority == COL_MAJOR ) sdb_p->set_majority( DBUFFER_COL_MAJOR );
    else sdb_p->set_majority( DBUFFER_NIL_MAJOR );

// NOTE: for string data we DO NOT read more than one
// record in at a time (just to keep program simple!)

// byte buffer to be used by CDFlib
    CDFVarBuffer read_buffer( v );

    for( int r=0; r<num_rec; r++ )
    {
      s = CDFlib( SELECT_, zVAR_RECNUMBER_, rec_start+r,
                GET_, zVAR_HYPERDATA_, read_buffer.p,
                NULL_ );
      if( s != CDF_OK )
        throw CDFFileBadRead(s,"Failed to read data from CDF variable: " + var_name );

      for(int i=0;i<v.num_elts;i++)
      {
        xstr.assign( (const char*)read_buffer.p + i * v.num_bytes_elt,
                     v.num_bytes_elt );
        sdb_p->push_back(xstr);
      }
    }
  } else
  {
    NumericVDataBuffer* ndb_p = new NumericVDataBuffer();
    data_buffer = ndb_p;
    vector<int> dims;
    for(int i=0;i<v.num_dims;i++)
      dims.push_back( v.dim_sizes[i] );
    ndb_p->set_dims( dims );
    if( cdf_majority == ROW_MAJOR ) ndb_p->set_majority( DBUFFER_ROW_MAJOR );
    else if( cdf_majority == COL_MAJOR ) ndb_p->set_majority( DBUFFER_COL_MAJOR );
    else ndb_p->set_majority( DBUFFER_NIL_MAJOR );

// actually, reserving space for the vector in the data buffer
//   didn't improve speed
//    ndb_p->reserve( v.num_elts*num_rec );

// rec_count is the number of records to be read in one call to
//   GET_, zVAR_HYPERDATA_

    long rec_count = 1000;

// byte buffer to be used by CDFlib
    CDFVarBuffer read_buffer( v, rec_count );

    s = CDFlib( SELECT_, zVAR_RECCOUNT_, rec_count, NULL_ );
    if( s != CDF_OK )
      throw CDFFileBadRead(s,"Failed select zVAR_RECCOUNT_ for CDF variable: " + var_name );

//cout << "cdf select rec_count: " << rec_count << endl;

    for( int r=0; r<num_rec; r+=rec_count )
    {

// deal with end of sequence where number of remaining records
// is less than initial rec_count
      if( (r+rec_count)>=num_rec+1 )
      {
        rec_count = num_rec - r;
//cout << "new rec_count value: " << rec_count << endl;
        s = CDFlib( SELECT_, zVAR_RECCOUNT_, rec_count, NULL_ );
        if( s != CDF_OK )
          throw CDFFileBadRead(s,"Failed select zVAR_RECCOUNT_ for CDF variable: " + var_name );
      }

      s = CDFlib( SELECT_, zVAR_RECNUMBER_, rec_start+r,
                GET_, zVAR_HYPERDATA_, read_buffer.p,
                NULL_ );
      if( s != CDF_OK )
        throw CDFFileBadRead(s,"Failed to read data from CDF variable: " + var_name );

// for numeric types use element by element static casting
//   and let ndb_p->push_back(double) do the conversion to double

      switch( v.data_type )
      {
        case CDF_BYTE:
        case CDF_INT1:
          for(int i=0;i<v.num_elts*rec_count;i++)
            ndb_p->push_back( *( (char*)read_buffer.p + i ) );
          break;
        case CDF_UINT1:
          for(int i=0;i<v.num_elts*rec_count;i++)
            ndb_p->push_back( *( (unsigned char*)read_buffer.p + i ) );
          break;
        case CDF_INT2:
          for(int i=0;i<v.num_elts*rec_count;i++)
            ndb_p->push_back( *( (short*)read_buffer.p + i ) );
          break;
        case CDF_UINT2:
          for(int i=0;i<v.num_elts*rec_count;i++)
            ndb_p->push_back( *( (unsigned short*)read_buffer.p + i ) );
          break;
        case CDF_INT4:
          for(int i=0;i<v.num_elts*rec_count;i++)
            ndb_p->push_back( *( (int*)read_buffer.p + i ) );
          break;
        case CDF_UINT4:
          for(int i=0;i<v.num_elts*rec_count;i++)
            ndb_p->push_back( *( (unsigned int*)read_buffer.p + i ) );
          break;
        case CDF_REAL4:
        case CDF_FLOAT:
          for(int i=0;i<v.num_elts*rec_count;i++)
            ndb_p->push_back( *( (float*)read_buffer.p + i ) );
          break;
        case CDF_REAL8:
        case CDF_DOUBLE:
        case CDF_EPOCH:
          for(int i=0;i<v.num_elts*rec_count;i++)
            ndb_p->push_back( *( (double*)read_buffer.p + i ) );
          break;
      }
    }  // loop over records
  }
  return data_buffer;
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------- cdf_variable_exists
bool CDFFileDataSource::cdf_variable_exists( const string& var_name ) const
{
  select();
  return CDFlib( CONFIRM_,
                   zVAR_EXISTENCE_, var_name.c_str(), NULL_ ) == CDF_OK ;
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------- cdf_vattribute_exists
bool CDFFileDataSource::cdf_vattribute_exists(
       const string& var_name, const string& attr_name ) const
{
  select();
  //string cdf_vattr_name(attr_name);

  CDFstatus s = CDFlib( CONFIRM_, ATTR_EXISTENCE_, attr_name.c_str(), NULL_ );
  if( s != CDF_OK )
    return false;
  if( get_cdf_attribute_scope(attr_name) != VARIABLE_SCOPE
      || !cdf_variable_exists(var_name) )
    return false;
  s=CDFlib( SELECT_,
      ATTR_NAME_, attr_name.c_str(),
      CONFIRM_,
      zENTRY_EXISTENCE_, get_cdf_variable_number(var_name),
       NULL_ );
  if( s != CDF_OK )
    return false;
  return (s == CDF_OK) ;
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------- cdf_vattribute_exists
bool CDFFileDataSource::cdf_vattribute_exists( int i_var, int i_attr ) const
{
  select();
  CDFstatus s = CDFlib( SELECT_, ATTR_, i_attr, NULL_ );
  if( s != CDF_OK )
    return false;
  if( get_cdf_attribute_scope( i_attr ) != VARIABLE_SCOPE )
    return false;
  return CDFlib( CONFIRM_,
                   zENTRY_EXISTENCE_, i_var,
                 NULL_ ) == CDF_OK ;
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------- cdf_gattribute_exists
bool CDFFileDataSource::cdf_gattribute_exists( const string& attr_name ) const
{
  select();
  //string cdf_gattr_name(attr_name);
  CDFstatus s = CDFlib( CONFIRM_, ATTR_EXISTENCE_, attr_name.c_str(), NULL_ );
  if( s != CDF_OK )
    return false;
  return get_cdf_attribute_scope(attr_name) == GLOBAL_SCOPE ;
}
// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------- cdf_gattr_entry_exists
bool CDFFileDataSource::cdf_gattr_entry_exists( int i_attr, int i_entry ) const
{
  select();
  if( get_cdf_attribute_scope(i_attr) != GLOBAL_SCOPE )
    return false;
  CDFstatus s = CDFlib( SELECT_, ATTR_, i_attr, NULL_ );
  if( s != CDF_OK )
      throw CDFFileBadRead(s,"Failed to SELECT_ ATTR_ " );

  return CDFlib( CONFIRM_, gENTRY_EXISTENCE_, i_entry, NULL_ ) == CDF_OK;
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------- get_num_cdf_variables
int CDFFileDataSource::get_num_cdf_variables(void) const
{
// get number of vars
  select();
  long num_vars;
  CDFstatus s = CDFlib( GET_, CDF_NUMzVARS_, &num_vars, NULL_ );
  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to GET_ CDF_NUMzVARS_" );
  return num_vars;
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------- get_cdf_variable_name
string CDFFileDataSource::get_cdf_variable_name( int i_var ) const
{
  select();
// get the name of corresponding variable
  char var_name[ CDF_VAR_NAME_LEN + 1 ];
  CDFstatus s = CDFlib( SELECT_, zVAR_, i_var,
                        GET_, zVAR_NAME_, var_name, NULL_ );
  if( s != CDF_OK )
      throw CDFFileBadRead(s,"Failed to SELECT_ & GET_ zVAR_NAME_" );
  return var_name;
}
// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------- get_cdf_variable_name
int CDFFileDataSource::get_cdf_variable_number( const string& var_name ) const
{
  select();
  long var_num;
// get the number of corresponding variable
  CDFstatus s = CDFlib( SELECT_, zVAR_NAME_, var_name.c_str(),
                        GET_, zVAR_NUMBER_, var_name.c_str(), &var_num, NULL_ );
  if( s != CDF_OK )
      throw CDFFileBadRead(s,"Failed to SELECT_ & GET_ zVAR_NUMBER_" );
  return var_num;
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------- get_num_cdf_attributes
int CDFFileDataSource::get_num_cdf_attributes(void) const
{
  select();
  long num_attr;
// get the total number of (global AND variable) atrributes
  CDFstatus s = CDFlib( GET_, CDF_NUMATTRS_, &num_attr, NULL_ );
  if( s != CDF_OK )
      throw CDFFileBadRead(s,"Failed to GET_ CDF_NUMATTRS_" );
  return num_attr;
}
// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------ get_cdf_attribute_scope
int CDFFileDataSource::get_cdf_attribute_scope( int i_attr ) const
{
  select();
  long attr_scope;
  CDFstatus s = CDFlib( SELECT_, ATTR_, i_attr,
                        GET_, ATTR_SCOPE_, &attr_scope, NULL_ );
  if( s != CDF_OK )
      throw CDFFileBadRead(s,"Failed to SELECT_, ATTR_ & GET_, ATTR_SCOPE_" );
  return attr_scope;
}
// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------ get_cdf_attribute_scope
int CDFFileDataSource::get_cdf_attribute_scope( const string& attr_name ) const
{
  select();
  long attr_scope;
  //string cdf_attr_name(attr_name);
  CDFstatus s = CDFlib( SELECT_, ATTR_NAME_, attr_name.c_str(),
                        GET_, ATTR_SCOPE_, &attr_scope, NULL_ );
  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to SELECT_, ATTR_NAME_ & GET_, ATTR_SCOPE_" );
  return attr_scope;
}
// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------- get_cdf_attribute_name
string CDFFileDataSource::get_cdf_attribute_name( int i_attr ) const
{
  select();
  char cdf_attr_name[ CDF_ATTR_NAME_LEN + 1 ];
  CDFstatus s = CDFlib( SELECT_, ATTR_, i_attr,
                GET_, ATTR_NAME_, cdf_attr_name, NULL_ );
  if( s != CDF_OK )
      throw CDFFileBadRead(s,"Failed to SELECT_ ATTR_ & GET_ ATTR_NAME_" );
  //string attr_name=get_attr_name(cdf_attr_name);
  return cdf_attr_name;
}
// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ----------------------------------------- get_cdf_attribute_number
int CDFFileDataSource::get_cdf_attribute_number( const string& attr_name ) const
{
  select();
  long attr_num;
  string cdf_attr_name(attr_name);
  CDFstatus s = CDFlib( SELECT_, ATTR_NAME_, attr_name.c_str(),
                GET_, ATTR_NUMBER_, attr_name.c_str(), &attr_num, NULL_ );
  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to SELECT_ ATTR_NAME_ & GET_ ATTR_NUMBER_" );
  return attr_num;
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------- get_gattr_data_type
//
// get data type and number of elements of specified attribute number
//   and entry number.
// the attribute must be global in scop
//
void CDFFileDataSource::get_gattr_data_type(
     int i_attr, int i_entry, long& data_type, long &num_elems ) const
{
  CDFstatus s;
  select();
  if( get_cdf_attribute_scope(i_attr) != GLOBAL_SCOPE )
    throw CDFFileBadRead(s,"get_gattr_data_type: Attribute is not global" );
  s = CDFlib( SELECT_, ATTR_, i_attr,
              SELECT_, gENTRY_, i_entry, NULL_ );
  if( s != CDF_OK )
    throw CDFFileBadRead(s,
    "get_gattr_data_type: Failed to SELECT_ ATTR_ SELECT_ gENTRY_" );
  s = CDFlib( GET_,
              gENTRY_DATATYPE_, &data_type,
              gENTRY_NUMELEMS_, &num_elems,      NULL_ );
  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to get info on G attr " );
}

// ++++++++++++++++++++++++++++++++++++++++++ class CDFFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------ create_data_names_table

void CDFFileDataSource::create_data_names_table(void)
{
  data_names_table = new DataNamesTable;

  select();
  int num_vars = get_num_cdf_variables();

// loop over variables
  for( int i_var=0; i_var<num_vars; i_var++ )
  {
    string var_name = get_cdf_variable_name( i_var );
    data_names_table->bindx( "variables/" + var_name, "" );
    data_names_table->bindx( "cdf_data_type/variables/" + var_name, "" );
  } // end of loop over variables

// loop over attributes
// this goes through ALL attr (global and variable)

  int num_attr = get_num_cdf_attributes();

  for( int i_attr=0; i_attr < num_attr; i_attr++ )
  {
    string attr_name = get_cdf_attribute_name(i_attr);
    if( get_cdf_attribute_scope(i_attr) == GLOBAL_SCOPE )
    {
      data_names_table->bindx( string(GLOBAL_ATTRIBUTES)+"/" + attr_name, "" );
      data_names_table->bindx( "cdf_data_type/" + string(GLOBAL_ATTRIBUTES)+"/" + attr_name, "" );
    } else
    {
// only add variable attribute if it exists for a particular variable
      for( int i_var=0; i_var<num_vars; i_var++ )
      {
// if no such zENTRY then skip the rest of this loop
        if( !cdf_vattribute_exists( i_var, i_attr ) )
          continue;
        string var_name = get_cdf_variable_name( i_var );
        data_names_table->bindx( string(VARIABLE_ATTRIBUTES)+"/" + var_name + "/" + attr_name, "" );
        data_names_table->bindx( "cdf_data_type/"+string(VARIABLE_ATTRIBUTES)+"/"
                                + var_name + "/" + attr_name, "" );
      }
    }
  } // end of loop over attributes

}


// +++++++++++++++++++++++++++++++++++++++ class CDF_TSFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------- CDF_TSFileDataSource

CDF_TSFileDataSource::CDF_TSFileDataSource ( string cdf_file_name )
  :   CDFFileDataSource( cdf_file_name )
{
// there is no harm in assuming that the file is an ISTP CDF ????

  set_ISTP_time_variables_default();

}


// +++++++++++++++++++++++++++++++++++++++ class CDF_TSFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------ set_time_variable

// for the cdf variable named variable_name, store the corresponding
// variable  name to be used for time tag, and a scaling and offset
// which are to be applied before computing the time as a Time
// if CDF EPOCH is used by the time variable,  then scaling=0.001
// and offset = 0
//
// NOTE: there is NO checking of variable existence or data type
//       all such conformance tests are left to the get_data functions

void CDF_TSFileDataSource::set_time_variable(
    const string& variable_name,
    const string& time_variable_name,
    double scaling, double offset )
{
  time_variable_table[ variable_name ] =
     CDFTimeVariableInfo( time_variable_name, scaling, offset );
}


// +++++++++++++++++++++++++++++++++++++++ class CDF_TSFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------- diag_print
void CDF_TSFileDataSource::diag_print(void)
{
  cout << "CDF_TSFileDataSource diag_print: " << endl;

  map<string,CDFTimeVariableInfo>::const_iterator ci;

  for( ci  = time_variable_table.begin();
       ci != time_variable_table.end(); ++ci )
  {
    cout << ci->first << ": "
         << ci->second.time_variable_name
         << "  [scaling: " << ci->second.scaling_to_seconds
         << ", offset: " << ci->second.offset_seconds_epoch_0AD << "]"<<endl;
  }

}


// +++++++++++++++++++++++++++++++++++++++ class CDF_TSFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------- set_ISTP_time_variables_default
//
// for variables with ISTP-like DEPEND_0 attribute, put an entry
//  into the time_variable_table. (leave other pre-existing
//  entries untouched)
//
// check cdf file is open
//   if not open, then try to re-open using file_name
// select cdf file
// get number of vars
// if number of vars is zero then exit
// foreach var number (starting at zero)
// select attribute by name "DEPEND_0"
//   if this fails then exit - ok, no ISTP-like variables in the file
// get number of vars
// if number of variables is zero then exit now!
// foreach var number (starting at zero)
//    use the var num as a zentry num
//    confirm zENTRY_EXISTENCE_ of zentry num
//    if zEntry exists (ie there is a DEPEND_0 for this var)
//      then: select zentry num (should work!)
//          : get zENTRY_DATATYPE_
//          : if it isn't CDF_CHAR or CDF_UCHAR
//          : then:  skip this var - DEPEND_0 MUST be string name of variable
//          : get zENTRY_NUMELEMS_ (gives num of chars in attribute)
//          : allocate buffer
//          : get zENTRY_DATA_ and convert to string for depend_0_var_name
//          : try to select zVAR_NAME_ of depend_0_var_name
//          : if DEPEND_0 attribute string value is NOT a variable name
//          : then:  skip this var - DEPEND_0 MUST be name of variable
//          : OK: add this var and depend_0_var_name to the
//          :   : time_variable_table, with offset=0 and scale=0.001
//               (NB: there is check of depend_0 datatype at this stage)
// end of loop over variables
//


void CDF_TSFileDataSource::set_ISTP_time_variables_default( void )
{
  CDFstatus s;

  if( open_status == false )
  {
    open( file_name );
  }
  select();

  string time_var_name;
  long time_var_max_rec_num=0;

  // get number of vars
  long num_vars;
  s = CDFlib( GET_, CDF_NUMzVARS_, &num_vars, NULL_ );
  if( s != CDF_OK )
    throw CDFFileBadRead(s,"Failed to GET_ CDF_NUMzVARS_" );

  if( num_vars == 0 ) return; // don't do anything!

  // find epoch variable
  for( long ivar=0; ivar < num_vars; ivar++ ) // loop over variables
  {

    // get data type of var
    long var_data_type,var_num_elems,var_max_rec_num;
    char var_name[ CDF_VAR_NAME_LEN + 1 ];
    s = CDFlib( SELECT_, zVAR_, ivar,
                GET_, zVAR_DATATYPE_, &var_data_type,
                zVAR_NUMELEMS_,  &var_num_elems,
		zVAR_MAXREC_, &var_max_rec_num,
          	zVAR_NAME_,var_name,NULL_);
    if( s != CDF_OK )
      continue;
    // epoch variable
    if(var_data_type == CDF_EPOCH)
    {
      // set name of default epoch variable to this one
      time_var_name=var_name;
      time_var_max_rec_num=var_max_rec_num;
      break;
    }
  }

  // loop through variables
  for( long ivar=0; ivar < num_vars; ivar++ ) // loop over variables
  {
    // get data type of var
    long var_data_type,var_num_elems,var_max_rec_num;
    char var_name[ CDF_VAR_NAME_LEN + 1 ];
    s = CDFlib( SELECT_, zVAR_, ivar,
                GET_, zVAR_DATATYPE_, &var_data_type,
                zVAR_NUMELEMS_,  &var_num_elems,
		zVAR_MAXREC_, &var_max_rec_num,
        	zVAR_NAME_,var_name,NULL_);
    if( s != CDF_OK )
      continue;
    // if epoch data then no associated time var
    if(var_data_type==CDF_EPOCH)
      continue;

    // select variable attribute
    bool no_depend0=false;
    s = CDFlib( SELECT_, ATTR_NAME_, "DEPEND_0", NULL_ );
    if(s!=CDF_OK)
      no_depend0=true;
    else
    {
      // check if there is a var entry for DEPEND_0 for this variable
      // if no such zENTRY then skip the rest of this loop
      if( CDFlib( CONFIRM_, zENTRY_EXISTENCE_, ivar, NULL_ ) != CDF_OK )
        no_depend0=true;
    }
    if(no_depend0)
    {
       if((var_max_rec_num!=time_var_max_rec_num) || time_var_max_rec_num==0)
         continue;
       if( time_var_name!="")
         set_time_variable( var_name, time_var_name, 0.001, 0.0 );
       continue;
    }
    // get data type of var attribute
    long vattr_data_type,vattr_num_elems;
    s = CDFlib( SELECT_, zENTRY_, ivar,
                GET_, zENTRY_DATATYPE_, &vattr_data_type,
                      zENTRY_NUMELEMS_, &vattr_num_elems,NULL_);
    if(s!=CDF_OK)
    {
      if( time_var_name!="")
        set_time_variable( var_name, time_var_name, 0.001, 0.0 );
      continue;
    }
    // skip rest of loop if this variable if not a string
    if( !(vattr_data_type == CDF_CHAR || vattr_data_type == CDF_UCHAR) )
    {
      if( time_var_name!="")
         set_time_variable( var_name, time_var_name, 0.001, 0.0 );
      continue;
    }

    // allocate string buffer and fill with null
    char* depend0_var_name = (char*)malloc( vattr_num_elems + 1 );
    for( int i=0; i<=vattr_num_elems; i++ ) depend0_var_name[i]='\0';

    s = CDFlib( GET_, zENTRY_DATA_, static_cast<void*>(depend0_var_name), NULL_ );
    if( s != CDF_OK )
    {
      if( time_var_name!="")
        set_time_variable( var_name, time_var_name, 0.001, 0.0 );
      free( depend0_var_name );
      continue;
    }

    // if there does NOT exist a zVar with this name then free the var_name string
    // and skip rest of loop
    s = CDFlib( SELECT_, zVAR_NAME_, depend0_var_name, NULL_ );
    if( s != CDF_OK )
    {
      if( time_var_name!="")
        set_time_variable( var_name, time_var_name, 0.001, 0.0 );
      free( depend0_var_name );
      continue;
    }

    // check that the datatype of DEPEND_0 is ISTP like
    // ie EPOCH datatype and zero dimensions
    long depend0_datatype, num_dims;
    s = CDFlib( GET_, zVAR_DATATYPE_, &depend0_datatype,
                     zVAR_NUMDIMS_,  &num_dims, NULL_ );
    if( s != CDF_OK )
    {
      if( time_var_name!="")
        set_time_variable( var_name, time_var_name, 0.001, 0.0 );
      free( depend0_var_name );
      continue;
    }

    // datatype wrong, or num dims >=2, then skip rest of loop
    // Actually ... we will accept ANY numeric type!
    if( (    depend0_datatype ==CDF_CHAR
          || depend0_datatype == CDF_UCHAR )
        || num_dims >=2 )
    {
      free( depend0_var_name );
      continue;
    }

    // if num dims == 1 then skip rest of loop if dims size != 1
    if( num_dims == 1 )
    {
      long dim_sizes[1];
      s = CDFlib( GET_, zVAR_DIMSIZES_, dim_sizes, NULL_ );
      if( s != CDF_OK )
      {
        free( depend0_var_name );
	continue;
      }
      if( dim_sizes[0] != 1 )
      {
        free( depend0_var_name );
        continue;
      }
    }

     // at this point: depend0_var_name is name of variable var_name in CDF file
     //                AND ... it is of right data type
     //                AND ... it is either dimensionality 0:[] or 1:[1]

     set_time_variable( var_name, depend0_var_name, 0.001, 0.0 );

     free( depend0_var_name );

  } // end of loop over variables ivar

}



// +++++++++++++++++++++++++++++++++++++++ class CDF_TSFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------- get_data_by_variable
//
// return TSDataBuffer for data in a given time interval
//

DataBuffer_var CDF_TSFileDataSource::get_data_by_variable(
    const string& var_name, const TimeInterval& req_time_interval,
    int subsample_cycle, int subsample_offset )
{
  TSDataBuffer_var tsdata_var = new TSDataBuffer; // return object

  CDFVarInfo cdf_var_info( var_name );

  if( time_variable_table.count( var_name ) == 0 )
    throw CDF_TSNoTimeVariable(0,"No time variable entry for variable: "+var_name);

  CDFTimeVariableInfo time_var_info = time_variable_table[ var_name ];
  CDFVarInfo cdf_time_var_info( time_var_info.time_variable_name );

  if( cdf_time_var_info.data_type == CDF_CHAR
      || cdf_time_var_info.data_type == CDF_UCHAR )
    throw CDF_TSTimeTagsNonConformant(0,
       "Non-numeric time variables not supported" );

  bool data_is_numeric = !( cdf_var_info.data_type == CDF_CHAR
                           || cdf_time_var_info.data_type == CDF_UCHAR );

  if( cdf_var_info.max_rec_num != cdf_time_var_info.max_rec_num )
  {
    throw CDF_TSTimeTagsNonConformant(0,
       "Data and time variables do not have same number of records" );
  }

  int num_rec = cdf_time_var_info.max_rec_num + 1; // number of records to read

  int rec_count = 1000; // size of chunk of records

  int i_okdata = -1; // count of acceptable data (for subsampling algorithm)

// prettify subsample inputs
  if( subsample_cycle < 0 ) subsample_cycle = -subsample_cycle;
  if( subsample_cycle == 0 ) subsample_cycle = 1;
  if( subsample_offset < 0 ) subsample_offset = -subsample_offset;
  subsample_offset = subsample_offset % subsample_cycle;

  if( data_is_numeric )
  {
    NumericVDataBuffer* data_buff_p;
    NumericVDataBuffer* ttags_buff_p;
    DataBuffer_var file_data_buff_var;
    DataBuffer_var file_ttags_buff_var;

// get start and end epochs of time interval ranges
    double epch_end=cdfEpoch(req_time_interval.end_abs());
    double epch_start=cdfEpoch(req_time_interval.start_abs());

// get the data ignoration functor
    IgnoreNumericDataFunctor_var ignore_data_function = numeric_ignoration_table[var_name];
// loop over records
    for( int r=0; r<num_rec;  r+=rec_count )
    {
// adjust n_rec_chunk for end of sequence
      if( (r+rec_count) >= (num_rec+1) )
        rec_count = num_rec - r;

      file_data_buff_var = CDFFileDataSource::get_data_by_variable(
                             var_name, r, rec_count );
      file_ttags_buff_var = CDFFileDataSource::get_data_by_variable(
                              time_var_info.time_variable_name, r, rec_count );

      NumericVDataBuffer_var data_nvdbf_var =
        NumericVDataBuffer_var::narrow( file_data_buff_var );
      if( data_nvdbf_var.is_nil() )
        throw CDF_TSDataNonConformant(0,"Data buffer unexpectedly not numeric");
      NumericVDataBuffer_var ttags_nvdbf_var =
        NumericVDataBuffer_var::narrow( file_ttags_buff_var );
      if( ttags_nvdbf_var.is_nil() )
        throw CDF_TSTimeTagsNonConformant(0,
          "Non-numeric time variables not supported" );
//      cout << ttags_nvdbf_var->size() << endl;
// set up data buffers for return data, after first read of records
      if( r==0 )
      {
// data dimensionality taken from first read of records
        data_buff_p = new NumericVDataBuffer( *(data_nvdbf_var->descriptor()) );
// ttags assumed scalar
        ttags_buff_p = new NumericVDataBuffer();
      }
// loop over items (an item is an object like a matrix with one or
//   more elements)
      int num_items = data_nvdbf_var->num_items();
      int num_item_elts = data_nvdbf_var->num_item_elts();
//      cout << "number of items to check: " << num_items << endl;
//      cout << "number of elements per item : " << num_item_elts << endl;
      for(int i=0; i<num_items; i++ )
      {
        double epch_i=(*ttags_nvdbf_var)[i];
	Time t_rec_t=cdfEpoch(epch_i);
        bool ignore_item = ignore_data_function.is_nil() ?
                             false :
                             ignore_data_function->ignore_data_item( *data_nvdbf_var, i );
	if(((floor(epch_end+0.5)-floor(epch_i+0.5))>=0.) &&
	   ((floor(epch_i+0.5)-floor(epch_start+0.5))>=0.))
        {
          i_okdata++;
          if( !ignore_item && (i_okdata - subsample_offset )% subsample_cycle == 0 )
          {
            ttags_buff_p->push_back( Seconds(t_rec_t ) );
            for( int ielt = i*num_item_elts; ielt<(i+1)*num_item_elts; ielt++)
              data_buff_p->push_back( (*data_nvdbf_var)[ielt] );
          }
        }
      }

    } // end of loop over file records

    tsdata_var->data = data_buff_p;
    tsdata_var->ttags = ttags_buff_p;
    if( ttags_buff_p->num_items() != 0 )
      tsdata_var->time_interval.set(
         Seconds((*ttags_buff_p)[0]),
         Seconds((*ttags_buff_p)[ttags_buff_p->size()-1] )  );

  } else  // end of numeric, start of string data
  {
// THIS IS SAME AS NUMERIC SECTION BUT WITH StringVDataBuffer
// AND DATA IGNORATION IS NOT YET IMPLEMENTED
    StringVDataBuffer* data_buff_p;
    NumericVDataBuffer* ttags_buff_p;
    DataBuffer_var file_data_buff_var;
    DataBuffer_var file_ttags_buff_var;

// get the data ignoration functor
//    IgnoreStringDataFunctor_var ignore_data_function =
//       string_ignoration_table[var_name];
// loop over records
    for( int r=0; r<num_rec;  r+=rec_count )
    {
// adjust n_rec_chunk for end of sequence
      if( (r+rec_count) >= (num_rec+1) )
        rec_count = num_rec - r;

      file_data_buff_var = CDFFileDataSource::get_data_by_variable(
                             var_name, r, rec_count );
      file_ttags_buff_var = CDFFileDataSource::get_data_by_variable(
                              time_var_info.time_variable_name, r, rec_count );

      StringVDataBuffer_var data_svdbf_var =
        StringVDataBuffer_var::narrow( file_data_buff_var );
      if( data_svdbf_var.is_nil() )
        throw CDF_TSDataNonConformant(0,"Data buffer unexpectedly not string");
      NumericVDataBuffer_var ttags_nvdbf_var =
        NumericVDataBuffer_var::narrow( file_ttags_buff_var );
      if( ttags_nvdbf_var.is_nil() )
        throw CDF_TSTimeTagsNonConformant(0,
          "Non-numeric time variables not supported" );
// set up data buffers for return data, after first read of records
      if( r==0 )
      {
// data dimensionality taken from first read of records
        data_buff_p = new StringVDataBuffer( *(data_svdbf_var->descriptor()) );
// ttags assumed scalar
        ttags_buff_p = new NumericVDataBuffer();
      }
// loop over items (an item is an object like a matrix with one or
//   more elements)
      int num_items = data_svdbf_var->num_items();
      int num_item_elts = data_svdbf_var->num_item_elts();
//      cout << "number of items to check: " << num_items << endl;
//      cout << "number of elements per item : " << num_item_elts << endl;
      for(int i=0; i<num_items; i++ )
      {
        double t_rec = Seconds(cdfEpoch((*ttags_nvdbf_var)[i] ));
        bool ignore_item = false;

        if( req_time_interval.contains(Seconds( t_rec )) )
        {
          i_okdata++;
          if( !ignore_item && (i_okdata - subsample_offset )% subsample_cycle == 0 )
          {
            ttags_buff_p->push_back( t_rec );
            for( int ielt = i*num_item_elts; ielt<(i+1)*num_item_elts; ielt++)
              data_buff_p->push_back( (*data_svdbf_var)[ielt] );
          }
        }
      }

    } // end of loop over file records

    tsdata_var->data = data_buff_p;
    tsdata_var->ttags = ttags_buff_p;
    if( ttags_buff_p->num_items() != 0 )
      tsdata_var->time_interval.set(
          Seconds((*ttags_buff_p)[0]) ,
          Seconds((*ttags_buff_p)[ttags_buff_p->size()-1] ) );
  } // end of getting string data

  return tsdata_var;
}
// +++++++++++++++++++++++++++++++++++++++ class CDF_TSFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------- get_time_range_by_variable
//
// return time range for data
//

TimeInterval CDF_TSFileDataSource::get_time_range_by_variable(
                 DataBuffer_var file_ttags_buff_var)
{
  TimeInterval time_range; // return value

    NumericVDataBuffer_var ttags_nvdbf_var =
      NumericVDataBuffer_var::narrow( file_ttags_buff_var );
    if( ttags_nvdbf_var.is_nil() )
      throw CDF_TSTimeTagsNonConformant(0,
        "Non-numeric time variables not supported" );

    int num_items = ttags_nvdbf_var->num_items();

    if(num_items<1)
    {
      cout << "CDF_TSFileDataSource::get_time_range_by_variable: " \
	   << " no records" << endl;
      return time_range;
    }

// loop over items (an item is an object like a matrix with one or
//   more elements)

    double epch1 = (*ttags_nvdbf_var)[0];
    double epch2 = (*ttags_nvdbf_var)[num_items-1];

      Time t_rec1 = Seconds(cdfEpoch( epch1 ));
      Time t_rec2 = Seconds(cdfEpoch( epch2 ));
      try{
        Time t1(t_rec1.iso_srep());
        Time t2(t_rec2.iso_srep());
	bool bad_t1=t1.get_time_rep()<-6.e10;
	bool bad_t2=t2.get_time_rep()<-6.e10;
	if(!bad_t1 && !bad_t2)
          time_range.set( t_rec1, t_rec2 );
	else if(bad_t1)
	  time_range.set(t_rec2, t_rec2);
	else if(bad_t2)
	  time_range.set(t_rec1, t_rec1);
	else ;
      }
      catch(Exception& e)
      {
        cout << "Invalid time " << endl;
        e.print_msg();
      }

// handle case when not monotonic
    double epch;
    for(int i=1; i<num_items-1; i++ )
    {
      epch = (*ttags_nvdbf_var)[i];
      if( epch < epch1 || epch > epch2){
        Time t_rec = Seconds(cdfEpoch( epch ));
        try{
          Time t(t_rec.iso_srep());
	  // bad time tag
	  if(t.get_time_rep()<-6.e10)
	    continue;
        }
        catch(Exception& e)
        {
          cout << "Invalid time at record " << i << " for variable " << endl;
          e.print_msg();
          continue;
        }

        if( t_rec < time_range.start() ) time_range.set_start( t_rec );
        if( t_rec > time_range.end() ) time_range.set_end( t_rec );
      }
    } // end loop over items


  return time_range;
}
// +++++++++++++++++++++++++++++++++++++++ class CDF_TSFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------- get_data_availability_by_variable
//
//  data_name and resolution of the  input data_availability are
//  not changed. resolution is used in this function.
//  data_name is not used in this function.
//  time_interval is set from that in cfd_ts file
//

void CDF_TSFileDataSource::get_data_availability_by_variable(
   const string& var_name, DataAvailability& data_availability )
{

  CDFVarInfo cdf_var_info( var_name );

  if( time_variable_table.count( var_name ) == 0 )
    throw CDF_TSNoTimeVariable(0,"No time variable entry for variable: "+var_name);

  CDFTimeVariableInfo time_var_info = time_variable_table[ var_name ];
  CDFVarInfo cdf_time_var_info( time_var_info.time_variable_name );

  if( cdf_time_var_info.data_type == CDF_CHAR
      || cdf_time_var_info.data_type == CDF_UCHAR )
    throw CDF_TSTimeTagsNonConformant(0,
       "Non-numeric time variables not supported" );

  if( cdf_var_info.max_rec_num != cdf_time_var_info.max_rec_num )
    throw CDF_TSTimeTagsNonConformant(0,
       "Data and time variables do not have same number of records" );

  DataBuffer_var file_ttags_buff_var;

  int num_rec = cdf_time_var_info.max_rec_num + 1; // number of records to read

  file_ttags_buff_var = CDFFileDataSource::get_data_by_variable(
                            time_var_info.time_variable_name, 0, num_rec );

  TimeInterval time_range=get_time_range_by_variable(file_ttags_buff_var);
  //cout << "time range is " << time_range << endl;
  double resolution=time_range.duration_abs().get_time_rep()*.01;
  data_availability.set_resolution(resolution);
  //cout << "resolution is " << resolution << endl;
  data_availability.set_time_interval(time_range);


    NumericVDataBuffer_var ttags_nvdbf_var =
      NumericVDataBuffer_var::narrow( file_ttags_buff_var );
    if( ttags_nvdbf_var.is_nil() )
      throw CDF_TSTimeTagsNonConformant(0,
        "Non-numeric time variables not supported" );

// loop over items (an item is an object like a matrix with one or
// more elements) These are the buffered time tags in blocks of num_items

    int num_items = ttags_nvdbf_var->num_items();
    if(num_items<1)
      return;
// modified to only update availability array if tt is in next bin
// working in epoch double is MUCH faster than cast to Time

    double step = resolution*1000; // number msecs between array entries
    double last;                   // push 1 tag into interval
                                   // to avoid bin edges
    if(num_items>1)
      last=(*ttags_nvdbf_var)[1] - step;
    else
      last=(*ttags_nvdbf_var)[0] - step;

    double prev = (*ttags_nvdbf_var)[0];
    for(int i=0; i<num_items; i++ )
    {
     double next = (*ttags_nvdbf_var)[i];
     // trap non-monotonic sequence - go back and start here
     if (prev > next) last = next - step;

     if (i == num_items-1 || next >= last + step ){
       // first, last or one bin further on
       Time t_rec = Seconds(cdfEpoch(next));
       try{
        Time t(t_rec.iso_srep());
       }
       catch(Exception& e)
       { e.print_msg();continue;}

       data_availability.set_data_available( t_rec );
       last += step;
      }
    } // end loop over items



}

// +++++++++++++++++++++++++++++++++++++++ class CDF_TSFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// --------------------------------------------------------- get_data
DataBuffer_var CDF_TSFileDataSource::get_data(
   const string& data_name, const TimeInterval& time_interval )
{
  DataBuffer_var nil_object; // for nil return
  Name dname( data_name ); // split into components
  if(dname.num_components()==1)
  {
     if(cdf_variable_exists(data_name))
       return get_data_by_variable(data_name,time_interval);
     else
       return nil_object;
  }
  if( dname.num_components() == 2 )
  {
    if( dname[0] == "variables" )
      return get_data_by_variable( dname[1].get_id(), time_interval );
    else
      return nil_object;
  }
  else
    return nil_object;

// lots more to go here for other data name components

}

// +++++++++++++++++++++++++++++++++++++++ class CDF_TSFileDataSource
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------- get_dcm_entries
void CDF_TSFileDataSource::get_dcm_entries( FAR::FARData& dcm_entries,
                              const string& data_name_root,
                              const string& file_name_root,
                              double data_availability_resolution)
{
//  dcm_entries.clear();

  time_t clock = time(0);
  string now_time = asctime(localtime(&clock));
// get rid of the trailing newline!
  now_time.erase(24);

  FAR::Record comment;
  comment.add_field("comment");
  comment.add_field("make_dcm_entry " + file_name);
  comment.add_field(now_time);
  dcm_entries.add_record(comment);

  int nvar = get_num_cdf_variables();
//  int nattr= get_num_cdf_attributes();
  for( int ivar=0; ivar<nvar; ivar++ )
  {
    try{
      string var_name = get_cdf_variable_name(ivar);
      CDFVarInfo cdf_var_info( var_name );

      // No entries for variables without an associated time variable
      if( time_variable_table.count( var_name ) == 0 )
      {
        //cout << "No time variable for cdf variable: " << var_name << endl;
        continue;
      }
      CDFTimeVariableInfo time_var_info = time_variable_table[ var_name ];
      //CDFVarInfo cdf_time_var_info( time_var_info.time_variable_name );

      DataAvailability da;
      da.set_resolution( data_availability_resolution );
      get_data_availability_by_variable( var_name, da );
      TimeInterval time_range = da.get_time_interval();

      char cbuff[50]; sprintf(cbuff,"%d", cdf_var_info.max_rec_num+1);
      string rec_num_str = cbuff;

      const int rec_sz(DCM_CDF_TS_DATA_NUM_FLDS);
      vector<string> flds(rec_sz);
      FAR::Record rec(flds);

      rec[(int)DCM_RECORD_TYPE_FLD]=string(DCM_DATA_RECORD);
      rec[(int)DCM_DATA_TYPE_FLD]=string(DCM_CDF_TS_DATA);
      // we need to pass in priority as an argument also
      rec[(int)DCM_CDF_TS_DATA_PRIORITY_FLD]="0";
      rec[(int)DCM_CDF_TS_DATA_DATASOURCE_FLD]=data_name_root + var_name;
      rec[(int)DCM_CDF_TS_DATA_VAR_NAME_FLD]=var_name;
      rec[(int)DCM_CDF_TS_DATA_TIME_VAR_NAME_FLD]=\
        time_var_info.time_variable_name;
      rec[(int)DCM_CDF_TS_DATA_TIME_TYPE_FLD]="cdf_epoch";
      Name file_name_name(file_name);
      rec[(int)DCM_CDF_TS_DATA_FILE_NAME_FLD]=file_name_root + \
        file_name_name.get_component(file_name_name.num_components()-1).get_id();
      rec[(int)DCM_CDF_TS_DATA_TIME_INT_START_FLD]=\
        time_range.start().iso_srep();
      rec[(int)DCM_CDF_TS_DATA_TIME_INT_END_FLD]=\
        time_range.end().iso_srep();
      rec[(int)DCM_CDF_TS_DATA_NUM_ITEMS_FLD]=rec_num_str;
      rec[(int)DCM_CDF_TS_DATA_AVAILABILITY_FLD]=da.srep();
      dcm_entries.add_record(rec);
    }
    catch(Exception& e)
    { e.print_msg(); continue;}

  }

}

// ---------------------- END OF NAMESPACE --------------------------
// ------------------------------------------------------------------
} // namespace QDOS
