#include "MatrixSVD.h"
//#include <g2c.h>

namespace QDOS{

// --- Cover functions for CLAPACK routines
int svdcomp(char *jobu, char *jobvt, long int *m, long int *n, \
    double *a, long int *lda, double *s, double *u, \
    long int *ldu, double *vt, long int *ldvt, long int *info)
{
    // Cover function for doubles 
    int res;
    long int mindim=std::min(*m,*n);
    long int lwork=2*std::max(3*mindim+std::max(*m,*n),5*mindim-4);        
    #ifdef DIAG
    cout << "creating working array" << endl;
    #endif
    double* work=new double [lwork];
    res=dgesvd_(jobu,jobvt,m,n,a,lda,s,u,ldu,vt,ldvt,work,&lwork,info);
    if (0!=*info)
    {
        throw MatrixValueError(ExceptionMsg()<< \
            "Internal error in CLAPACK: DGESVD()");
    }             
    delete work;
    return res;
}
   
int svdcomp(char *jobu, char *jobvt, long int *m, long int *n, \
    double_complex *a, long int *lda, double *s, double_complex *u, \
    long int *ldu, double_complex *vt, long int *ldvt, long int *info)
{
    // Cover function for complex<double> 
    int res;
    long int mindim=std::min(*m,*n);
    long int lwork=2*(2*mindim+std::max(*m,*n));        
    #ifdef DIAG
    cout << "creating working array" << endl;
    #endif
    doublecomplex* work=new doublecomplex [lwork];
    doublecomplex* a_cmp=reinterpret_cast<doublecomplex*>(a);
    doublecomplex* u_cmp=reinterpret_cast<doublecomplex*>(u);
    doublecomplex* vt_cmp=reinterpret_cast<doublecomplex*>(vt);
      
    double* rwork=new double [std::max(3*mindim,5*mindim-4)];
    res=zgesvd_(jobu,jobvt,m,n,a_cmp,lda,s,u_cmp,ldu,vt_cmp,ldvt,work,\
        &lwork,rwork,info);
    if (0!=*info)
    {
        throw MatrixValueError(ExceptionMsg()<< \
            "Internal error in CLAPACK: ZGESVD()");
    }             
    delete work;
    delete rwork;
    return res;
}

}  // end namespace QDOS        

  
