//
//  QObject.cc
//
//  Qdos2 - base object class
//
//  David Burgess
//  July 1999
//

#include <iostream>

#include "QObject.h"
#include "Adapter.h"
#include "StringRep.h"

namespace QDOS {

//TypeInfo_var QObject::_typeinfo=new TypeInfo("QdObject");

int QdObject::qobject_class_counter = 0;

//DEF_QDOBJECT(QdObjectSeq,"QdObjectSeq")

QdObject::QdObject(void) : xrefs( new ObjectNamesTable )
{
  object_id=++qobject_class_counter;
#ifdef DIAG
  cout<<"QdObject constructor (Object #"<<object_id<<") refs#"
      << get_ref_count()
      << endl;
#endif
}

QdObject::~QdObject()
{
#ifdef DIAG
 cout << "QObject destructor (Object # " << object_id << ") refs#"
      << get_ref_count()
      << endl;
#endif
 }

void QdObject::list_xref_names( vector<string>& sl )
{
  xrefs->list_names_r( sl );
}

void QdObject::copy_xrefs_from(QdObject_var from_obj)
{
    vector<string> sl;
    from_obj->list_xref_names(sl);
    for(int i=0;i<sl.size();i++)
    {
        if(sl[i][0]==CACHE_XREF_CHAR) continue;
		
        QdObject_var xref_obj=from_obj->get_xref(sl[i]);
        change_xref(sl[i],xref_obj);
    }
}

void QdObject::copy_xrefs_to(QdObject_var to_obj)
{
    vector<string> sl;
    list_xref_names(sl);
    for(int i=0;i<sl.size();i++)
    {
        if(sl[i][0]==CACHE_XREF_CHAR)
	  continue;
        QdObject_var xref_obj=get_xref(sl[i]);
        to_obj->change_xref(sl[i],xref_obj);
    }
}

string QdObject::size_srep(const StringRep& stringrep, int sz ) const
{ string s; return s;}

FunctionAdapter* QdObject::join(QdObject_var x_in, QdObject_var y_in,\
    QdObject_var x_out, QdObject_var y_out, \
    QdObject_var join_method, QdObject_var gap_value, \
    QdObject_var gap_handler) const
{ throw Exception("[QObjectException][FunctionNotDefined]");}

FunctionAdapter* QdObject::multijoin(QdObjectSeq_var x_in, \
    QdObjectSeq_var y_in,QdObject_var x_out,QdObjectSeq_var y_out,\
    QdObjectSeq_var join_method,\
    QdObjectSeq_var gap_value, QdObjectSeq_var gap_handler,\
    int triage_cnt) const
{ throw Exception("[QObjectException][FunctionNotDefined]");}

void join(QdObject_var x_in, QdObject_var y_in,\
      QdObject_var x_out, QdObject_var y_out,\
      QdObject_var join_method, QdObject_var gap_value, \
      QdObject_var gap_handler)
{ FunctionAdapter* join_adapter=y_in->join(x_in,y_in,x_out,y_out,\
    join_method,gap_value,gap_handler);
    join_adapter->run();
    delete join_adapter;
}

void multijoin(QdObjectSeq_var x_in,QdObjectSeq_var y_in,\
    QdObject_var x_out, QdObjectSeq_var y_out, \
    QdObjectSeq_var join_method, QdObjectSeq_var gap_value, \
    QdObjectSeq_var gap_handler, int triage_cnt)
{
    FunctionAdapter* multijoin_adapter=\
        x_out->multijoin(x_in,y_in,x_out,y_out,join_method,\
        gap_value,gap_handler,triage_cnt);
    multijoin_adapter->run();
    delete multijoin_adapter;
}

Vistream& operator>>(Vistream& v_is,QdObject_var obj)
{
  static int restoreCallCounter=0;
#ifdef DIAG
  cout << "restoreCallCounter is " << restoreCallCounter << endl;
#endif

  static map<int,QdObject_var> restoredObjectIds;
  if(restoreCallCounter==0)
  {
#ifdef DIAG
     cout << "resetting map " << endl;
#endif
     restoredObjectIds.clear();
  }

#ifdef DIAG
  cout << "size of restored Objects is " <<
     distance(restoredObjectIds.begin(),restoredObjectIds.end()) << endl;
#endif
  if(obj.is_nil())
    return v_is;
  obj->get(v_is);
  int sz=0;
  v_is >> sz;
  vector<string> sl(sz);
  for(int i=0;i<sz;i++)
  {
    v_is >> sl[i];
    string type;
    v_is >> type;
    int obj_id=0;
    QdObject_var xref_obj;
    string obj_type;
    if(type==NULL_OBJECT)
      continue;
    else if(type==OBJECT_ID || type==OBJECT)
      v_is >> obj_id;
    if(type==OBJECT_ID)
    {
      map<int,QdObject_var>::iterator itr=restoredObjectIds.find(obj_id);
      if(itr==restoredObjectIds.end())
      {
        cout << "Could not find saved object id " << obj_id << " on restore" << endl;
	continue;
      }
      else
      {
        xref_obj=itr->second;
        obj->change_xref(sl[i],xref_obj);
	continue;
      }
    }
    else if(type==OBJECT)
      v_is >> obj_type;
    else
      obj_type=type;

    if(xref_obj.is_nil() && obj_type!="")
      xref_obj=QdObject::create(obj_type);
    if(!xref_obj.is_nil())
    {
      restoredObjectIds[obj_id]=xref_obj;
      restoreCallCounter++;
      v_is >> xref_obj;
      if(restoreCallCounter)
        restoreCallCounter--;
      else
        cerr << "cannot decrement restoreCallCounter as zero!!" << endl;

    }
    else
      continue;
    //restoredObjectIds[obj_id]=xref_obj;
    obj->change_xref(sl[i],xref_obj);
  }
  return v_is;
}

Vostream& operator<<(Vostream& v_os, QdObject_var obj)
{
  static int saveCallCounter=0;
#ifdef DIAG
  cout << "saveCallCounter is " << saveCallCounter << endl;
#endif

  static map<int,QdObject_var> savedObjectIds;
  if(saveCallCounter==0)
  {
#ifdef DIAG
     cout << "resetting map " << endl;
#endif
     savedObjectIds.clear();
  }

#ifdef DIAG
  cout << "size of saved Objects is " <<
     distance(savedObjectIds.begin(),savedObjectIds.end()) << endl;
#endif

   if(obj.is_nil())
     return v_os;
   obj->put(v_os);
   vector<string> sl;
   obj->list_xref_names(sl);
   int sz=sl.size();
   v_os << sz;
   for(int i=0;i<sz;i++)
   {
     string name=sl[i];
     int obj_id=0;

     QdObject_var xref_obj;
     if(sl[i][0]!=CACHE_XREF_CHAR)
       xref_obj=obj->get_xref(sl[i]);
     string type;
     if(xref_obj.is_nil())
       type=NULL_OBJECT;
     else
     {
       obj_id=xref_obj->get_id();
       map<int,QdObject_var>::iterator itr=savedObjectIds.find(obj_id);
       if(itr!=savedObjectIds.end())
       {
          if(xref_obj.ptr_()==itr->second.ptr_())
	    type=OBJECT_ID;
	  else
            type=OBJECT;
       }
       else
         type=OBJECT;
     }
     v_os << name;
     v_os << type;
     if(type==NULL_OBJECT)
       continue;
     v_os << obj_id;
     if(type==OBJECT_ID)
       continue;
     string obj_type=xref_obj->object_type();
     v_os << obj_type;
#ifdef DIAG
     cout << "Saving " << name << " with obj id " << obj_id << endl;
#endif
     savedObjectIds[obj_id]=xref_obj;
     saveCallCounter++;
     v_os << xref_obj;
     if(saveCallCounter)
       saveCallCounter--;
     else
       cerr << "cannot decrement saveCallCounter as zero!!" << endl;
   }
   return v_os;
}

} // end namespace QDOS
