#include "qdutil_copies.h"

//a small number
static double CLOSE_ENOUGH = 1.e-4;

void QuSetTxtAttr(char *attrName, char *value, QdObject_var obj)
{
    //create Xref DO
    try
    {
      if(obj->xref_exists(attrName))
         obj->delete_xref(attrName); 
	 
      QdString_var DOx = new QdString(value);
    
      // Add Xref called "attrName" to parent Object
      // In this case DOx (the attribute) does NOT have a name of its own
      obj->set_xref(attrName, DOx);
    }
    catch(Exception &e)
    {
     return; //no warning 
    }
   
  return;

 } // end QuSetTextAttr



bool QuIsThreeVector(QdObject_var obj)
{
  QdRMatrix_var rmatrix=QdRMatrix_var::narrow(obj);
  if(rmatrix.is_nil())
  {
    QdRMatrixSeq_var rmatrixseq=QdRMatrixSeq_var::narrow(obj);
    if(rmatrixseq.is_nil())
      return false;
    if(!rmatrixseq->sequence_size())
      return false;

    try
    {
      if((rmatrixseq[0].dims()).size()!=1)
        return false;
      if((rmatrixseq[0].dims())[0]!= 3)
        return false;
      if(!rmatrixseq->is_homogeneous_seq())
        return false;
      return true;
    }
    catch(Exception &e)
    {
       e.print_msg();
       return false;
    }
  } 
  
  try
  {
    if((rmatrix->dims()).size()!=1)
      return false;
    if((rmatrix->dims())[0]!=3)
      return false;
    return true;
  }
  catch(Exception &e)
  {
    e.print_msg();
    return false;
  }
}

bool QuIsTSRegular(QdTimeSeq_var tt, double *spacing){
  double err;
  double Gap, Gaptest;
   
  *spacing = 0;
    
  if ( tt.is_nil() ) return false;
  if( !tt->is_sequence() ) return false;

  int n_seq = tt->sequence_size();

  if( n_seq<2 ) return false;
  
  
  Gap = (double) mSeconds(tt[1]) - (double) mSeconds(tt[0]);
  
  *spacing = Gap/1000.;
  if (Gap < 1) return false;  // work to no better than msec accuracy

  
  for (int i=2; i<n_seq; i++){
    Gaptest = (double)mSeconds(tt[i]) - (double)mSeconds(tt[i-1]);
  
    err = fabs((Gap - Gaptest)/Gap);
    if (err > CLOSE_ENOUGH ) return false;
  }
  return true;
}


bool Qu_is3Vector(QdObject_var obj)
{ return QuIsThreeVector(obj);}

bool QuIsBadData(double value, double fill)
{
 
    if( fill == 0 && value == fill) return true;
    else if ( fabs((value - fill)/fill) < CLOSE_ENOUGH ) return true;
   

    if ( !finite(value) ) return true; // finds NaN, Infinity, -Infinity
  
  
  return false;
}

QdRScalarSeq_var QuCleanObject(QdRScalarSeq_var dobj){
	
	  QdObject_var fill_xref;
  	  double fill;
  	  bool notify = false;
  	  QdRScalarSeq_var obj_out;
      

    if (dobj->xref_exists("FILLVAL")) 
    {fill_xref = dobj->get_xref("FILLVAL"); }
    else
    fill = -3.4e38; // in case xref not set
  
    QdTimeSeq_var ts_in=get_timetags(dobj);
    QdTimeSeq_var ts_out=new QdTimeSeq;
    
    // scalar sequence case
    
   QdRScalarSeq_var ss_in = dobj;
   if(!ss_in.is_nil())
   {
     if(ss_in->empty())
       {  return obj_out; }
          
      QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);

      if( !fill_v.is_nil())
      {   fill = (double) *fill_v; }
      else{ fill = -3.4e38;}
      
      // initialise mask
      SequenceMask msk(ss_in->size(),true); 
      
      for(int i=0;i<ss_in->size();i++)
          msk[i] = !QuIsBadData(ss_in[i], fill);
       
      QdRScalarSeq_var ss_out=new QdRScalarSeq;
      ss_in->get_subsequence(*ss_out,msk);
      ss_out->copy_xrefs_from(ss_in);
      
      if( !ts_in.is_nil() ){
        ts_in->get_subsequence(*ts_out,msk);
        ts_out->copy_xrefs_from(ts_in);
        set_timetags(ss_out,ts_out);
      }
      return ss_out;
   }
   
   QdRScalar_var s_in = QdRScalar_var::narrow(dobj);
   if(!s_in.is_nil())
   {
      QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);
      if( !fill_v.is_nil())
      { fill = (double) *fill_v; }

      if( QuIsBadData( *s_in, fill))
          return obj_out; // empty ptr
      else return dobj; // return original object
   }

   // string and other input types not cleaned - return input object
   
   return dobj;
   
}

QdObject_var QuCleanObjectMatrix(QdObject_var dobj){
  QdObject_var fill_xref;
  double fill;
  bool notify = false;
  bool hasTT;
  QdObject_var obj_out;
      

    if (dobj->xref_exists("FILLVAL")) 
      fill_xref = dobj->get_xref("FILLVAL");
    
    fill = -3.4e38; // in case xref not set
  
    
    
    QdTimeSeq_var ts_out=new QdTimeSeq;
    
    // is sequence a matrix sequence
    QdRMatrixSeq_var ms_in=QdRMatrixSeq_var::narrow(dobj);
    QdTimeSeq_var ts_in=get_timetags(dobj);
    hasTT = !(ts_in.is_nil());
    if(!ms_in.is_nil())
    {
      if(ms_in->empty())
          return obj_out;
	  
	  
          
      // if so, do we have a single fill-value or a matrix of different fill     
      // values for each dimension
      
      QdRMatrix_var fill_m=QdRMatrix_var::narrow(fill_xref);
      
      // ...either way create matrix of fill values   
      // this assumes the MatrixSeq is homogeneous i.e. each matrix
      // element has the same no. and sizes of dimensions 
   
      if(fill_m.is_nil())
      {
          QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);
          if(fill_v.is_nil())
              fill_v = new QdRScalar(fill);
               
          fill_m=new QdRMatrix(ms_in[0].dimdata(),fill_v->scalar());
          
      }
      
      // initialise mask
      SequenceMask msk(ms_in->size(),true); 
      
      // iterator for fill values
      RMatrix::iterator fitr=fill_m->begin();    
  
     // cout << ms_in->size() << '\t' << ts_in->size() << endl;
  
      for(int i=0;i<ms_in->size();i++)
      {
          // iterator for matrix
          RMatrix::iterator itr=ms_in[i].begin(); 
          int j=0;     
          while(itr!=ms_in[i].end() && msk[i])
          {
              const double& d=*itr++;
              msk[i] = !QuIsBadData(d, *(fitr+j++) ) ;
          }
	//  cout << i << '\r';
	  // catch bad time tags
	  if(hasTT && ts_in[i].get_time_rep() < -6.e10) msk[i] = false;
      }       
         
	// cout << endl;
      QdRMatrixSeq_var ms_out=new QdRMatrixSeq;
      ms_in->get_subsequence(*ms_out,msk);
      ms_out->copy_xrefs_from(ms_in);

      if( !ts_in.is_nil() ){
        ts_in->get_subsequence(*ts_out,msk);
        ts_out->copy_xrefs_from(ts_in);
        set_timetags(ms_out,ts_out);
      }
      return ms_out;
    }
    
    // scalar sequence case
    
   QdRScalarSeq_var ss_in=QdRScalarSeq_var::narrow(dobj);
   if(!ss_in.is_nil())
   {
     if(ss_in->empty())
          return obj_out;
          
      QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);
      if( !fill_v.is_nil())
          fill = (double) *fill_v;

      // initialise mask
      SequenceMask msk(ss_in->size(),true); 
      
      for(int i=0;i<ss_in->size();i++){
          msk[i] = !QuIsBadData(ss_in[i], fill);
	  // catch bad time tags
	  if(hasTT && ts_in[i].get_time_rep() < -6.e10) msk[i] = false;
      }
       
      QdRScalarSeq_var ss_out=new QdRScalarSeq;
      ss_in->get_subsequence(*ss_out,msk);
      ss_out->copy_xrefs_from(ss_in);
      
      if( !ts_in.is_nil() ){
        ts_in->get_subsequence(*ts_out,msk);
        ts_out->copy_xrefs_from(ts_in);
        set_timetags(ss_out,ts_out);
      }
      return ss_out;
   }
   

    QdRMatrix_var m_in=QdRMatrix_var::narrow(dobj);
    if(!m_in.is_nil())
    {
          
      // if so, do we have a single fill-value or a matrix of different fill     
      // values for each dimension
      
      QdRMatrix_var fill_m=QdRMatrix_var::narrow(fill_xref);
      
      // ...either way create matrix of fill values   
      // this assumes the MatrixSeq is homogeneous i.e. each matrix
      // element has the same no. and sizes of dimensions 
   
      if(fill_m.is_nil())
      {
          QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);
          if(fill_v.is_nil())
              fill_v = new QdRScalar(fill);
               
          fill_m=new QdRMatrix(m_in->dimdata(),fill_v->scalar());
          
      }
      
      
      // iterator for fill values
      RMatrix::iterator fitr=fill_m->begin();    
       
      // iterator for matrix
      RMatrix::iterator itr=m_in->begin(); 
      int j=0;     
      while(itr!=m_in->end() )
      {
          const double& d=*itr++;
          if( QuIsBadData(d, *(fitr+j++) ) )
             return obj_out; // empty ptr
      }
      
      // catch bad time tags
      if(hasTT && ts_in[0].get_time_rep() < -6.e10) return obj_out;

      return dobj;    // return original object
         
    }

   QdRScalar_var s_in=QdRScalar_var::narrow(dobj);
   if(!s_in.is_nil())
   {
      QdRScalar_var fill_v=QdRScalar_var::narrow(fill_xref);
      if( !fill_v.is_nil())
          fill = (double) *fill_v;

      if( QuIsBadData( *s_in, fill))
          return obj_out; // empty ptr
      
      // catch bad time tags
      if(hasTT && ts_in[0].get_time_rep() < -6.e10)  return obj_out;

      return dobj; // return original object as clean
   }

   // string and other input types not cleaned - return input object
   
   return dobj;
   
}


