#include "tmc_data.h"
#include "tmc_routines.h"
#include "time.h"


#define LEEA 0
#define HEEA 1

//set up lineage strings

const char * CL_AUX[5] = {
    "CLUSTERII",
    "CLUSTER_SP",
    "AUXILIARY",
    "SUMMARY",
    "CAUSUM" };

const char * C1_EFW[5] = {
    "CLUSTERII",
    "CLUSTER-1",
    "EFW",
    "PRIME",
    "CEFPRIME" };

const char * C2_EFW[5] = {
    "CLUSTERII",
    "CLUSTER-2",
    "EFW",
    "PRIME",
    "CEFPRIME" };

const char * C3_EFW[5] = {
    "CLUSTERII",
    "CLUSTER-3",
    "EFW",
    "PRIME",
    "CEFPRIME" };

const char * C4_EFW[5] = {
    "CLUSTERII",
    "CLUSTER-4",
    "EFW",
    "PRIME",
    "CEFPRIME" };

const char * C1_EFW_CAA[5] = {
    "CLUSTERII",
    "CLUSTER-1",
    "EFW",
    "CAA",
    "SCPOT_SP" };

const char * C2_EFW_CAA[5] = {
    "CLUSTERII",
    "CLUSTER-2",
    "EFW",
    "CAA",
    "SCPOT_SP" };

const char * C3_EFW_CAA[5] = {
    "CLUSTERII",
    "CLUSTER-3",
    "EFW",
    "CAA",
    "SCPOT_SP" };

const char * C4_EFW_CAA[5] = {
    "CLUSTERII",
    "CLUSTER-4",
    "EFW",
    "CAA",
    "SCPOT_SP" };


const char * C1_FGM[5] = {
    "CLUSTERII",
    "CLUSTER-1",
    "FGM",
    "PRIME",
    "CFPRIME" };

const char * C2_FGM[5] = {
    "CLUSTERII",
    "CLUSTER-2",
    "FGM",
    "PRIME",
    "CFPRIME" };

const char * C3_FGM[5] = {
    "CLUSTERII",
    "CLUSTER-3",
    "FGM",
    "PRIME",
    "CFPRIME" };

const char * C4_FGM[5] = {
    "CLUSTERII",
    "CLUSTER-4",
    "FGM",
    "PRIME",
    "CFPRIME" };


const char * C1_CPCMOMH[5] = {
    "CLUSTERII",
    "CLUSTER-1",
    "PEACE",
    "CORE",
    "CPCMOMH" };

const char * C2_CPCMOMH[5] = {
    "CLUSTERII",
    "CLUSTER-2",
    "PEACE",
    "CORE",
    "CPCMOMH" };

const char * C3_CPCMOMH[5] = {
    "CLUSTERII",
    "CLUSTER-3",
    "PEACE",
    "CORE",
    "CPCMOMH" };

const char * C4_CPCMOMH[5] = {
    "CLUSTERII",
    "CLUSTER-4",
    "PEACE",
    "CORE",
    "CPCMOMH" };

const char * C1_CPCMOMF[5] = {
    "CLUSTERII",
    "CLUSTER-1",
    "PEACE",
    "CORE",
    "CPCMOMF" };

const char * C2_CPCMOMF[5] = {
    "CLUSTERII",
    "CLUSTER-2",
    "PEACE",
    "CORE",
    "CPCMOMF" };

const char * C3_CPCMOMF[5] = {
    "CLUSTERII",
    "CLUSTER-3",
    "PEACE",
    "CORE",
    "CPCMOMF" };

const char * C4_CPCMOMF[5] = {
    "CLUSTERII",
    "CLUSTER-4",
    "PEACE",
    "CORE",
    "CPCMOMF" };

const char * C1_CPCDM[5] = {
    "CLUSTERII",
    "CLUSTER-1",
    "PEACE",
    "CORE",
    "CPCDM" };

const char * C2_CPCDM[5] = {
    "CLUSTERII",
    "CLUSTER-2",
    "PEACE",
    "CORE",
    "CPCDM" };

const char * C3_CPCDM[5] = {
    "CLUSTERII",
    "CLUSTER-3",
    "PEACE",
    "CORE",
    "CPCDM" };

const char * C4_CPCDM[5] = {
    "CLUSTERII",
    "CLUSTER-4",
    "PEACE",
    "CORE",
    "CPCDM" };

    //assign correct set of lineages
    //based on spacecraft number
    void tmc_data::set_lineages()
    {
	      
	      
	      ASSIGN_LINEAGE(&aux, CL_AUX);
	      
	      switch( spacecraft )
	      {
	      		case 1:      
				    ASSIGN_LINEAGE(&cpcmomh, C1_CPCMOMH);
   				    ASSIGN_LINEAGE(&cpcmomf, C1_CPCMOMF);
				    ASSIGN_LINEAGE(&fgm, C1_FGM);
				    ASSIGN_LINEAGE(&efw, C1_EFW);
				    ASSIGN_LINEAGE(&efw_caa, C1_EFW_CAA);
				    ASSIGN_LINEAGE(&cpcdm, C1_CPCDM);
				    break;

			
			case 2:
				    ASSIGN_LINEAGE(&cpcmomh, C2_CPCMOMH);
   				    ASSIGN_LINEAGE(&cpcmomf, C2_CPCMOMF);
				    ASSIGN_LINEAGE(&fgm, C2_FGM);
				    ASSIGN_LINEAGE(&efw, C2_EFW);
				    ASSIGN_LINEAGE(&efw_caa, C2_EFW_CAA);
				    ASSIGN_LINEAGE(&cpcdm, C2_CPCDM);
				    break;
			
			
			case 3:
	   			    ASSIGN_LINEAGE(&cpcmomh, C3_CPCMOMH);
   				    ASSIGN_LINEAGE(&cpcmomf, C3_CPCMOMF);
				    ASSIGN_LINEAGE(&fgm, C3_FGM);
				    ASSIGN_LINEAGE(&efw, C3_EFW);
				    ASSIGN_LINEAGE(&efw_caa, C3_EFW_CAA);
				    ASSIGN_LINEAGE(&cpcdm, C3_CPCDM);
				    break;
			
			case 4:
			 	    ASSIGN_LINEAGE(&cpcmomh, C4_CPCMOMH);
   				    ASSIGN_LINEAGE(&cpcmomf, C4_CPCMOMF);
				    ASSIGN_LINEAGE(&fgm, C4_FGM);
				    ASSIGN_LINEAGE(&efw, C4_EFW);
				    ASSIGN_LINEAGE(&efw_caa, C4_EFW_CAA);
				    ASSIGN_LINEAGE(&cpcdm, C4_CPCDM);
				    break;
	      
	      
	         default: break;
	      }
		 
      }
		 
     
      //use the idfs routines to populate all the raw data
      //positions -- from start to end. report if data cannot be
      //found, and take correct action if EFW or FGM data cannot
      //be found.
      TMC_ERROR tmc_data::get_data(IDFS_TIME start, IDFS_TIME end, bool constant_phi, double phi)
      {
	       IDFS_TIME aux_start, aux_end;

        aux_start.year = start.year;
        aux_start.doy = start.doy;
        aux_start.sec = start.sec;
        aux_start.nano = start.nano;

        aux_end.year = end.year;
        aux_end.doy = end.doy;
        aux_end.sec = end.sec;
        aux_end.nano = end.nano;

        if((aux_end.sec-aux_start.sec < 150))
        {
            //cout << aux_end.sec-aux_start.sec << endl;
            aux_end.sec+=150;
            aux_start.sec-=150;

            if(aux_end.sec >= 86400)
            {aux_end.sec = 0+150; aux_end.doy++;}


            if(aux_start.sec < 0)
            {aux_start.sec = 86400-150; aux_end.doy--;}
           
          }
        
      if(leea_overlap == TRUE)
	    {
	    overlap_N = ret_obj(start, end, cpcmomh, LDENSITY, SENSOR, true);  
	    if(overlap_N.is_nil())
	    {DebugCout("Couldn't get data for LEEA density"); return TMC_IDFS_RET_ERR;}
    
	    overlap_Vx = ret_obj(start, end, cpcmomh, LVELOCITYX, SENSOR, true);
            if(overlap_Vx.is_nil())
	    {DebugCout("Couldn't get data for LEEA vx"); return TMC_IDFS_RET_ERR;}

	    overlap_Vy = ret_obj(start, end, cpcmomh, LVELOCITYY, SENSOR, true);
	    if(overlap_Vy.is_nil())
	    {DebugCout("Couldn't get data for LEEA vy"); return TMC_IDFS_RET_ERR;}
	    
	    overlap_Vz = ret_obj(start, end, cpcmomh, LVELOCITYZ, SENSOR, true);
	    if(overlap_Vz.is_nil())
	    {DebugCout("Couldn't get data for LEEA vz"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pxx = ret_obj(start, end, cpcmomh, LPXX, SENSOR, true);
	    if(overlap_Pxx.is_nil())
	    {DebugCout("Couldn't get data for LEEA pxx"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pyy = ret_obj(start, end, cpcmomh, LPYY, SENSOR, true);
	    if(overlap_Pyy.is_nil())
	    {DebugCout("Couldn't get data for LEEA pyy"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pzz = ret_obj(start, end, cpcmomh, LPZZ, SENSOR, true);
	    if(overlap_Pzz.is_nil())
	    {DebugCout("Couldn't get data for LEEA pzz"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pxy = ret_obj(start, end, cpcmomh, LPXY, SENSOR, true);
	    if(overlap_Pxy.is_nil())
	    {DebugCout("Couldn't get data for LEEA pxy"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pxz = ret_obj(start, end, cpcmomh, LPXZ, SENSOR, true);
	    if(overlap_Pxz.is_nil())
	    {DebugCout("Couldn't get data for LEEA pxz"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pyz = ret_obj(start, end, cpcmomh, LPYZ, SENSOR, true);
	    if(overlap_Pyz.is_nil())
	    {DebugCout("Couldn't get data for LEEA pyz"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_spinmode = ret_obj(start, end, cpcmomh, SPIN_MODE, MODE, true);
	    if(overlap_spinmode.is_nil())
	    {DebugCout("Couldn't get data for spin mode"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_energymode = ret_obj(start, end, cpcmomh, LMODE, MODE, true);
	    if(overlap_energymode.is_nil())
	    {DebugCout("Couldn't get data for LEEA mode"); return TMC_IDFS_RET_ERR;}

	    overlap_EL = ret_obj(start, end, cpcmomh, EEND, SENSOR, true);
	    if(overlap_EL.is_nil())
	    {DebugCout("Couldn't get data for LEEA EL"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_EU = ret_obj(start, end, cpcmomh, EBEG, SENSOR, true);
	    if(overlap_EU.is_nil())
	    {DebugCout("Couldn't get data for LEEA EU"); return TMC_IDFS_RET_ERR;}

	}
	else
	{
	    overlap_N = ret_obj(start, end, cpcmomh, HDENSITY, SENSOR, true);
	    if(overlap_N.is_nil())
	    {DebugCout("Couldn't get data for HEEA density"); return TMC_IDFS_RET_ERR;}

	    overlap_Vx = ret_obj(start, end, cpcmomh, HVELOCITYX, SENSOR, true);
            if(overlap_Vx.is_nil())
	    {DebugCout("Couldn't get data for HEEA vx"); return TMC_IDFS_RET_ERR;}

	    overlap_Vy = ret_obj(start, end, cpcmomh, HVELOCITYY, SENSOR, true);
	    if(overlap_Vy.is_nil())
	    {DebugCout("Couldn't get data for HEEA vy"); return TMC_IDFS_RET_ERR;}

	    overlap_Vz = ret_obj(start, end, cpcmomh, HVELOCITYZ, SENSOR, true);
	    if(overlap_Vz.is_nil())
	    {DebugCout("Couldn't get data for HEEA vz"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pxx = ret_obj(start, end, cpcmomh, HPXX, SENSOR, true);
	    if(overlap_Pxx.is_nil())
	    {DebugCout("Couldn't get data for HEEA pxx"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pyy = ret_obj(start, end, cpcmomh, HPYY, SENSOR, true);
	    if(overlap_Pyy.is_nil())
	    {DebugCout("Couldn't get data for HEEA pyy"); return TMC_IDFS_RET_ERR;}

	    overlap_Pzz = ret_obj(start, end, cpcmomh, HPZZ, SENSOR, true);
	    if(overlap_Pzz.is_nil())
	    {DebugCout("Couldn't get data for HEEA pzz"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pxy = ret_obj(start, end, cpcmomh, HPXY, SENSOR, true);
	    if(overlap_Pxy.is_nil())
	    {DebugCout("Couldn't get data for HEEA pxy"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pxz = ret_obj(start, end, cpcmomh, HPXZ, SENSOR, true);
	    if(overlap_Pxz.is_nil())
	    {DebugCout("Couldn't get data for HEEA pxz"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_Pyz = ret_obj(start, end, cpcmomh, HPYZ, SENSOR, true);
	    if(overlap_Pyz.is_nil())
	    {DebugCout("Couldn't get data for HEEA pyz"); return TMC_IDFS_RET_ERR;}

	    overlap_spinmode = ret_obj(start, end, cpcmomh, SPIN_MODE, MODE, true);
	    if(overlap_spinmode.is_nil())
	    {DebugCout("Couldn't get data for spin mode"); return TMC_IDFS_RET_ERR;}

	    overlap_energymode = ret_obj(start, end, cpcmomh, HMODE, MODE, true);
	    if(overlap_energymode.is_nil())
	    {DebugCout("Couldn't get data for HEEA mode"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_EL = ret_obj(start, end, cpcmomh, EEND, SENSOR, true);
	    if(overlap_EL.is_nil())
	    {DebugCout("Couldn't get data for HEEA EH"); return TMC_IDFS_RET_ERR;}
	   
	    overlap_EU = ret_obj(start, end, cpcmomh, EBEG, SENSOR, true);
	    if(overlap_EU.is_nil())
	    {DebugCout("Couldn't get data for HEEA EU"); return TMC_IDFS_RET_ERR;}
	}

	QdRScalarSeq_var Bx = new QdRScalarSeq(overlap_N->sequence_size());
	QdRScalarSeq_var By = new QdRScalarSeq(overlap_N->sequence_size());
	QdRScalarSeq_var Bz = new QdRScalarSeq(overlap_N->sequence_size());

	if(top==TRUE)
	{
	    top_N = ret_obj(start, end, cpcmomf, TDENSITY, SENSOR, true);
	    if(top_N.is_nil())
	    {DebugCout("Couldn't get data for T density"); return TMC_IDFS_RET_ERR;}

	    top_Vx = ret_obj(start, end, cpcmomf, TVELOCITYX, SENSOR, true);
	    if(top_Vx.is_nil())
	    {DebugCout("Couldn't get data for T vx"); return TMC_IDFS_RET_ERR;}

	    top_Vy = ret_obj(start, end, cpcmomf, TVELOCITYY, SENSOR, true);
	    if(top_Vy.is_nil())
	    {DebugCout("Couldn't get data for T vy"); return TMC_IDFS_RET_ERR;}

	    top_Vz = ret_obj(start, end, cpcmomf, TVELOCITYZ, SENSOR, true);
	    if(top_Vz.is_nil())
	    {DebugCout("Couldn't get data for T vz"); return TMC_IDFS_RET_ERR;}

	    top_Pxx = ret_obj(start, end, cpcmomf, TPXX, SENSOR, true);
	    if(top_Pxx.is_nil())
	    {DebugCout("Couldn't get data for T pxx"); return TMC_IDFS_RET_ERR;}

	    top_Pyy = ret_obj(start, end, cpcmomf, TPYY, SENSOR, true);
	    if(top_Pyy.is_nil())
	    {DebugCout("Couldn't get data for T pyy"); return TMC_IDFS_RET_ERR;}

	    top_Pzz = ret_obj(start, end, cpcmomf, TPZZ, SENSOR, true);
	    if(top_Pzz.is_nil())
	    {DebugCout("Couldn't get data for T pzz"); return TMC_IDFS_RET_ERR;}

	    top_Pxy = ret_obj(start, end, cpcmomf, TPXY, SENSOR, true);
	    if(top_Pxy.is_nil())
	    {DebugCout("Couldn't get data for T pxy"); return TMC_IDFS_RET_ERR;}

	    top_Pxz = ret_obj(start, end, cpcmomf, TPXZ, SENSOR, true);
	    if(top_Pxz.is_nil())
	    {DebugCout("Couldn't get data for T pxz"); return TMC_IDFS_RET_ERR;}

	    top_Pyz = ret_obj(start, end, cpcmomf, TPYZ, SENSOR, true);
	    if(top_Pyz.is_nil())
	    {DebugCout("Couldn't get data for T pyz"); return TMC_IDFS_RET_ERR;}

	    top_EL = ret_obj(start, end, cpcmomf, TEEND, SENSOR, true);
	    if(top_EL.is_nil())
	    {DebugCout("Couldn't get data for T EL"); return TMC_IDFS_RET_ERR;}

	    top_EU = ret_obj(start, end, cpcmomf, TEBEG, SENSOR, true);
	    if(top_EU.is_nil())
	    {DebugCout("Couldn't get data for T EU"); return TMC_IDFS_RET_ERR;}

	    top_energymode = ret_obj(start, end, cpcmomf, TMODE, MODE, true);
	    if(top_energymode.is_nil())
	    {DebugCout("Couldn't get data for T mode"); return TMC_IDFS_RET_ERR;}

	    top_source = ret_obj(start, end, cpcmomf, TSOURCE, MODE, true);
	    if(top_source.is_nil())
	    {DebugCout("Couldn't get data for T source"); return TMC_IDFS_RET_ERR;}

	}

	if(bottom == TRUE)
	{

	    bottom_N = ret_obj(start, end, cpcmomf, BDENSITY, SENSOR, true);
	    if(bottom_N.is_nil())
	    {DebugCout("Couldn't get data for B density"); return TMC_IDFS_RET_ERR;}

	    bottom_Vx = ret_obj(start, end, cpcmomf, BVELOCITYX, SENSOR, true);
	    if(bottom_Vx.is_nil())
	    {DebugCout("Couldn't get data for B vx"); return TMC_IDFS_RET_ERR;}

	    bottom_Vy = ret_obj(start, end, cpcmomf, BVELOCITYY, SENSOR, true);
	    if(bottom_Vy.is_nil())
	    {DebugCout("Couldn't get data for B vy"); return TMC_IDFS_RET_ERR;}

	    bottom_Vz = ret_obj(start, end, cpcmomf, BVELOCITYZ, SENSOR, true);
	    if(bottom_Vz.is_nil())
	    {DebugCout("Couldn't get data for B vz"); return TMC_IDFS_RET_ERR;}

	    bottom_Pxx = ret_obj(start, end, cpcmomf, BPXX, SENSOR, true);
	    if(bottom_Pxx.is_nil())
	    {DebugCout("Couldn't get data for B pxx"); return TMC_IDFS_RET_ERR;}

	    bottom_Pyy = ret_obj(start, end, cpcmomf, BPYY, SENSOR, true);
	    if(bottom_Pyy.is_nil())
	    {DebugCout("Couldn't get data for B pyy"); return TMC_IDFS_RET_ERR;}

	    bottom_Pzz = ret_obj(start, end, cpcmomf, BPZZ, SENSOR, true);
	    if(bottom_Pzz.is_nil())
	    {DebugCout("Couldn't get data for B pzz"); return TMC_IDFS_RET_ERR;}

	    bottom_Pxy = ret_obj(start, end, cpcmomf, BPXY, SENSOR, true);
	    if(bottom_Pxy.is_nil())
	    {DebugCout("Couldn't get data for B pxy"); return TMC_IDFS_RET_ERR;}

	    bottom_Pxz = ret_obj(start, end, cpcmomf, BPXZ, SENSOR, true);
	    if(bottom_Pxz.is_nil())
	    {DebugCout("Couldn't get data for B pxz"); return TMC_IDFS_RET_ERR;}

	    bottom_Pyz = ret_obj(start, end, cpcmomf, BPYZ, SENSOR, true);
	    if(bottom_Pyz.is_nil())
	    {DebugCout("Couldn't get data for B pyz"); return TMC_IDFS_RET_ERR;}

	    bottom_EL = ret_obj(start, end, cpcmomf, BEEND, SENSOR, true);
	    if(bottom_EL.is_nil())
	    {DebugCout("Couldn't get data for B EL"); return TMC_IDFS_RET_ERR;}

	    bottom_EU = ret_obj(start, end, cpcmomf, BEBEG, SENSOR, true);
	    if(bottom_EU.is_nil())
	    {DebugCout("Couldn't get data for B EU"); return TMC_IDFS_RET_ERR;}

	    bottom_energymode = ret_obj(start, end, cpcmomf, BMODE, MODE, true);
	    if(bottom_energymode.is_nil())
	    {DebugCout("Couldn't get data for B mode"); return TMC_IDFS_RET_ERR;}

	    bottom_source = ret_obj(start, end, cpcmomf, BSOURCE, MODE, true);
	    if(bottom_source.is_nil())
	    {DebugCout("Couldn't get data for B source"); return TMC_IDFS_RET_ERR;}

	}

	if(constant_phi==FALSE)
	{
		/* efw_U_probe_sc = ret_obj(start, end, efw_caa, EFW_CAA_SCPOT, SENSOR, true); 
                   if(efw_U_probe_sc is good, manipulate it to get from SCPOT to U_probe (see
                   below). ELSE go try to pull in the EFW PP U_probe and continue, and
                   fall through the rest of the else's below*/

		/* When the caa holding of L3_P is in better shape, and held in idfs, it
                 * will be better to use that. Note that elsewhere in qtmc, the s/c potential
                 * is derived from U_probe as: Phi_sc = 1.0-U_probe. So here after
                 * fetching SCPOT we need to modify it so that efw_U_probe_sc will carry
                 * something that, when subtracted from one, gives Phi_sc. This is slightly
                 * klunky, but it avoids changing all the joining code, filling of
                 * tmc structure, etc., which is all done with U_probe. sjs July 2007 */

		efw_U_probe_sc = ret_obj(start, end, efw, U_PROBE_SC, SENSOR, true);
		if(efw_U_probe_sc.is_nil() || (efw_U_probe_sc->sequence_size()<2))
		{
			 cout << "NOTE!!! Couldn't get data for efw_U_probe_sc, using user-supplied value = ";
			 cout << phi << " V" << endl;
			 efw_U_probe_sc = new QdRScalarSeq(overlap_N->sequence_size());
			 for(int z=0; z<overlap_N->sequence_size(); z++)
			 	efw_U_probe_sc[z] = 1 - phi;

				set_timetags( efw_U_probe_sc, get_timetags(overlap_N) );

		}

	}
	else
	{

			efw_U_probe_sc = new QdRScalarSeq(overlap_N->sequence_size());
			//cout << EEA_n->sequence_size() << endl;
			for(int z=0; z<overlap_N->sequence_size(); z++)
			  {
			  	 efw_U_probe_sc[z] = 1 - phi;
				// cout << U_probe_sc[z] << endl;
			  }
	set_timetags( efw_U_probe_sc, get_timetags(overlap_N) );


	}

	switch(spacecraft)
	{

	case 1:

	aux_latitude = ret_obj(aux_start, aux_end, aux, SC1_LATITUDE, SENSOR, true);
	if(aux_latitude.is_nil())
	{DebugCout("Couldn't get data for latitude"); return TMC_IDFS_RET_ERR;}

	aux_longitude = ret_obj(aux_start, aux_end, aux, SC1_LONGITUDE, SENSOR, 
true); 	if(aux_longitude.is_nil())
	{DebugCout("Couldn't get data for longitude"); return TMC_IDFS_RET_ERR;}

	break;

	case 2:

	aux_latitude = ret_obj(aux_start, aux_end, aux, SC3_LATITUDE, SENSOR, true);
	if(aux_latitude.is_nil())
	{DebugCout("Couldn't get data for latitude"); return TMC_IDFS_RET_ERR;}

	aux_longitude = ret_obj(aux_start, aux_end, aux, SC3_LONGITUDE, SENSOR, 
true); 	if(aux_longitude.is_nil())
	{DebugCout("Couldn't get data for longitude"); return TMC_IDFS_RET_ERR;}

	break;

	case 3:

	aux_latitude = ret_obj(aux_start, aux_end, aux, SC3_LATITUDE, SENSOR, true);
	if(aux_latitude.is_nil())
	{DebugCout("Couldn't get data for latitude"); return TMC_IDFS_RET_ERR;}

	aux_longitude = ret_obj(aux_start, aux_end, aux, SC3_LONGITUDE, SENSOR, 
true); 	if(aux_longitude.is_nil())
	{DebugCout("Couldn't get data for longitude"); return TMC_IDFS_RET_ERR;}

	break;

	case 4:

	aux_latitude = ret_obj(aux_start, aux_end, aux, SC4_LATITUDE, SENSOR, true);
	if(aux_latitude.is_nil())
	{DebugCout("Couldn't get data for latitude"); return TMC_IDFS_RET_ERR;}

	aux_longitude = ret_obj(aux_start, aux_end, aux, SC4_LONGITUDE, SENSOR, 
true); 	if(aux_longitude.is_nil())
	{DebugCout("Couldn't get data for longitude"); return TMC_IDFS_RET_ERR;}

	break;

	default: break;
	}

        cpcdm_spos = ret_obj(start, end, cpcdm, SPOS, SENSOR, true);
        if(cpcdm_spos.is_nil()){
           DebugCout("Couldn't get SPOS"); return TMC_IDFS_RET_ERR;
        }
        // cout << "\n SPOS reg_obj success\n";

	Bx = ret_obj(start, end, fgm, BX_GSE, SENSOR, false);
	By = ret_obj(start, end, fgm, BY_GSE, SENSOR, false);
	Bz = ret_obj(start, end, fgm, BZ_GSE, SENSOR, false);

	if(Bx.is_nil() || By.is_nil() || Bz.is_nil() || Bx->sequence_size()<2 || By->sequence_size()<2 || Bz ->sequence_size()<2)
	{
          cout << "\nNOTE!!! Couldn't get FGM B-vector, so will not estimate anisotropy\n";
          fgm_B_vec = new QdRMatrixSeq(overlap_N->sequence_size(), Dim(3));

	  for(int w=0; w<overlap_N->sequence_size(); w++)
	  {
	        fgm_B_vec[w][0] = 0.0; fgm_B_vec[w][1] = 0.0; fgm_B_vec[w][2] = 0.0;
    }

	  set_timetags( fgm_B_vec, get_timetags(overlap_N) );


	}
	else
	   fgm_B_vec = components_to_vector(Bx, By, Bz);


	if(fgm_B_vec.is_nil())
	    {DebugCout("Couldn't get data for fgm_B_vec"); return TMC_IDFS_RET_ERR;}

	  return TMC_OK;

	}
        //get calibration sequence for leea
	      void tmc_data::get_leea_calibration_info(string name)
	      {

	      	if(bottom==TRUE || leea_overlap==TRUE)
			leea_gain_correction = read_n_rtn_cal_file(name);
	       	else leea_gain_correction = NULL;

	      }

       //same for heea 
       void tmc_data::get_heea_calibration_info(string name)
	      {
	        if(top==TRUE || heea_overlap==TRUE)
			heea_gain_correction = read_n_rtn_cal_file(name);
	        else heea_gain_correction = NULL;
              }
	
      //take L1L2 or H1H2 sequence, separate first and
      //second half of spin, shift to centre tagging and sum
      //to get L1+L2 or H1+H2 sequence at spin resolution
	    TMC_ERROR tmc_data::separate_and_sum_overlap()
	    {

	         int i;
		       int size = overlap_N->sequence_size();
	         SequenceMask mask(size, TRUE);
    
     //setup mask
     for(i=0; i<size; i++)
   		 {
	  		if((int)overlap_spinmode[i]==0) 
			    mask[i]=TRUE;
		  	else
			    mask[i]=FALSE;
   		 }

	      
		int first_half  = 0;
		int second_half = 0;
		int total = 0;

    		total = size;

    		for(i=0; i<total; i++)
	      {
		      if(mask[i]==TRUE)
		       first_half++;
	
		      if(mask[i]==FALSE)
		       second_half++;
    		}
    
	  QdObjectSeq_var a =  new QdObjectSeq( 13 );
		QdObjectSeq_var b =  new QdObjectSeq( 13 ); 
	        
		QdObjectSeq_var ca =  new QdObjectSeq( 13 );
		QdObjectSeq_var cb =  new QdObjectSeq( 13 ); 
	        

		QdObjectSeq_var j = new QdObjectSeq( 26 );

    
    
    		for(i=0; i<13; i++)
   	        {
		  (a)[i] = new QdRScalarSeq( first_half ); 
		  (b)[i] = new QdRScalarSeq( second_half );
		}

    	mask_series(overlap_N,   a[0], b[0], mask);
    	mask_series(overlap_Vx,  a[1], b[1], mask);
	mask_series(overlap_Vy,  a[2], b[2], mask);
	mask_series(overlap_Vz,  a[3], b[3], mask);
	mask_series(overlap_Pxx, a[4], b[4], mask);
	mask_series(overlap_Pxy, a[5], b[5], mask);
	mask_series(overlap_Pxz, a[6], b[6], mask);
	mask_series(overlap_Pyy, a[7], b[7], mask);					
  mask_series(overlap_Pyz, a[8], b[8], mask);
	mask_series(overlap_Pzz, a[9], b[9], mask);
	mask_series(overlap_energymode, a[10], b[10], mask);
	mask_series(overlap_EL, a[11], b[11], mask);
	mask_series(overlap_EU, a[12], b[12], mask);
	
      for(i=0; i<13; i++)
      {
      	if( shift_time_series( (QdObject_var)(a[i]) ) !=TMC_OK )
	    return TMC_ERR;

	    //    cout << a[i]->sequence_size() << '\t' << b[i]->sequence_size() << endl;
      }


     try
     {
     	 for(i=0;i<13;i++)
     	 {
      		ca[i] = check_for_non_conformity(a[i]);
      		cb[i] = check_for_non_conformity(b[i]);
      	 }
      
     }
     catch(Exception &e)
     {
          return TMC_ERR;
     }
    
    QdObjectSeq_var allclean = new QdObjectSeq(26);

    for(i=0;i<13;i++)
    allclean[i] = ca[i];
    
    for(i=0;i<13;i++)
    allclean[i+13] = cb[i];


    if ( join_a_to_b(allclean, cb[0], j, 26) != TMC_OK )
	return TMC_ERR;  
    
  //  cout << "done join 1" << endl;

  //  if ( join_a_to_b(cb, cb[0], bj, 13) != TMC_OK )
  //	return TMC_ERR;
    
    	       total = j[0]->sequence_size();
    	     //  cout << "total = " << total << endl;
	
	       ss_overlap_N = new QdRScalarSeq( total );
	       ss_overlap_Vx = new QdRScalarSeq( total );
	       ss_overlap_Vy = new QdRScalarSeq( total );
	       ss_overlap_Vz = new QdRScalarSeq( total );
	       ss_overlap_Pxx = new QdRScalarSeq( total );
	       ss_overlap_Pxy = new QdRScalarSeq( total );
	       ss_overlap_Pxz = new QdRScalarSeq( total );
	       ss_overlap_Pyy = new QdRScalarSeq( total );
	       ss_overlap_Pyz = new QdRScalarSeq( total );
	       ss_overlap_Pzz = new QdRScalarSeq( total );
	       ss_overlap_EL = new QdRScalarSeq( total );
	       ss_overlap_EU = new QdRScalarSeq( total );
	       ss_overlap_energymode = new QdRScalarSeq( total );
    
	    
	     add(total,  
		QdRScalarSeq_var::narrow(j[0]),
		QdRScalarSeq_var::narrow(j[13]),
		ss_overlap_N,
		0);

	     add(total,  
		QdRScalarSeq_var::narrow(j[1]),
		QdRScalarSeq_var::narrow(j[1+13]),
		ss_overlap_Vx,
		0);
		
   	     add(total,
		QdRScalarSeq_var::narrow(j[2]),
		QdRScalarSeq_var::narrow(j[2+13]), 
		ss_overlap_Vy,
		0);
		
	     add(total,
		QdRScalarSeq_var::narrow(j[3]), 
		QdRScalarSeq_var::narrow(j[3+13]),
		ss_overlap_Vz,
		0);

	     add(total,  
		QdRScalarSeq_var::narrow(j[4]), 
		QdRScalarSeq_var::narrow(j[4+13]),
		ss_overlap_Pxx,
		0);

	     add(total,
		QdRScalarSeq_var::narrow(j[5]),
		QdRScalarSeq_var::narrow(j[5+13]), 
		ss_overlap_Pxy,
		0);
		
            add(total,
		QdRScalarSeq_var::narrow(j[6]),
		QdRScalarSeq_var::narrow(j[6+13]),
		ss_overlap_Pxz,
		0);
		
	    add(total,
		QdRScalarSeq_var::narrow(j[7]),
		QdRScalarSeq_var::narrow(j[7+13]), 
		ss_overlap_Pyy,
		0);

	    add(total,  
		QdRScalarSeq_var::narrow(j[8]), 
		QdRScalarSeq_var::narrow(j[8+13]),
		ss_overlap_Pyz,
		0);

	    add(total,  
		QdRScalarSeq_var::narrow(j[9]),
		QdRScalarSeq_var::narrow(j[9+13]),
		ss_overlap_Pzz,
		0);
	    
	    add(total,  
		QdRScalarSeq_var::narrow(j[10]),
	        QdRScalarSeq_var::narrow(j[10+13]),
		ss_overlap_energymode,
		1);

	    add(total,
		QdRScalarSeq_var::narrow(j[11]), 
	        QdRScalarSeq_var::narrow(j[11+13]), 
		ss_overlap_EL,
		1);

	    add(total,
		QdRScalarSeq_var::narrow(j[12]),
	        QdRScalarSeq_var::narrow(j[12+13]), 
		ss_overlap_EU,
		1);

           // cout << "ss_overlap_N = " << ss_overlap_N->sequence_size() << endl;
	    //Now set the time tags, which we define as coming from the second half of the spin
    	    QdTimeSeq_var tt_in  = new QdTimeSeq();
	    tt_in = get_timetags(j[0]);


	set_timetags(ss_overlap_N, tt_in);
	set_timetags(ss_overlap_Vx, tt_in);
	set_timetags(ss_overlap_Vy, tt_in);
	set_timetags(ss_overlap_Vz, tt_in);
	set_timetags(ss_overlap_Pxx, tt_in);
	set_timetags(ss_overlap_Pxy, tt_in);
	set_timetags(ss_overlap_Pxz, tt_in);
	set_timetags(ss_overlap_Pyy, tt_in);
	set_timetags(ss_overlap_Pyz, tt_in);
	set_timetags(ss_overlap_Pzz, tt_in);
	set_timetags(ss_overlap_energymode, tt_in);
	set_timetags(ss_overlap_EL, tt_in);
	set_timetags(ss_overlap_EU, tt_in);

	 //cout << "ss_overlap_N = " << ss_overlap_N->sequence_size() << endl;
	
	return TMC_OK;
}

//method for adding two sequences
void tmc_data::add(int nSize, QdRScalarSeq_var qdObj1, QdRScalarSeq_var qdObj2, QdRScalarSeq_var qdOut, int add_or_pass)
{
    int nIndex = 0;
    QdObject_var fillobj = new QdRScalar(-3.4e38);
    double dVal;
    for ( nIndex = 0; nIndex < nSize; nIndex++ )
    {
	if(add_or_pass==0)
	{
	    double     dVal1 = (double)qdObj1[ nIndex ];
	    double     dVal2 = (double)qdObj2[ nIndex ];
	    dVal           = dVal1 + dVal2;
	   // cout << dVal1 << '\t' << dVal2;
	    qdOut[ nIndex ] = dVal;
	   // cout << '\t' << (double)qdOut[nIndex] << endl;

	}
	else
	{
	  qdOut[nIndex] = (double)qdObj2[ nIndex ];	
	}
    }

    try{ if(qdOut->xref_exists("FILLVAL"))
	   qdOut->delete_xref("FILLVAL"); 
	 
	qdOut->set_xref("FILLVAL", fillobj);
       } catch(Exception &e){return;}
}
	      

//method for extracting every other point in a sequence to form two sequences
void tmc_data::mask_series(QdRScalarSeq_var in, QdObject_var A, QdObject_var B, SequenceMask mask)
{
	    extract_half_spin(in, mask,  QdRScalarSeq_var::narrow(A));
	    extract_half_spin(in, !mask, QdRScalarSeq_var::narrow(B));
}

	      
//method for above
void  tmc_data::extract_half_spin( QdRScalarSeq_var in, SequenceMask mask, QdRScalarSeq_var out)
{
	    QdTimeSeq_var tt_out = new QdTimeSeq();
	    QdTimeSeq_var tt_in  = new QdTimeSeq();
    
	    tt_in = get_timetags(in);
	    tt_in->get_subsequence(*tt_out, mask);
    
	    in->get_subsequence(*out, mask);
	    set_timetags(out, tt_out);
}


//method to join sequence a to sequence b and output aj
TMC_ERROR tmc_data::join_a_to_b(QdObjectSeq_var a, QdObject_var b, QdObjectSeq_var aj, int count)
{
    int triage_cnt = 0;
    int i = 0;

 //   for(i=0;i<count;i++)
 //     cout << a[i]->sequence_size() << '\t' << b->sequence_size() << endl;


    //Time line of the time series to be joined
    QdObjectSeq_var ttIn = new QdObjectSeq(count);
    if ( ttIn.is_nil() )
    	return TMC_ERR;

    // Set of time series to be joined
    QdObjectSeq_var toJoin = new QdObjectSeq(count);
    if ( toJoin.is_nil() )
    	return TMC_ERR;

    //Join method
    QdObjectSeq_var joinMethod = new QdObjectSeq(count);
    if ( joinMethod.is_nil() )
    	return TMC_ERR;

    //Gap value if the time series is not regular
    QdObjectSeq_var gapValue = new QdObjectSeq(count);
    if ( gapValue.is_nil() )
    	return TMC_ERR;

    // Gap handler if the time series is not regular
    QdObjectSeq_var gapHandler = new QdObjectSeq(count);
    if ( gapHandler.is_nil() )
    	return TMC_ERR;

    //Tags to join onto
    QdTimeSeq_var join_pts = get_timetags( (b) );
    if (join_pts.is_nil())
    	return TMC_ERR;

    //A copy
    QdTimeSeq_var ttJoined = new QdTimeSeq( *join_pts );

    double dSpacing  = 0.;

    QdTimeSeq_var toJoinTags;  // for the sake of initialization


    //Remove gaps
    triage_cnt = count;

    for(i=0; i<count; i++)
    {

	toJoin[i]     = ( (a)[i] );
	ttIn[i]       = get_timetags( (a)[i] );
	aj[i]        = toJoin[i]->create();
	toJoinTags    = get_timetags( (a)[i] );
	joinMethod[i] = new QdString(LINEAR);
	
	if ( !QuIsTSRegular( toJoinTags , &dSpacing ) );
	      //  cout << dSpacing << endl;
	
	gapValue[i]   = new QdTime(Seconds(1.1*dSpacing));
    }
    //Hand the results to the multijoin function
    try 
    {
	multijoin( ttIn, toJoin, ttJoined, aj, joinMethod, gapValue, gapHandler, triage_cnt);
    }
    catch( Exception &e )
    {
   	e.print_msg();
   	return TMC_ERR;
    }

    for(i=0; i<count; i++)
    {
	
	(aj[i])->copy_xrefs_from( toJoin[i] );
	set_timetags( aj[i], ttJoined );
    }

    return TMC_OK;
}
     
	      
//prepare output sequences with correct time tags, etc.
TMC_ERROR tmc_data::set_up_outputs()
{

    int nSize = j_ss_overlap_N->sequence_size();

    QdTimeSeq_var tLine = get_timetags( j_ss_overlap_N );
    if ( tLine.is_nil() )
	return TMC_IS_NIL_ERR;     

    Ne = new QdRScalarSeq( nSize );
    if ( Ne.is_nil() )
	return TMC_IS_NIL_ERR;
    
    Ve = new QdRMatrixSeq( nSize, Dim(3) );
    if ( Ve.is_nil() )
	return TMC_IS_NIL_ERR;

    VT = new QdRScalarSeq( nSize );
    if ( VT.is_nil() )
	return TMC_IS_NIL_ERR;
    
    Te = new QdRScalarSeq( nSize );
    if ( Te.is_nil() )
	return TMC_IS_NIL_ERR;
    
    Te_perp  = new QdRScalarSeq( nSize );
    if ( Te_perp.is_nil() )
	return TMC_IS_NIL_ERR;

    Te_para = new QdRScalarSeq( nSize );
    if ( Te_para.is_nil() )
	return TMC_IS_NIL_ERR;
    
    rT = new QdRScalarSeq( nSize );
    if ( rT.is_nil() )
	return TMC_IS_NIL_ERR;

    rN = new QdRScalarSeq( nSize );
    if ( rN.is_nil() )
	return TMC_IS_NIL_ERR;
    
    rV = new QdRScalarSeq( nSize );
    if ( rV.is_nil() )
	return TMC_IS_NIL_ERR;
    
    Ie = new QdRScalarSeq( nSize );
    if ( Ie.is_nil() )
	return TMC_IS_NIL_ERR;

    err = new QdRScalarSeq( nSize );
    if ( err.is_nil() )
	return TMC_IS_NIL_ERR;
    
    p_off_on = new QdRScalarSeq( nSize );
    if ( p_off_on.is_nil() )
	return TMC_IS_NIL_ERR;
    
    p_perp_ratio = new QdRScalarSeq( nSize );
    if ( p_perp_ratio.is_nil() )
	return TMC_IS_NIL_ERR;


    Ne->copy_xrefs_from( overlap_N );
    Ve->copy_xrefs_from( fgm_B_vec );
    VT->copy_xrefs_from( overlap_N );
    Te->copy_xrefs_from( overlap_N );
    Te_perp->copy_xrefs_from( overlap_N );
    Te_para->copy_xrefs_from( overlap_N );
    err->copy_xrefs_from( overlap_N );
    
    // Set time tags
    set_timetags( Ne, tLine );
    set_timetags( Ve, tLine );
    set_timetags( VT, tLine );
    set_timetags( Te, tLine );
    set_timetags( Te_perp, tLine );
    set_timetags( Te_para, tLine );
    set_timetags( err, tLine );
    set_timetags( rT, tLine );
    set_timetags( rN, tLine );
    set_timetags( rV, tLine );
    set_timetags( Ie, tLine );
    set_timetags( p_off_on, tLine );
    set_timetags( p_perp_ratio, tLine );
    
    return TMC_OK;

}

//main correction loop
TMC_ERROR tmc_data::correct_moments(double GL, double GU, double GR, double phi_offset, 
    string *destination, string *logname, bool logging, bool clobber, double const_sc_pot)
{

    //initialise all the necessary stuff we need

    string D = *destination;
    D += *logname;

    //struct to handle bottom & overlap moment calculation
    TRUNC_STRUCT calculation_struct;
    //struct to handle adding the top moments
    TRUNC_STRUCT add_top_struct;
    //struct to handle final parameters
    TRUNC_STRUCT final_struct;

    //intermediate things
    double tmc_N, tmc_top_N, tmc_lat, tmc_long, tmc_Tperp, tmc_Tpara, tmc_EL, tmc_EU, tmc_raw_U;
    double tmc_sunaz;  //sjs july2007
    double tmc_b_field[3] = { 0., 0., 0. };
    double tmc_v[3] = { 0., 0., 0. };
    double tmc_top_v[3] = { 0., 0., 0. };
    double tmc_v0[3] = { 0., 0., 0. };
    double tmc_Vsc, tmc_top_Vsc;
    double tmc_P[3][3] = {{ 0., 0., 0. },{ 0., 0., 0. },{ 0., 0., 0. }};
    double tmc_top_P[3][3] = {{ 0., 0., 0. },{ 0., 0., 0. },{ 0., 0., 0. }};
    double tmc_rT = 0., tmc_rN = 0., tmc_rV = 0., tmc_Ie = 0.;
    double totally_raw_P[3][3] = {{ 0., 0., 0. },{ 0., 0., 0. },{ 0., 0., 0. }};
    double totally_raw_v[3] = { 0., 0., 0. };
    double totally_raw_N = 0.;
    int i,j;
    bool heea_or_leea;
    double cf;
    double av_Phi = 0.;
    double av_b[3] = {0., 0., 0.};
    double av_rT = 0.;
    double av_rV = 0.;
    double av_rN = 0.;
    double N_from_solver = 0.;
    double NV_from_solver[3] = {0.,0.,0.};
    double P_from_solver[3][3] = {{ 0., 0., 0. },{ 0., 0., 0. },{ 0., 0., 0. }};
    double N_with_top = 0.;
    double NV_with_top[3] = {0.,0.,0.};
    double P_with_top[3][3] = {{ 0., 0., 0. },{ 0., 0., 0. },{ 0., 0., 0. }};
    int identity[3][3] =  {{ 1, 0, 0 },{ 0, 1, 0 },{ 0, 0, 1 }};
    double direction_of_vsc[3] = {0.,0.,0.};
    double small_p = 0.;
    double NV_from_solver_mag = 0.;
    double NV_with_top_mag = 0.;
    double raw_V[3] = {0., 0., 0.};
    double max_Phi = -1000.;
    double min_Phi = 1000.;
    time_t begin;
    time_t finish;
    double leea_gain_correction_factor;
    double heea_gain_correction_factor;
    double av_b_mag;
    double field_aligned_p_tensor[3][3] = {{ 0., 0., 0. },{ 0., 0., 0. },{ 0., 0., 0. }};
    double T_para_check, T_perp_check, A_sc;
    int p=0, q=0;
    //mark when we begin calculating
    time(&begin);
    double glcopy = GL, gucopy = GU, grcopy = GR;
    SENSOR_MODE m = DOUBLE_SENSOR_MODE;
    double tcf = 1.;
    double bcf = 1.;
    double ocf = 1.;
    int bottom_source = LEEA;
    int top_source = HEEA;

    //to count how many times we fail and succeed
    int nSuccess = 0;
    int nFailure = 0;
    char dBuf[ 1024 ];

    int nSize = j_ss_overlap_N->sequence_size();

    if(heea_overlap == TRUE)
    	heea_or_leea = HEEA;
    else
	    heea_or_leea = LEEA;


    TRUNC_ERR ret_val;
    ofstream  debug;

    if(logging)
    {

    if(!clobber)
    {
    ifstream test_exists(D.c_str());

    //check if there is a logfile with the same name, if so add an 'A' to the name
    while(1!=2)
    {
    	ifstream test_exists(D.c_str());

      if(!test_exists.is_open()){ test_exists.close(); break; }
	    else{D.insert(D.length()-4, "A");}

    	test_exists.close();
    }
    
    }

    debug.open( D.c_str(), ofstream::out );
    if(!debug.is_open())
    {cout << "\nError writing log file!\n"; return TMC_ERR;}

    }
    else
    {

    cout << "\nIn reticent mode: no log file will be generated\n";

    }

    
    cout << endl;

    //*************************************
    //main loop
    //*************************************
    for (int nIndex = 0; nIndex < nSize; nIndex++)
    {


	m = determine_sensor_mode(nIndex);

	if(bottom)
	{
	    if((int)j_bottom_source[nIndex]==1)
    		bottom_source=LEEA;
    	    else
    		bottom_source=HEEA;
	}

	if(top)
	{
    	    if((int)j_top_source[nIndex]==1)
    		top_source=LEEA;
    	    else
	 	top_source=HEEA;
	}

	//test each time whether we should use B moments
	if( (bottom==true) && (m==DOUBLE_SENSOR_MODE) ){bottom = use_B_moments_at_all((double)j_bottom_N[nIndex]);}

	// Initialise the structure to carry out the calculations
	set_defaults_tstruct(&calculation_struct);

	//set the spacecraft potential (note this is not yet Phi, where Phi = 1-U)
	tmc_raw_U  = (double)j_efw_U_probe_sc[ nIndex ];
	if (tmc_raw_U > 1000. || tmc_raw_U < -1000. ) tmc_raw_U = 1.0-const_sc_pot;  
	// protection sjs apr2005; NB min_Phi = +1000, max_Phi = -1000. used
	// elsewhere as initialisers and changed in loop! 

	//calibration info for leea -- if no calibration data set the gain correction to unity,
	//otherwise set the correct value given the current time
	// sjs: gain_correction routine treats gain as steps from start to next start
	// which is ok - it doesn't join/interpolate. July 2007
	if( leea_gain_correction.is_nil() )
	  leea_gain_correction_factor = 1;
	else
	  leea_gain_correction_factor = gain_correction( current_time_calc(nIndex), leea_gain_correction);

	//ditto for heea
	if( heea_gain_correction.is_nil() )
	  heea_gain_correction_factor = 1;
	else
	  heea_gain_correction_factor = gain_correction( current_time_calc(nIndex), heea_gain_correction);
	// gain factors for early part of mission need special treatment:
	fiddle_gain_factors(  current_time_calc(nIndex), &leea_gain_correction_factor, &heea_gain_correction_factor );

	populate_totally_raw_moments(m, &totally_raw_N, totally_raw_v, totally_raw_P, nIndex, heea_or_leea, leea_gain_correction_factor, heea_gain_correction_factor, phi_offset);

	populate_low_energy_moments(m, &tmc_EL, &tmc_EU, &tmc_N, tmc_v, tmc_P, leea_gain_correction_factor, heea_gain_correction_factor, phi_offset, nIndex, heea_or_leea, &tcf, &bcf, &ocf);

	if(tmc_N==0 && vector_mag(tmc_v)==0)
	{
			GL = -2; GU = 2; GR = 1;
	}
	else
	{
	  		GL = glcopy; GU = gucopy; GR = grcopy;
	}

	//get the high moments (i.e. T)
	if(m == DOUBLE_SENSOR_MODE)
	populate_high_energy_moments(&tmc_top_N, tmc_top_v, tmc_top_P, tcf, nIndex);
	else
	{
		tmc_top_N = 0.;
		tmc_top_v[0] = 0.; tmc_top_v[1] = 0.; tmc_top_v[2] = 0.;
		for(i=0;i<3;i++)
			for(j=0;j<3;j++)
				tmc_top_P[i][j] = 0.;
	}

	cout << "Solving... Progress: " << (100*nIndex)/(nSize-1) << " % at point " << current_time_calc(nIndex);

	//set auxillary data lat and long
	tmc_lat = (double)j_aux_latitude[ nIndex ];
	tmc_long = (double)j_aux_longitude[ nIndex ];

        // sunaz is 26.2 - SPOS * 360/(16*1024)
        tmc_sunaz = 26.2 - 0.02197265625*(double)j_cpcdm_spos[ nIndex ];
        // cout << "Sunaz = " << tmc_sunaz << endl;  

	//set fgm mag field vector
	tmc_b_field[ 0 ] = (double)j_fgm_B_vec[ nIndex ][ 0 ] ;
	tmc_b_field[ 1 ] = (double)j_fgm_B_vec[ nIndex ][ 1 ] ;
	tmc_b_field[ 2 ] = (double)j_fgm_B_vec[ nIndex ][ 2 ] ;

	//populate calculation structure
	set_Nsc( &calculation_struct, tmc_N );
	set_Psc( &calculation_struct, tmc_P );
	set_vxsc( &calculation_struct, tmc_v[ 0 ]);
	set_vysc( &calculation_struct, tmc_v[ 1 ] );
	set_vzsc( &calculation_struct, tmc_v[ 2 ]);
	set_vl( &calculation_struct, tmc_EL );
	set_vu( &calculation_struct, tmc_EU  );
	set_phi_sc( &calculation_struct, tmc_raw_U  );
	set_Vsc( &calculation_struct );
	set_VTsc( &calculation_struct );

	//recall that the direction of v is left unchanged. keep track of it with this array
	direction_of_vsc[0] = calculation_struct.vxsc/calculation_struct.Vsc;
	direction_of_vsc[1] = calculation_struct.vysc/calculation_struct.Vsc;
	direction_of_vsc[2] = calculation_struct.vzsc/calculation_struct.Vsc;

	//calc curlyE, vL,U from phi_sc, vl,u
	calc_intermediary( &calculation_struct );
	//takes above and fills zeta,vl,u,eps
	prepare_normalised_nl_inputs( &calculation_struct );
	//solve
	ret_val = normalised_nl_solver( &calculation_struct, GL, GU, GR );
	//get maxwellian values from normalised results
	dimensional_from_nl_solution( &calculation_struct );

	//now to add the top moments...
	//...do we think there are top moments?
	if(top)
	{
	    //yes we do, but shall we actually use them?
	  if(use_T_moments((double)j_top_N[nIndex]))
	  {

		cout <<'\r';

		//default structure to handle things
		set_defaults_tstruct(&add_top_struct);

	  //populate structure with solved data from overlap and bottom moments if present
                copy_trunc_struct_maxwellians(&calculation_struct, &add_top_struct);

          //set phi=0 as we have corrected for the spacecraft potential
      		add_top_struct.phi_sc = 0;

		//integrate from zero...
		set_vl( &add_top_struct, 0 );

		//...to the upper limit of the overlap
		set_vu( &add_top_struct, tmc_EU  );

		//calculate some data and populate
		calc_intermediary(&add_top_struct);

		//get Vsc, Nsc, etc from solved quantities
		calc_moments_from_maxwellian(&add_top_struct);

		//set the overlap and bottom 'moments'
		N_from_solver = add_top_struct.Nsc;

		NV_from_solver[0] = add_top_struct.Vsc*add_top_struct.Nsc * direction_of_vsc[0];
		NV_from_solver[1] = add_top_struct.Vsc*add_top_struct.Nsc * direction_of_vsc[1];
		NV_from_solver[2] = add_top_struct.Vsc*add_top_struct.Nsc * direction_of_vsc[2];

		NV_from_solver_mag = sqrt(NV_from_solver[0]*NV_from_solver[0] + NV_from_solver[1]*NV_from_solver[1] + NV_from_solver[2]*NV_from_solver[2]);

		//now add on the top moments
		N_with_top = N_from_solver + tmc_top_N;

	        add_top_struct.n0 = N_with_top;

	        NV_with_top[0] = NV_from_solver[0] + tmc_top_v[0];
		NV_with_top[1] = NV_from_solver[1] + tmc_top_v[1];
		NV_with_top[2] = NV_from_solver[2] + tmc_top_v[2];
    
		NV_with_top_mag = vector_mag(NV_with_top);

    //calculate Tr(P) = 3p0 + mN(2V0Ve-Ve**2) --> get p0
		//small_p = (add_top_struct.trPsc + M_ELECTRON*N_from_solver*((2*NV_from_solver_mag*add_top_struct.v0/N_from_solver) - add_top_struct.v0*add_top_struct.v0))/3;

                 small_p = (add_top_struct.trPsc - 2*M_ELECTRON*NV_from_solver_mag*calculation_struct.v0 + M_ELECTRON*N_from_solver*calculation_struct.v0*calculation_struct.v0)/3;

    //cout << add_top_struct.trPsc << '\t' << calculation_struct.v0 << '\t' << NV_from_solver_mag << '\t' << N_from_solver << '\t' << small_p << endl;
        
		//now can reconstruct P0
		for(p=0;p<3;p++)
			for(q=0;q<3;q++)
				{
            //P_from_solver[p][q] = small_p*identity[p][q] - M_ELECTRON*N_from_solver*((2*NV_from_solver_mag*add_top_struct.v0/N_from_solver) - add_top_struct.v0*add_top_struct.v0)*direction_of_vsc[p]*direction_of_vsc[q];
            P_from_solver[p][q] = small_p*identity[p][q] + 2*M_ELECTRON*NV_from_solver_mag*calculation_struct.v0*direction_of_vsc[p]*direction_of_vsc[q] - M_ELECTRON*N_from_solver*calculation_struct.v0*calculation_struct.v0*direction_of_vsc[p]*direction_of_vsc[q];

        }


		for(p=0;p<3;p++)
			for(q=0;q<3;q++)
				{
					P_with_top[p][q] = P_from_solver[p][q]+tmc_top_P[p][q];
					tmc_P[p][q] = P_with_top[p][q];
				}


		add_top_struct.VTsc = sqrt( (2./3.) * ( (trace(P_with_top)/(N_with_top*M_ELECTRON)) -  (NV_with_top_mag*NV_with_top_mag)/(N_with_top*N_with_top) ) );

		add_top_struct.T0 = 0.5*M_ELECTRON*add_top_struct.VTsc*add_top_struct.VTsc/K_BOLTZ;

	        tmc_v[0] = add_top_struct.vxsc + tmc_top_v[0]/N_with_top;
		tmc_v[1] = add_top_struct.vysc + tmc_top_v[1]/N_with_top;
		tmc_v[2] = add_top_struct.vzsc + tmc_top_v[2]/N_with_top;

		// get Maxwellian components of velocity vector in GSE
		transform_to_gse_vector( tmc_lat, tmc_long , tmc_sunaz, tmc_v0 );  
                // sjs july 2007: this looks null, as tmc_v0[] recalculated and transformed_to_gse
		// by handover_params, which actually popoulates Ve[] vector

	        tmc_v0[0] = (NV_with_top_mag/N_with_top) * direction_of_vsc[0];  // sjs these look right
	        tmc_v0[1] = (NV_with_top_mag/N_with_top) * direction_of_vsc[1];
	 	tmc_v0[2] = (NV_with_top_mag/N_with_top) * direction_of_vsc[2];
		
		// homdover...: renormalises by v0{solver}/Vsc,  transforms to gse, and handed over to Ve[] vector
	        handover_params(&add_top_struct, tmc_v0, tmc_lat, tmc_long, tmc_sunaz, nIndex); 
   
		transform_to_gse_vector( tmc_lat, tmc_long , tmc_sunaz, tmc_v );         // transform v (velocity x, y, z) with lat,  long
		transform_to_gse_tensor( tmc_lat, tmc_long, tmc_sunaz, tmc_P );          // transform P (pressure) with lat (latitude), long (longitude)
		// sjs I don't see where tmc_v and tmc_P are used below.


    }  // end if we are supposed to use top moments
    else  // else there are top moments but we aren't supposed to use them
    {

	    cout << '\r';

  	    handover_params(&calculation_struct, tmc_v0, tmc_lat, tmc_long, tmc_sunaz, nIndex);
		// sjs: so we take tmc_v0, which gets scaled and transformed by handover, &put into Ve[]vec
   
	    transform_to_gse_vector( tmc_lat, tmc_long , tmc_sunaz, tmc_v );         // transform v (velocity x, y, z) with lat,  long
	    transform_to_gse_tensor( tmc_lat, tmc_long, tmc_sunaz, tmc_P );   // transform P (pressure) with lat (latitude), long (longitude)
		// again I don't see where these are used below

	    }

	} 
    else  // else there aren't top moments
	{
	    cout << '\r';

      handover_params(&calculation_struct, tmc_v0, tmc_lat, tmc_long, tmc_sunaz, nIndex);

            transform_to_gse_vector( tmc_lat, tmc_long, tmc_sunaz, tmc_v );         // transform v (velocity x, y, z) with lat,  long
	    transform_to_gse_tensor( tmc_lat, tmc_long, tmc_sunaz, tmc_P );   // transform P (pressure) with lat (latitude), long (longitude)
		// again I don't see where these are used below

	 }

   transform_to_gse_vector( tmc_lat, tmc_long, tmc_sunaz, totally_raw_v);         // transform v (velocity x, y, z) with lat,  long
   transform_to_gse_tensor( tmc_lat, tmc_long, tmc_sunaz, totally_raw_P);

  set_defaults_tstruct(&final_struct);

	//populate structure with solved data from overlap and bottom moments if present
  //also populate 'raw' slots with 'totally raw data'
   final_struct.Nsc = totally_raw_N;
   final_struct.Vsc = vector_mag(totally_raw_v);
   final_struct.vxsc = totally_raw_v[0];
	 final_struct.vysc = totally_raw_v[1];
	 final_struct.vzsc = totally_raw_v[2];

    for(i=0;i<3;i++)
		  for(j=0;j<3;j++)
		  	final_struct.Psc[i][j] = totally_raw_P[i][j];

	 final_struct.n0 = Ne[nIndex];
	 final_struct.v0 = sqrt(Ve[nIndex][0]*Ve[nIndex][0] + Ve[nIndex][1]*Ve[nIndex][1] + Ve[nIndex][2]*Ve[nIndex][2]);
	 final_struct.vT = VT[nIndex];
	 final_struct.T0 = Te[nIndex];
	 final_struct.trPsc = trace(final_struct.Psc);


	if(tmc_b_field[0] == 0 &&  tmc_b_field[1] == 0 && tmc_b_field[2] == 0)
	{
	   tmc_Tpara = 0.;
	   tmc_Tperp = 0.;
	   T_para_check = 0.;
	   T_perp_check = 0.;

	   for(i=0;i<3;i++)
	   	  for(j=0;j<3;j++)
			      field_aligned_p_tensor[i][j] = 0.;

	     p_off_on[nIndex] = 0.;
  	   p_perp_ratio[nIndex] = 0.;

	}
	else
	{
	   tmc_Tpara = estimate_anisotropy_T0_para( &final_struct, tmc_b_field, totally_raw_P, totally_raw_v );
	   tmc_Tperp = estimate_anisotropy_T0_perp( &final_struct, tmc_b_field, totally_raw_P, totally_raw_v );
	   p_special_transform(&add_top_struct, tmc_b_field, totally_raw_P, totally_raw_v, field_aligned_p_tensor);
	   A_sc = (0.5*(field_aligned_p_tensor[1][1]+field_aligned_p_tensor[2][2]))/field_aligned_p_tensor[0][0];
	   T_para_check = 3*final_struct.T0/(1+2*A_sc);
	   T_perp_check = A_sc*T_para_check;
     p_off_on[nIndex] = (fabs(( field_aligned_p_tensor[0][1] + field_aligned_p_tensor[0][2] + field_aligned_p_tensor[1][2] )/3)) / (( field_aligned_p_tensor[0][0] + field_aligned_p_tensor[1][1] + field_aligned_p_tensor[2][2] )/3);
  	 p_perp_ratio[nIndex] = (field_aligned_p_tensor[1][1] / field_aligned_p_tensor[2][2]);

	}

	Te_para[ nIndex ] = tmc_Tpara;
	Te_perp[ nIndex ] = tmc_Tperp;

	calculate_estimation_ratios(&final_struct, &tmc_rN, &tmc_rV, &tmc_rT, &tmc_Ie);

	rT[ nIndex ] = (double)tmc_rT;
	rN[ nIndex ] = (double)tmc_rN;
	rV[ nIndex ] = (double)tmc_rV;
	Ie[ nIndex ] = (double)tmc_Ie;
  err[ nIndex ] = (double)ret_val;

	if(bottom==true && m==DOUBLE_SENSOR_MODE)
	{
	    if(!use_B_moments(1-tmc_raw_U, lower_energy_lookup((int)j_bottom_EL[ nIndex ],
	                                                       (int)j_bottom_energymode[nIndex],
							       spacecraft, bottom_source),
			      phi_offset,(double)j_bottom_N[nIndex]))
	    {
	    if( ((int)err[nIndex] == 0) || ((int)err[nIndex] == 6))
	    {err[nIndex] = TRUNC_NO_B_MOMS; ret_val = TRUNC_NO_B_MOMS;}
	    }

	}

  //count success and failure
  if( (ret_val==0) || (ret_val==6) || (ret_val==9)  )
	{
    nSuccess++;
	  av_rV += rV[nIndex];
    av_rN += rN[nIndex];
	  av_rT += rT[nIndex];
	}
  else
	  nFailure++;

	av_Phi += 1 - tmc_raw_U;

	if( (1 - tmc_raw_U) < min_Phi)
		min_Phi = 1 - tmc_raw_U;

	if( (1 - tmc_raw_U) > max_Phi)
		max_Phi = 1 - tmc_raw_U;

	av_b[0] += tmc_b_field[ 0 ];
	av_b[1] += tmc_b_field[ 1 ];
	av_b[2] += tmc_b_field[ 2 ];

  //get correct units
  
	VT[ nIndex ] /= 1.e3; //km/s
	Ne[ nIndex ] /= 1.e6; //cm**-3
	Ve[ nIndex ][ 0 ]  /= 1.e3; //km/s
	Ve[ nIndex ][ 1 ]  /= 1.e3;
	Ve[ nIndex ][ 2 ]  /= 1.e3;

  raw_V[0] = totally_raw_v[0]/1.e3;
  raw_V[1] = totally_raw_v[1]/1.e3;
  raw_V[2] = totally_raw_v[2]/1.e3;

  if(logging) // print a log if logging is on
  {
  #ifdef DEBUG_LOG

  debug.precision(6);

	debug << "---------------INDEX = " << nIndex << " @ TIME = " <<  current_time_calc(nIndex) << endl;

	debug << "Error code = " << (int)ret_val << ": " <<  trunc_err_messages(ret_val) << endl;
	debug << endl;
	debug << "RAW ------> CORRECTED MOMENTS" << endl;
	debug << endl;
	debug << "Ne (/cm**3) = " << totally_raw_N/1.e6 << " ------> " << Ne[nIndex] << endl;
	debug << "Ve_x (km/s) = " << raw_V[0] << " ------> " << Ve[ nIndex ][ 0 ] << endl;
	debug << "Ve_y (km/s) = " << raw_V[1] << " ------> " << Ve[ nIndex ][ 1 ] << endl;
	debug << "Ve_z (km/s) = " << raw_V[2] << " ------> " << Ve[ nIndex ][ 2 ] << endl;
	debug << "Te (K)      = " << get_raw_T(rT[nIndex], Te[nIndex]) << " ------> " << Te[nIndex] << endl;
  debug << "v_T (km/s) = " << VT[nIndex] << endl;
	debug << endl;
	debug << "Raw Pressure tensor (Pa)     " << endl;
	debug <<  totally_raw_P[ 0 ][ 0 ] << "  " << totally_raw_P[ 0 ][ 1 ] << "  " << totally_raw_P[ 0 ][ 2 ] << endl;
	debug <<  totally_raw_P[ 1 ][ 0 ] << "  " << totally_raw_P[ 1 ][ 1 ] << "  " << totally_raw_P[ 1 ][ 2 ] << endl;
	debug <<  totally_raw_P[ 2 ][ 0 ] << "  " << totally_raw_P[ 2 ][ 1 ] << "  " << totally_raw_P[ 2 ][ 2 ] << endl;
	debug << endl;
	if(tmc_b_field[0] == 0 &&  tmc_b_field[1] == 0 && tmc_b_field[2] == 0)
  {debug << "No FGM B-vector, therefore no anisotropy estimation\n" << endl;}
	else
	{
	debug << "Estimated Te_para (K) = " << Te_para[ nIndex ] << endl;
	debug << "Estimated Te_perp (K) = " << Te_perp[ nIndex ] << endl;
	debug << "p tensor (field aligned)" << endl;
	debug <<  field_aligned_p_tensor[ 0 ][ 0 ] << "  " << field_aligned_p_tensor[ 0 ][ 1 ] << "  " << field_aligned_p_tensor[ 0 ][ 2 ] << endl;
	debug <<  field_aligned_p_tensor[ 1 ][ 0 ] << "  " << field_aligned_p_tensor[ 1 ][ 1 ] << "  " << field_aligned_p_tensor[ 1 ][ 2 ] << endl;
	debug <<  field_aligned_p_tensor[ 2 ][ 0 ] << "  " << field_aligned_p_tensor[ 2 ][ 1 ] << "  " << field_aligned_p_tensor[ 2 ][ 2 ] << endl;
	debug << endl;
	debug << "rp_off_on = " << p_off_on[nIndex] << '\t' << "rp_perp_ratio = " << p_perp_ratio[nIndex] << endl;
  debug << endl;
	}
	debug << "Estimation parameters" << endl;
	debug << "rT = " << rT[nIndex] << endl;
	debug << "rV = " << rV[nIndex] << endl;
	debug << "rN = " << rN[nIndex] << endl;
	debug << "Ie = " << Ie[nIndex] << endl;
	debug << endl;
	debug << "Supplementary inputs" << endl;
	debug << "Latitude (deg) = " << tmc_lat << endl;
	debug << "Longitude (deg) = " << tmc_long << endl;
	debug << "Delta Phi (V) = " << phi_offset << endl;
	debug << "U (V) = " << tmc_raw_U << '\t' << "Phi (V) = " << 1 - tmc_raw_U << endl;
	debug << "bx = " << tmc_b_field[ 0 ] << endl;
	debug << "by = " << tmc_b_field[ 1 ] << endl;
	debug << "bz = " << tmc_b_field[ 2 ] << endl;
	debug << endl;
	debug << "Guess window parameters   " << endl;
	debug << "Glower (%) = " << GL << endl;
	debug << "Gupper (%) = " << GU << endl;
	debug << "Gres   (%) = " << GR << endl;
	debug << endl;
	debug << "Integration limits" << endl;
	debug << endl;
	if(m==SINGLE_SENSOR_MODE_TOP)
	debug << "SINGLE_SENSOR_MODE_TOP" << endl;
	else if(m==SINGLE_SENSOR_MODE_BOTTOM)
	debug << "SINGLE_SENSOR_MODE_BOTTOM" << endl;
	else
	debug << "DOUBLE_SENSOR_MODE" << endl;

	if(heea_or_leea==LEEA && m==DOUBLE_SENSOR_MODE)
	{
	debug << "L1L2 EL (eV) - lookup = " << lower_energy_lookup((int)j_ss_overlap_EL[ nIndex ], (int)j_ss_overlap_energymode[nIndex], spacecraft, heea_or_leea) << '\t';
	debug << "EU (eV) - lookup = " <<  upper_energy_lookup((int)j_ss_overlap_EU[ nIndex ], (int)j_ss_overlap_energymode[nIndex], spacecraft, heea_or_leea)  << endl;
	}
	else if(heea_or_leea==HEEA && m==DOUBLE_SENSOR_MODE)
	{
	debug << "H1H2 EL (eV) - lookup = " << lower_energy_lookup((int)j_ss_overlap_EL[ nIndex ], (int)j_ss_overlap_energymode[nIndex], spacecraft, heea_or_leea) << '\t';
        debug << "EU (eV) - lookup = " <<  upper_energy_lookup((int)j_ss_overlap_EU[ nIndex ], (int)j_ss_overlap_energymode[nIndex], spacecraft, heea_or_leea)  << endl;
	}
        debug << endl;
	debug << "used EL (eV) = " << tmc_EL << '\t';
	debug << "EU (eV) = " << tmc_EU << endl;
	debug << endl;

	if(!bottom)
	    debug << "No B-moments (user specified or non-existent)" << endl;
	else if(bottom==true && m==DOUBLE_SENSOR_MODE)
	{
	    if(!use_B_moments(1-tmc_raw_U,lower_energy_lookup((int)j_bottom_EL[ nIndex ],  (int)j_bottom_energymode[nIndex], spacecraft, 0),phi_offset,(double)j_bottom_N[nIndex]))
	    {
		debug << "No B-moments (solver's decision)" << endl;
		debug << "B EL - lookup (eV) = " <<  lower_energy_lookup((int)j_bottom_EL[ nIndex ],  (int)j_bottom_energymode[nIndex], spacecraft, 0) << endl;

	    }
	    else
	    	debug << "B-moments present" << endl;

	}
	else if(bottom && m==SINGLE_SENSOR_MODE_BOTTOM)
		debug << "B-moments present" << endl;

	if(!top)
	    debug << "No T-moments" << endl;
	else if(top)
	{
	    if(use_T_moments((double)j_top_N[nIndex]) == FALSE)
		debug << "No T-moments" << endl;
	    else
		debug << "T-moments present" << endl;
	}

	debug << "Bottom gain correction = " << bcf << endl;
	debug << "Overlap gain correction = " << ocf << endl;
	debug << "Top gain correction = " << tcf << endl;


	debug << endl;




#endif
} //end logging if

    }


    time(&finish);

    debug.close();

    for(i=0;i<3;i++)
  	av_b[i] /= nSize;

    av_b_mag = vector_mag(av_b);

    //print summary

    cout.precision(3);
    cout << endl;

    cout << "\nSummary:" << endl;
    cout <<   "          Calculation duration:            ";
    if(difftime(finish, begin) > 90)
    cout << (float)(difftime(finish, begin)/60) << " mins" << endl;
    else
    cout << difftime(finish, begin) << " secs" << endl;
    cout <<   "          Actual data start:               " << current_time_calc(0) << endl;
    cout <<   "          Actual data end:                 " << current_time_calc(nSize-1) << endl;
    cout <<   "          Number of points solved:         " << nSize << endl;
    cout <<   "          Number of successes:             " << nSuccess << endl;
    cout <<   "          Number of failures:              " << nFailure << endl;
    if(av_rN<0)
    cout <<   "          Average under-estimation of Ne:  " << (int)(-100*av_rN/nSize) << " %" << endl;
    else
    cout <<   "          Average over-estimation of Ne:   " << (int)(100*av_rN/nSize)<< " %" << endl;
    if(av_rT<0)
    cout <<   "          Average under-estimation of Te:  " << (int)(-100*av_rT/nSize)<< " %" << endl;
    else
    cout <<   "          Average over-estimation of Te:   " << (int)(100*av_rT/nSize)<< " %" << endl;
    if(av_rV<0)
    cout <<   "          Average under-estimation of Ve:  " << (int)(-100*av_rV/nSize)<< " %" << endl;
    else
    cout <<   "          Average over-estimation of Ve:   " << (int)(100*av_rV/nSize)<< " %" << endl;
    cout << endl;
    cout <<   "          Average spacecraft potential:    " << av_Phi/nSize << " V  [hi = " << max_Phi << " V, lo = " << min_Phi << " V]" << endl;
    if(tmc_b_field[0] == 0 &&  tmc_b_field[1] == 0 && tmc_b_field[2] == 0)
    {
    cout <<   "          Average direction of B field:    " << "(none)" << endl;
    }
    else
    {
    cout <<   "          Average direction of B field:    " << "(" << av_b[0]/(av_b_mag);
    						       cout << " , " << av_b[1]/(av_b_mag);
						       cout << " , " << av_b[2]/(av_b_mag) << ")" << endl;
    }
    cout << endl;
    return TMC_OK;

}

//
//join all the necessary data seqs
//(this just loops over the above join_a_to_b)
//the number of sequences which need to be joined (and in which order)
//depends on whether there are B, T, overlap moments e.g. BL1L2 vs BL1L2T,
//etc
//
	TMC_ERROR tmc_data::join_all()
	{

	QdObjectSeq_var toJoin;
	QdObjectSeq_var Joined;
	QdObjectSeq_var joinOnto;

	int i;

  //
  //now go through the different combinations of moments
  //
  if(bottom==TRUE && top==TRUE)
	{

	toJoin = new QdObjectSeq(46);
	Joined = new QdObjectSeq(46);
	joinOnto = new QdObjectSeq(46);

	for(i=0; i<46; i++)
		joinOnto[i] = ss_overlap_N;

	toJoin[0] = ss_overlap_N;
	toJoin[1] = ss_overlap_Vx;
	toJoin[2] = ss_overlap_Vy;
	toJoin[3] = ss_overlap_Vz;
	toJoin[4] = ss_overlap_Pxx;
	toJoin[5] = ss_overlap_Pxy;
	toJoin[6] = ss_overlap_Pxz;
	toJoin[7] = ss_overlap_Pyy;
	toJoin[8] = ss_overlap_Pyz;
	toJoin[9] = ss_overlap_Pzz;
	toJoin[10] = ss_overlap_EL;
	toJoin[11] = ss_overlap_EU;
	toJoin[12] = ss_overlap_energymode;

	toJoin[13] = bottom_N;
	toJoin[14] = bottom_Vx;
	toJoin[15] = bottom_Vy;
	toJoin[16] = bottom_Vz;
	toJoin[17] = bottom_Pxx;
	toJoin[18] = bottom_Pxy;
	toJoin[19] = bottom_Pxz;
	toJoin[20] = bottom_Pyy;
	toJoin[21] = bottom_Pyz;
	toJoin[22] = bottom_Pzz;
	toJoin[23] = bottom_EL;
	toJoin[24] = bottom_EU;
	toJoin[25] = bottom_energymode;
	toJoin[26] = bottom_source;

	toJoin[27] = top_N;
	toJoin[28] = top_Vx;
	toJoin[29] = top_Vy;
	toJoin[30] = top_Vz;
	toJoin[31] = top_Pxx;
	toJoin[32] = top_Pxy;
	toJoin[33] = top_Pxz;
	toJoin[34] = top_Pyy;
	toJoin[35] = top_Pyz;
	toJoin[36] = top_Pzz;
	toJoin[37] = top_EL;
	toJoin[38] = top_EU;
	toJoin[39] = top_energymode;
	toJoin[40] = top_source;


	toJoin[41] = fgm_B_vec;
	toJoin[42] = efw_U_probe_sc;
	toJoin[43] = aux_latitude;
	toJoin[44] = aux_longitude;
        toJoin[45] = cpcdm_spos;

	if(join_a_to_b(toJoin, joinOnto[0], Joined, 46) != TMC_OK)
			return TMC_ERR;


	j_ss_overlap_N = QdRScalarSeq_var::narrow(Joined[0]);
	j_ss_overlap_Vx = QdRScalarSeq_var::narrow(Joined[1]);
	j_ss_overlap_Vy = QdRScalarSeq_var::narrow(Joined[2]);
	j_ss_overlap_Vz = QdRScalarSeq_var::narrow(Joined[3]);
	j_ss_overlap_Pxx = QdRScalarSeq_var::narrow(Joined[4]);
	j_ss_overlap_Pxy = QdRScalarSeq_var::narrow(Joined[5]);
	j_ss_overlap_Pxz = QdRScalarSeq_var::narrow(Joined[6]);
	j_ss_overlap_Pyy = QdRScalarSeq_var::narrow(Joined[7]);
	j_ss_overlap_Pyz = QdRScalarSeq_var::narrow(Joined[8]);
	j_ss_overlap_Pzz = QdRScalarSeq_var::narrow(Joined[9]);
	j_ss_overlap_EL = QdRScalarSeq_var::narrow(Joined[10]);
	j_ss_overlap_EU = QdRScalarSeq_var::narrow(Joined[11]);
	j_ss_overlap_energymode = QdRScalarSeq_var::narrow(Joined[12]);

	j_bottom_N = QdRScalarSeq_var::narrow(Joined[13]);
	j_bottom_Vx = QdRScalarSeq_var::narrow(Joined[14]);
	j_bottom_Vy = QdRScalarSeq_var::narrow(Joined[15]);
	j_bottom_Vz = QdRScalarSeq_var::narrow(Joined[16]);
	j_bottom_Pxx = QdRScalarSeq_var::narrow(Joined[17]);
	j_bottom_Pxy = QdRScalarSeq_var::narrow(Joined[18]);
	j_bottom_Pxz = QdRScalarSeq_var::narrow(Joined[19]);
	j_bottom_Pyy = QdRScalarSeq_var::narrow(Joined[20]);
	j_bottom_Pyz = QdRScalarSeq_var::narrow(Joined[21]);
	j_bottom_Pzz = QdRScalarSeq_var::narrow(Joined[22]);
	j_bottom_EL = QdRScalarSeq_var::narrow(Joined[23]);
	j_bottom_EU = QdRScalarSeq_var::narrow(Joined[24]);
	j_bottom_energymode = QdRScalarSeq_var::narrow(Joined[25]);
	j_bottom_source = QdRScalarSeq_var::narrow(Joined[26]);

	j_top_N = QdRScalarSeq_var::narrow(Joined[27]);
	j_top_Vx = QdRScalarSeq_var::narrow(Joined[28]);
	j_top_Vy = QdRScalarSeq_var::narrow(Joined[29]);
	j_top_Vz = QdRScalarSeq_var::narrow(Joined[30]);
	j_top_Pxx = QdRScalarSeq_var::narrow(Joined[31]);
	j_top_Pxy = QdRScalarSeq_var::narrow(Joined[32]);
	j_top_Pxz = QdRScalarSeq_var::narrow(Joined[33]);
	j_top_Pyy = QdRScalarSeq_var::narrow(Joined[34]);
	j_top_Pyz = QdRScalarSeq_var::narrow(Joined[35]);
	j_top_Pzz = QdRScalarSeq_var::narrow(Joined[36]);
	j_top_EL = QdRScalarSeq_var::narrow(Joined[37]);
	j_top_EU = QdRScalarSeq_var::narrow(Joined[38]);
	j_top_energymode = QdRScalarSeq_var::narrow(Joined[39]);
	j_top_source = QdRScalarSeq_var::narrow(Joined[40]);

	j_fgm_B_vec = QdRMatrixSeq_var::narrow(Joined[41]);
	j_efw_U_probe_sc = QdRScalarSeq_var::narrow(Joined[42]);
	j_aux_latitude = QdRScalarSeq_var::narrow(Joined[43]);
	j_aux_longitude = QdRScalarSeq_var::narrow(Joined[44]);
	j_cpcdm_spos = QdRScalarSeq_var::narrow(Joined[45]);


	}
	else if(bottom==TRUE && top==FALSE)
	{

	toJoin = new QdObjectSeq(32);
	Joined = new QdObjectSeq(32);
	joinOnto = new QdObjectSeq(32);

	for(i=0; i<32; i++)
		joinOnto[i] = ss_overlap_N;

	toJoin[0] = ss_overlap_N;
	toJoin[1] = ss_overlap_Vx;
	toJoin[2] = ss_overlap_Vy;
	toJoin[3] = ss_overlap_Vz;
	toJoin[4] = ss_overlap_Pxx;
	toJoin[5] = ss_overlap_Pxy;
	toJoin[6] = ss_overlap_Pxz;
	toJoin[7] = ss_overlap_Pyy;
	toJoin[8] = ss_overlap_Pyz;
	toJoin[9] = ss_overlap_Pzz;
	toJoin[10] = ss_overlap_EL;
	toJoin[11] = ss_overlap_EU;
	toJoin[12] = ss_overlap_energymode;

	toJoin[13] = bottom_N;
	toJoin[14] = bottom_Vx;
	toJoin[15] = bottom_Vy;
	toJoin[16] = bottom_Vz;
	toJoin[17] = bottom_Pxx;
	toJoin[18] = bottom_Pxy;
	toJoin[19] = bottom_Pxz;
	toJoin[20] = bottom_Pyy;
	toJoin[21] = bottom_Pyz;
	toJoin[22] = bottom_Pzz;
	toJoin[23] = bottom_EL;
	toJoin[24] = bottom_EU;
	toJoin[25] = bottom_energymode;
	toJoin[26] = bottom_source;

	toJoin[27] = fgm_B_vec;
	toJoin[28] = efw_U_probe_sc;
	toJoin[29] = aux_latitude;
	toJoin[30] = aux_longitude;
        toJoin[31] = cpcdm_spos;

	if(join_a_to_b(toJoin, joinOnto[0], Joined, 32) != TMC_OK)
			return TMC_ERR;


	j_ss_overlap_N = QdRScalarSeq_var::narrow(Joined[0]);
	j_ss_overlap_Vx = QdRScalarSeq_var::narrow(Joined[1]);
	j_ss_overlap_Vy = QdRScalarSeq_var::narrow(Joined[2]);
	j_ss_overlap_Vz = QdRScalarSeq_var::narrow(Joined[3]);
	j_ss_overlap_Pxx = QdRScalarSeq_var::narrow(Joined[4]);
	j_ss_overlap_Pxy = QdRScalarSeq_var::narrow(Joined[5]);
	j_ss_overlap_Pxz = QdRScalarSeq_var::narrow(Joined[6]);
	j_ss_overlap_Pyy = QdRScalarSeq_var::narrow(Joined[7]);
	j_ss_overlap_Pyz = QdRScalarSeq_var::narrow(Joined[8]);
	j_ss_overlap_Pzz = QdRScalarSeq_var::narrow(Joined[9]);
	j_ss_overlap_EL = QdRScalarSeq_var::narrow(Joined[10]);
	j_ss_overlap_EU = QdRScalarSeq_var::narrow(Joined[11]);
	j_ss_overlap_energymode = QdRScalarSeq_var::narrow(Joined[12]);

	 j_bottom_N = QdRScalarSeq_var::narrow(Joined[13]);
	j_bottom_Vx = QdRScalarSeq_var::narrow(Joined[14]);
	j_bottom_Vy = QdRScalarSeq_var::narrow(Joined[15]);
	j_bottom_Vz = QdRScalarSeq_var::narrow(Joined[16]);
	j_bottom_Pxx = QdRScalarSeq_var::narrow(Joined[17]);
	j_bottom_Pxy = QdRScalarSeq_var::narrow(Joined[18]);
	j_bottom_Pxz = QdRScalarSeq_var::narrow(Joined[19]);
	j_bottom_Pyy = QdRScalarSeq_var::narrow(Joined[20]);
	j_bottom_Pyz = QdRScalarSeq_var::narrow(Joined[21]);
	j_bottom_Pzz = QdRScalarSeq_var::narrow(Joined[22]);
	j_bottom_EL = QdRScalarSeq_var::narrow(Joined[23]);
	j_bottom_EU = QdRScalarSeq_var::narrow(Joined[24]);
	j_bottom_energymode = QdRScalarSeq_var::narrow(Joined[25]);
	j_bottom_source = QdRScalarSeq_var::narrow(Joined[26]);

	j_fgm_B_vec = QdRMatrixSeq_var::narrow(Joined[27]);
	j_efw_U_probe_sc = QdRScalarSeq_var::narrow(Joined[28]);
	j_aux_latitude = QdRScalarSeq_var::narrow(Joined[29]);
	j_aux_longitude = QdRScalarSeq_var::narrow(Joined[30]);
	j_cpcdm_spos = QdRScalarSeq_var::narrow(Joined[31]);



	}
	else if(bottom==FALSE && top==TRUE)
	{

	toJoin = new QdObjectSeq(32);
	Joined = new QdObjectSeq(32);
	joinOnto = new QdObjectSeq(32);

	for(i=0; i<32; i++)
		joinOnto[i] = ss_overlap_N;

	toJoin[0] = ss_overlap_N;
	toJoin[1] = ss_overlap_Vx;
	toJoin[2] = ss_overlap_Vy;
	toJoin[3] = ss_overlap_Vz;
	toJoin[4] = ss_overlap_Pxx;
	toJoin[5] = ss_overlap_Pxy;
	toJoin[6] = ss_overlap_Pxz;
	toJoin[7] = ss_overlap_Pyy;
	toJoin[8] = ss_overlap_Pyz;
	toJoin[9] = ss_overlap_Pzz;
	toJoin[10] = ss_overlap_EL;
	toJoin[11] = ss_overlap_EU;
	toJoin[12] = ss_overlap_energymode;

	toJoin[13] = top_N;
	toJoin[14] = top_Vx;
	toJoin[15] = top_Vy;
	toJoin[16] = top_Vz;
	toJoin[17] = top_Pxx;
	toJoin[18] = top_Pxy;
	toJoin[19] = top_Pxz;
	toJoin[20] = top_Pyy;
	toJoin[21] = top_Pyz;
	toJoin[22] = top_Pzz;
	toJoin[23] = top_EL;
	toJoin[24] = top_EL;
	toJoin[25] = top_energymode;
	toJoin[26] = top_source;

	toJoin[27] = fgm_B_vec;
	toJoin[28] = efw_U_probe_sc;
	toJoin[29] = aux_latitude;
	toJoin[30] = aux_longitude;
	toJoin[31] = cpcdm_spos;

	if(join_a_to_b(toJoin, joinOnto[0], Joined, 32) != TMC_OK)
			return TMC_ERR;


	j_ss_overlap_N = QdRScalarSeq_var::narrow(Joined[0]);
	j_ss_overlap_Vx = QdRScalarSeq_var::narrow(Joined[1]);
	j_ss_overlap_Vy = QdRScalarSeq_var::narrow(Joined[2]);
	j_ss_overlap_Vz = QdRScalarSeq_var::narrow(Joined[3]);
	j_ss_overlap_Pxx = QdRScalarSeq_var::narrow(Joined[4]);
	j_ss_overlap_Pxy = QdRScalarSeq_var::narrow(Joined[5]);
	j_ss_overlap_Pxz = QdRScalarSeq_var::narrow(Joined[6]);
	j_ss_overlap_Pyy = QdRScalarSeq_var::narrow(Joined[7]);
	j_ss_overlap_Pyz = QdRScalarSeq_var::narrow(Joined[8]);
	j_ss_overlap_Pzz = QdRScalarSeq_var::narrow(Joined[9]);
	j_ss_overlap_EL = QdRScalarSeq_var::narrow(Joined[10]);
	j_ss_overlap_EU = QdRScalarSeq_var::narrow(Joined[11]);
	j_ss_overlap_energymode = QdRScalarSeq_var::narrow(Joined[12]);

	j_top_N = QdRScalarSeq_var::narrow(Joined[13]);
	j_top_Vx = QdRScalarSeq_var::narrow(Joined[14]);
	j_top_Vy = QdRScalarSeq_var::narrow(Joined[15]);
	j_top_Vz = QdRScalarSeq_var::narrow(Joined[16]);
	j_top_Pxx = QdRScalarSeq_var::narrow(Joined[17]);
	j_top_Pxy = QdRScalarSeq_var::narrow(Joined[18]);
	j_top_Pxz = QdRScalarSeq_var::narrow(Joined[19]);
	j_top_Pyy = QdRScalarSeq_var::narrow(Joined[20]);
	j_top_Pyz = QdRScalarSeq_var::narrow(Joined[21]);
	j_top_Pzz = QdRScalarSeq_var::narrow(Joined[22]);
	j_top_EL = QdRScalarSeq_var::narrow(Joined[23]);
	j_top_EU = QdRScalarSeq_var::narrow(Joined[24]);
	j_top_energymode = QdRScalarSeq_var::narrow(Joined[25]);
	j_top_source = QdRScalarSeq_var::narrow(Joined[26]);

	j_fgm_B_vec = QdRMatrixSeq_var::narrow(Joined[27]);
	j_efw_U_probe_sc = QdRScalarSeq_var::narrow(Joined[28]);
	j_aux_latitude = QdRScalarSeq_var::narrow(Joined[29]);
	j_aux_longitude = QdRScalarSeq_var::narrow(Joined[30]);
	j_cpcdm_spos = QdRScalarSeq_var::narrow(Joined[31]);


	}
	else
	{

	toJoin = new QdObjectSeq(18);
	Joined = new QdObjectSeq(18);
	joinOnto = new QdObjectSeq(18);

	for(i=0; i<18; i++)
		joinOnto[i] = ss_overlap_N;

	toJoin[0] = ss_overlap_N;
	toJoin[1] = ss_overlap_Vx;
	toJoin[2] = ss_overlap_Vy;
	toJoin[3] = ss_overlap_Vz;
	toJoin[4] = ss_overlap_Pxx;
	toJoin[5] = ss_overlap_Pxy;
	toJoin[6] = ss_overlap_Pxz;
	toJoin[7] = ss_overlap_Pyy;
	toJoin[8] = ss_overlap_Pyz;
	toJoin[9] = ss_overlap_Pzz;
	toJoin[10] = ss_overlap_EL;
	toJoin[11] = ss_overlap_EU;
	toJoin[12] = ss_overlap_energymode;


	toJoin[13] = fgm_B_vec;
	toJoin[14] = efw_U_probe_sc;
	toJoin[15] = aux_latitude;
	toJoin[16] = aux_longitude;
	toJoin[17] = cpcdm_spos;

	if(join_a_to_b(toJoin, joinOnto[0], Joined, 18) != TMC_OK)
			return TMC_ERR;


	j_ss_overlap_N = QdRScalarSeq_var::narrow(Joined[0]);
	j_ss_overlap_Vx = QdRScalarSeq_var::narrow(Joined[1]);
	j_ss_overlap_Vy = QdRScalarSeq_var::narrow(Joined[2]);
	j_ss_overlap_Vz = QdRScalarSeq_var::narrow(Joined[3]);
	j_ss_overlap_Pxx = QdRScalarSeq_var::narrow(Joined[4]);
	j_ss_overlap_Pxy = QdRScalarSeq_var::narrow(Joined[5]);
	j_ss_overlap_Pxz = QdRScalarSeq_var::narrow(Joined[6]);
	j_ss_overlap_Pyy = QdRScalarSeq_var::narrow(Joined[7]);
	j_ss_overlap_Pyz = QdRScalarSeq_var::narrow(Joined[8]);
	j_ss_overlap_Pzz = QdRScalarSeq_var::narrow(Joined[9]);
	j_ss_overlap_EL = QdRScalarSeq_var::narrow(Joined[10]);
	j_ss_overlap_EU = QdRScalarSeq_var::narrow(Joined[11]);
	j_ss_overlap_energymode = QdRScalarSeq_var::narrow(Joined[12]);


	j_fgm_B_vec = QdRMatrixSeq_var::narrow(Joined[13]);
	j_efw_U_probe_sc = QdRScalarSeq_var::narrow(Joined[14]);
	j_aux_latitude = QdRScalarSeq_var::narrow(Joined[15]);
	j_aux_longitude = QdRScalarSeq_var::narrow(Joined[16]);
	j_cpcdm_spos = QdRScalarSeq_var::narrow(Joined[17]);



	}

	return TMC_OK;

	}


        //
        //access methods
        //
 	      double* tmc_data::get_Ne(int i) const{ double *val = new double; *val = (double)cNe[i]; return val; }
	      
	      double* tmc_data::get_Vx(int i) const{ double *val = new double; *val = (double)cVe[i][0]; return val; }

	      double* tmc_data::get_Vy(int i) const{ double *val = new double; *val = (double)cVe[i][1]; return val; }

	      double* tmc_data::get_Vz(int i) const{ double *val = new double; *val = (double)cVe[i][2]; return val; }

	      double* tmc_data::get_VT(int i) const{ double *val = new double; *val = (double)cVT[i]; return val; }

        double* tmc_data::get_Te(int i) const{ double *val = new double; *val = (double)cTe[i]; return val; }

      	double* tmc_data::get_Te_perp(int i) const
	      { double *val = new double; *val = (double)cTe_perp[i]; return val; }

      	      double* tmc_data::get_Te_para(int i) const
	      { double *val = new double; *val = (double)cTe_para[i]; return val; }

       	      int*    tmc_data::get_err(int i) const
	      { int *val = new int; *val = (int)cerr[i]; return val; }
	      
	      double* tmc_data::get_p_off_on(int i) const
	      { double *val = new double; *val = (double)cp_off_on[i]; return val; }

	      double* tmc_data::get_p_perp_ratio(int i) const
	      { double *val = new double; *val = (double)cp_perp_ratio[i]; return val; }

	      double* tmc_data::get_rT(int i) const{ double *val = new double; *val = (double)crT[i]; return val; }
	      double* tmc_data::get_rN(int i) const{ double *val = new double; *val = (double)crN[i]; return val; }
	      double* tmc_data::get_rV(int i) const{ double *val = new double; *val = (double)crV[i]; return val; }
	      double* tmc_data::get_Ie(int i) const{ double *val = new double; *val = (double)cIe[i]; return val; }
        double* tmc_data::get_phi_sc(int i) const{ double *val = new double; *val = (double)c_phi_sc[i]; return val;}

   //remove failed points (i.e. if error != 0,6,9 )
   //so only output successes
   TMC_ERROR tmc_data::clean_outputs()
   {

          int error_seq_size = err->sequence_size();
  	  int success_count = 0;
  	  SequenceMask error_mask(error_seq_size, TRUE);

  	  int ind = 0;

  	  for(ind=0; ind<error_seq_size; ind++)
  	  {
		if( ((int)err[ind]==0) || ((int)err[ind] == 6) || ((int)err[ind] == 9) )
		    {error_mask[ind]=TRUE; success_count++;}
		else
		    error_mask[ind]=FALSE;
  	  }

    cNe = new QdRScalarSeq( success_count );
    if ( cNe.is_nil() )
	return TMC_IS_NIL_ERR;
    
    cVe = new QdRMatrixSeq( success_count, Dim(3) );
    if ( cVe.is_nil() )
	return TMC_IS_NIL_ERR;

    cTe = new QdRScalarSeq( success_count );
    if ( cTe.is_nil() )
	return TMC_IS_NIL_ERR;
	
    cVT = new QdRScalarSeq( success_count );
    if ( cVT.is_nil() )
	return TMC_IS_NIL_ERR;

    cTe_perp = new QdRScalarSeq( success_count );
    if ( cTe_perp.is_nil() )
	return TMC_IS_NIL_ERR;

    cTe_para = new QdRScalarSeq( success_count );
    if ( cTe_para.is_nil() )
	return TMC_IS_NIL_ERR;

    cerr = new QdRScalarSeq( success_count );
    if ( cerr.is_nil() )
	return TMC_IS_NIL_ERR;

    crT = new QdRScalarSeq( success_count );
    if ( crT.is_nil() )
	return TMC_IS_NIL_ERR;

    crN = new QdRScalarSeq( success_count );
    if ( crN.is_nil() )
	return TMC_IS_NIL_ERR;
    
    crV = new QdRScalarSeq( success_count );
    if ( crV.is_nil() )
	return TMC_IS_NIL_ERR;

    cIe = new QdRScalarSeq( success_count );
    if ( cIe.is_nil() )
	return TMC_IS_NIL_ERR;

    cp_perp_ratio = new QdRScalarSeq( success_count );
    if ( cp_perp_ratio.is_nil() )
	  return TMC_IS_NIL_ERR;

    cp_off_on = new QdRScalarSeq( success_count );
    if ( cp_off_on.is_nil() )
	  return TMC_IS_NIL_ERR;

    c_phi_sc = new QdRScalarSeq( success_count );
    if(c_phi_sc.is_nil() )
    return TMC_IS_NIL_ERR;
  
    cNe->copy_xrefs_from( Ne );
    cVe->copy_xrefs_from( Ve );
    cVT->copy_xrefs_from( Ne );
    cTe->copy_xrefs_from( Ne );
    cTe_perp->copy_xrefs_from( Ne );
    cTe_para->copy_xrefs_from( Ne );
    cerr->copy_xrefs_from( Ne );
    
    QdTimeSeq_var tt_masked = new QdTimeSeq();
    QdTimeSeq_var tLine = get_timetags( j_ss_overlap_N );

    tLine->get_subsequence(*tt_masked, error_mask);

    Ne->get_subsequence(*cNe, error_mask);
    Ve->get_subsequence(*cVe, error_mask);
    VT->get_subsequence(*cVT, error_mask);
    Te->get_subsequence(*cTe, error_mask);
    Te_perp->get_subsequence(*cTe_perp, error_mask);
    Te_para->get_subsequence(*cTe_para, error_mask);
    rT->get_subsequence(*crT, error_mask);
    rN->get_subsequence(*crN, error_mask);
    rV->get_subsequence(*crV, error_mask);
    Ie->get_subsequence(*cIe, error_mask);
    err->get_subsequence(*cerr, error_mask);
    p_off_on->get_subsequence(*cp_off_on, error_mask);
    p_perp_ratio->get_subsequence(*cp_perp_ratio, error_mask);
    j_efw_U_probe_sc->get_subsequence(*c_phi_sc, error_mask);
    set_timetags( cNe, tt_masked );
    set_timetags( cVe, tt_masked );
    set_timetags( cerr, tt_masked );

    set_timetags( cVT, tt_masked );
    set_timetags( cTe, tt_masked );
    set_timetags( cTe_perp, tt_masked );
    set_timetags( cTe_para, tt_masked );
    set_timetags( crT, tt_masked );
    set_timetags( crN, tt_masked );
    set_timetags( crV, tt_masked );
    set_timetags( cIe, tt_masked );
    set_timetags( cp_off_on, tt_masked );
    set_timetags( cp_perp_ratio, tt_masked );
    set_timetags( c_phi_sc, tt_masked );

    for(int i=0;i<success_count;i++)
      {c_phi_sc[i] *= -1;
       c_phi_sc[i] += 1;}
      
    return TMC_OK;

    }

//get the time of the point i in the correction loop
const char * tmc_data::current_time(int i) const
{
	    QdTimeSeq_var line = get_timetags( QdRScalarSeq_var::narrow(cerr) );
    	StringRep T;
    	QdTime_var time = new QdTime;
    	QdTimeSeq_var time_seq = new QdTimeSeq( cerr->sequence_size());
    	time_seq = QdTimeSeq_var::narrow(line);
	
       *time = time_seq[i];
       
       static char time_str[25]; //qdos doesn't insert T and Z in iso strings
       int y,m,d,h,min;
       float sec;
       
       time->split_ymd(y,m,d,h,min,sec);
       sprintf(time_str,"%4.4d-%2.2d-%2.2dT%2.2d:%2.2d:%06.3fZ",y,m,d,h,min,sec);

	//return T(time).c_str();
	return time_str;

}

//same as above, but for general sequence seq
const char * tmc_data::current_time_from_seq(int i, QdRScalarSeq_var seq) const
{
	    QdTimeSeq_var line = get_timetags( seq );
    	StringRep T;
    	QdTime_var time = new QdTime;
    	QdTimeSeq_var time_seq = new QdTimeSeq( seq->sequence_size());
    	time_seq = QdTimeSeq_var::narrow(line);

       *time = time_seq[i];
       
       static char time_str[25]; //qdos doesn't insert T and Z in iso strings
       int y,m,d,h,min;
       float sec;
       
       time->split_ymd(y,m,d,h,min,sec);
       sprintf(time_str,"%4d-%2d-%2dT%2d:%2d:%06.3fZ",y,m,d,h,min,sec);

	//return T(time).c_str();
	return time_str;

}

//
//return time string of position i in calc
//
const char * tmc_data::current_time_calc(int i) const
{
	QdTimeSeq_var line = get_timetags( QdRScalarSeq_var::narrow(j_ss_overlap_N) );
    	StringRep T;
    	QdTime_var time = new QdTime;
    	QdTimeSeq_var time_seq = new QdTimeSeq( j_ss_overlap_N->sequence_size());
    	time_seq = QdTimeSeq_var::narrow(line);

        *time = time_seq[i];

	return T(time).c_str();

}

//set all seqs to null for safety
void tmc_data::null_all()
{

 	      leea_gain_correction = NULL;
	      heea_gain_correction = NULL;


	      //the raw moments -- overlap
	       overlap_N = NULL;
	       overlap_Vx = NULL;
	       overlap_Vy = NULL;
	       overlap_Vz = NULL;
	       overlap_Pxx = NULL;
	       overlap_Pxy = NULL;
	       overlap_Pxz = NULL;
	       overlap_Pyy = NULL;
	       overlap_Pyz = NULL;
	       overlap_Pzz = NULL;
	       overlap_EL = NULL;
	       overlap_EU = NULL;
	       overlap_spinmode = NULL;
               overlap_energymode = NULL;

	      //subsetted and summed
               ss_overlap_N = NULL;
	       ss_overlap_Vx = NULL;
	       ss_overlap_Vy = NULL;
	       ss_overlap_Vz = NULL;
	       ss_overlap_Pxx = NULL;
	       ss_overlap_Pxy = NULL;
	       ss_overlap_Pxz = NULL;
	       ss_overlap_Pyy = NULL;
	       ss_overlap_Pyz = NULL;
	       ss_overlap_Pzz = NULL;
	       ss_overlap_EL = NULL;
	       ss_overlap_EU = NULL;
	       ss_overlap_energymode = NULL;

	      //the raw moments -- bottom and top
	       bottom_N = NULL;
	       bottom_Vx = NULL;
	       bottom_Vy = NULL;
	       bottom_Vz = NULL;
	       bottom_Pxx = NULL;
	       bottom_Pxy = NULL;
	       bottom_Pxz = NULL;
	       bottom_Pyy = NULL;
	       bottom_Pyz = NULL;
	       bottom_Pzz = NULL;
	       bottom_EL = NULL;
	       bottom_EU = NULL;
	       bottom_energymode = NULL;
	       bottom_source = NULL;

	       top_N = NULL;
	       top_Vx = NULL;
	       top_Vy = NULL;
	       top_Vz = NULL;
	       top_Pxx = NULL;
	       top_Pxy = NULL;
	       top_Pxz = NULL;
	       top_Pyy = NULL;
	       top_Pyz = NULL;
	       top_Pzz = NULL;
	       top_EL = NULL;
	       top_energymode = NULL;
 	       top_EU = NULL;
	       top_source = NULL;

	      //other data we need
	       fgm_B_vec = NULL;
	       efw_U_probe_sc = NULL;
	       aux_latitude = NULL;
	       aux_longitude = NULL;
	       cpcdm_spos = NULL;

	      //joined data
	       j_ss_overlap_N = NULL;
	       j_ss_overlap_Vx = NULL;
	       j_ss_overlap_Vy = NULL;
	       j_ss_overlap_Vz = NULL;
	       j_ss_overlap_Pxx = NULL;
	       j_ss_overlap_Pxy = NULL;
	       j_ss_overlap_Pxz = NULL;
	       j_ss_overlap_Pyy = NULL;
	       j_ss_overlap_Pyz = NULL;
	       j_ss_overlap_Pzz = NULL;
	       j_ss_overlap_EL = NULL;
	       j_ss_overlap_EU = NULL;
	       j_ss_overlap_energymode = NULL;

	       j_bottom_N = NULL;
	       j_bottom_Vx = NULL;
	       j_bottom_Vy = NULL;
	       j_bottom_Vz = NULL;
	       j_bottom_Pxx = NULL;
	       j_bottom_Pxy = NULL;
	       j_bottom_Pxz = NULL;
	       j_bottom_Pyy = NULL;
	       j_bottom_Pyz = NULL;
	       j_bottom_Pzz = NULL;
	       j_bottom_EL = NULL;
	       j_bottom_energymode = NULL;
	       j_bottom_EU = NULL;
	       j_bottom_source = NULL;

	       j_top_N = NULL;
	       j_top_Vx = NULL;
	       j_top_Vy = NULL;
	       j_top_Vz = NULL;
	       j_top_Pxx = NULL;
	       j_top_Pxy = NULL;
	       j_top_Pxz = NULL;
	       j_top_Pyy = NULL;
	       j_top_Pyz = NULL;
	       j_top_Pzz = NULL;
	       j_top_EL = NULL;
	       j_top_energymode = NULL;
               j_top_EU = NULL;
	       j_top_source = NULL;


	      //other data we need - joined
	       j_fgm_B_vec = NULL;
	       j_efw_U_probe_sc = NULL;
	       j_aux_latitude = NULL;
	       j_aux_longitude = NULL;
	       j_cpcdm_spos = NULL;

	      //corrected moments & output data
	       Ne = NULL;
	       Te = NULL;
	       Te_perp = NULL;
               Te_para = NULL;
	       Ve = NULL;
	       VT = NULL;
	       rN = NULL;
               rV = NULL;
	       rT = NULL;
      	       err = NULL;
               Ie = NULL;
	       p_off_on = NULL;
	       p_perp_ratio = NULL;

	      //clean corrected moments & output data
	       cNe = NULL;
	       cTe = NULL;
	       cTe_perp = NULL;
         cTe_para = NULL;
	       cVe = NULL;
	       cVT = NULL;
	       crN = NULL;
         crV = NULL;
	       crT = NULL;
      	 cerr = NULL;
         cIe = NULL;
	       cp_off_on = NULL;
	       cp_perp_ratio = NULL;
         c_phi_sc = NULL;

}

//shift sequence half a space to the right -- used for separation
//of overlap half moments
TMC_ERROR  tmc_data::shift_time_series( QdObject_var  rTobeShifted )
{

    if ( rTobeShifted.is_nil() )
    {
	    return TMC_ERR;
    }

    QdTimeSeq_var    tTime;
    tTime = get_timetags( rTobeShifted );

    if ( tTime.is_nil() )
    {
    	return TMC_ERR;
    }

    double dTimeInterval = 0.;

    // We do assume that there is no gap within this time series
    if ( !QuIsTSRegular( tTime, &dTimeInterval ) );

    dTimeInterval /= 2;
    *tTime += Seconds( dTimeInterval );
    set_timetags( rTobeShifted, tTime );

    return ( TMC_OK );
}

//
//shift a sequence of sequences by 1/2 a space
//
TMC_ERROR tmc_data::shift_all_time_series( QdObjectSeq_var  inputs, int nSeq )
{
    for ( int i = 0; i < nSeq; i++ )
    {
    	QdObject_var   dObject = QdObject_var::narrow((inputs)[i]);

      if ( shift_time_series( dObject ) != TMC_OK )
	    {
	      return ( TMC_ERR );
	    }

    }

    return ( TMC_OK );
}

//
//check whether we should use the bottom moments looking at delta phi
//(or value=0)
//
bool tmc_data::use_B_moments(double U, double EL, double phi_offset, double value)
{
    if(value==0)
	return FALSE;
    else
    {
	    if( ((EL-U) < phi_offset) || (EL == 26460))
		return FALSE; //don't use them
	    else
		return TRUE; //use them
	}

}

//
//are B and T present or set to 0?
//
bool tmc_data::use_T_moments(double value) //probably density would be best
{
    if(value==0.0)
	return FALSE;
    else
	return TRUE;
}

//
//the user may have requested B, but none may exist (B==0)
//
bool tmc_data::use_B_moments_at_all(double value)
{


       if(value == 0.)
	return FALSE;
       else
	return TRUE;


}

//
//the user may have requested T, but none may exist (T==0)
//
bool tmc_data::use_T_moments_at_all(QdObjectSeq_var inputs)
{

    QdRScalar_var T = QdRScalar_var::narrow((inputs)[26]);
    if ( T.is_nil() )
	      return TRUE;
    else
    {
	    if((double)(*T)==0)
	    	return FALSE;
	    else
		    return TRUE;
    }
}

//
//apply correct gain correction
//
double tmc_data::gain_correction(const char * ctime, QdRScalarSeq_var calibration)
{

    int i = 0, seq = 0;

    QdTime_var lookup_time = new QdTime(ctime);

    seq = calibration->sequence_size();

    QdTimeSeq_var cal_timeseq = get_timetags( calibration );

    QdTimeInterval_var cal_interval;

    for(i=0; i<seq-1; i++)
    {

	cal_interval = new QdTimeInterval(cal_timeseq[i+1], cal_timeseq[i]);

	if(cal_interval->contains(*lookup_time) == TRUE)
	    return (double)(calibration[i]);

	//cout << (double)(narrow_cal[i]) << endl;

    }

    return 1.;

}

//
//access methods for tensors
//
double tmc_data::get_j_ss_overlap_Pij(int i, int j, int N) const
{

switch( i )
{

	case 0: switch( j )
	       {
	       	 case 0:return (double)j_ss_overlap_Pxx[N];

	      	 case 1:return (double)j_ss_overlap_Pxy[N];

	      	 case 2:return (double)j_ss_overlap_Pxz[N];

		 default: break;
	        }

	case 1: switch( j )
	       {
	       	 case 0:return (double)j_ss_overlap_Pxy[N];

	      	 case 1:return (double)j_ss_overlap_Pyy[N];

	      	 case 2:return (double)j_ss_overlap_Pyz[N];
		 
		 default: break;
	        }


	case 2: switch( j )
	       {
	       	 case 0:return (double)j_ss_overlap_Pxz[N];

	      	 case 1:return (double)j_ss_overlap_Pyz[N];

	      	 case 2:return (double)j_ss_overlap_Pzz[N];
		 
		 default: break;
	        }

		default: break;



}


return 1.0;

}


double tmc_data::get_j_bottom_Pij(int i, int j, int N) const
{

switch( i )
{

	case 0: switch( j )
	       {
	       	 case 0:return (double)j_bottom_Pxx[N];

	      	 case 1:return (double)j_bottom_Pxy[N];

	      	 case 2:return (double)j_bottom_Pxz[N];
		 
		 default: break;
	        }

	case 1: switch( j )
	       {
	       	 case 0:return (double)j_bottom_Pxy[N];
	       
	      	 case 1:return (double)j_bottom_Pyy[N];

	      	 case 2:return (double)j_bottom_Pyz[N];
		 
		 default: break;
	        }


	case 2: switch( j )
	       {
	       	 case 0:return (double)j_bottom_Pxz[N];

	      	 case 1:return (double)j_bottom_Pyz[N];

	      	 case 2:return (double)j_bottom_Pzz[N];
		 
		 default: break;
	        }

		default: break;



}


return 1.0;

}

double tmc_data::get_j_top_Pij(int i, int j, int N) const
{

switch( i )
{

	case 0: switch( j )
	       {
	       	 case 0:return (double)j_top_Pxx[N];

	      	 case 1:return (double)j_top_Pxy[N];

	      	 case 2:return (double)j_top_Pxz[N];

		 default: break;
	        }

	case 1: switch( j )
	       {
	       	 case 0:return (double)j_top_Pxy[N];
	       
	      	 case 1:return (double)j_top_Pyy[N];

	      	 case 2:return (double)j_top_Pyz[N];

		 default: break;
	        }	


	case 2: switch( j )
	       {
	       	 case 0:return (double)j_top_Pxz[N];
	       
	      	 case 1:return (double)j_top_Pyz[N];

	      	 case 2:return (double)j_top_Pzz[N];

		 default: break;
	        }
		
		default: break;



}


return 1.0;

}

//remove errant t-tags
QdObject_var tmc_data::check_for_non_conformity(QdObject_var in)
{

int size = in->sequence_size();
int y1, y2, d1, d2, h1, h2, m1, m2;
double s1, s2;
int newsize = 0;
double dspace;

    SequenceMask mask(size, TRUE);

    QdRScalarSeq_var scalarin = QdRScalarSeq_var::narrow(in);

    Time timeY;
    Time timeZ;
    Time initial;
    Time tmp;

    QdTimeSeq_var timeseq = new QdTimeSeq(size);
    timeseq = get_timetags(in);
    if ( !QuIsTSRegular( timeseq , &dspace ) );

    QdTimeSeq_var line = QdTimeSeq_var::narrow(timeseq);

    for(int i=0;i<size-1;i++)
    {
	    timeY = line[i];
	    timeZ = line[i+1];
      initial = line[0];
    	timeY.split_ydoy(y1, d1, h1, m1, s1 );
    	timeZ.split_ydoy(y2, d2, h2, m2, s2 );

	//cout << i << '\t' << timeY << '\t' << timeZ << '\t' << timeZ-timeY << endl;
	//cout << (double)mSeconds(timeZ - timeY) << '\t' << 1000*dspace << endl;
	   if( (timeZ<initial) || (timeZ < timeY) )
	 	 {

		    	mask[i+1]=FALSE;

			for(int j=1;j<(size-1-i);j++)
			{
				tmp = line[i+j+1];
				if( (tmp < timeY)  )
				{mask[i+j+1]=FALSE; i+=j;}
				else{break;}
			}

		    }
		else
		    {
		    	mask[i+1]=TRUE; newsize++;

		    }

    } //end for

    	QdObject_var out;
      QdRScalarSeq_var outscalar = new QdRScalarSeq(newsize);
      QdTimeSeq_var tt_out = new QdTimeSeq();
	    timeseq->get_subsequence(*tt_out, mask);
      scalarin->get_subsequence(*outscalar, mask);
      set_timetags(outscalar, tt_out);

      out = outscalar;

      return out;

}

//calculate rN,T,V
void tmc_data::calculate_estimation_ratios(TRUNC_STRUCT *tsr, double *rN, double *rV, double *rT, double *Ie)
{

double Traw, Nraw, Vraw, p_raw[3][3] = {{0.,0.,0.},{0.,0.,0.},{0.,0.,0.}};
int i, j;

double v_vec_raw[3] = {0., 0., 0.};

v_vec_raw[0] = tsr->vxsc;
v_vec_raw[1] = tsr->vysc;
v_vec_raw[2] = tsr->vzsc;

for (i = 0; i<3; i++)
   for (j=0; j<3; j++)
   	p_raw[i][j] = (tsr->Psc[i][j] - tsr->Nsc*M_ELECTRON*v_vec_raw[i]*v_vec_raw[j]);


Traw = trace(p_raw)/(3*K_BOLTZ*tsr->Nsc);
Vraw = tsr->Vsc;
Nraw = tsr->Nsc;


	    *rT = (Traw - tsr->T0)/tsr->T0;
		if(isnan(*rT))
			*rT = 0;

 	    *rN = (Nraw - tsr->n0)/tsr->n0;
   	        if(isnan(*rN))
			*rN = -100;


	    *rV = (Vraw - tsr->v0)/tsr->v0;
   	        if(isnan(*rV))
			*rV = -100;


	    *Ie = sqrt(*rT * *rT + *rN * *rN + *rV * *rV);
return;
}

//extract 'raw' values from corrected ones
double tmc_data::get_raw_N(double rN, double Ne)
{

return Ne*(rN+1);

}

double tmc_data::get_raw_T(double rT, double Te)
{

return Te*(rT+1);

}

double tmc_data::get_raw_V(double rV, double Ve)
{

return Ve*(rV+1);

}


//as a reference we define the totally raw moments as the
//sum of raw B+overlap+T
void tmc_data::populate_totally_raw_moments(SENSOR_MODE m, double *raw_N, double raw_v[3], double raw_P[3][3], int nIndex, int heea_or_leea, double leea_gain_correction_factor, double heea_gain_correction_factor, double phi_offset)
{

	int i,j;

	double EL = 0, EU = 0,  NL = 0, vL[3]={0., 0., 0.}, PL[3][3] = {{0., 0., 0.}, {0., 0., 0.},{0., 0., 0.}};
	double NU = 0, vU[3]={0., 0., 0.}, PU[3][3] = {{0., 0., 0.}, {0., 0., 0.},{0., 0., 0.}};

  //specific gain correction factors to apply
  double tcf, bcf, ocf;

  //get the raw moments
  populate_low_energy_moments(m, &EL, &EU, &NL, vL, PL,leea_gain_correction_factor, heea_gain_correction_factor,phi_offset, nIndex, heea_or_leea, &tcf, &bcf, &ocf);
 	populate_high_energy_moments(&NU, vU, PU, tcf, nIndex);

  for(i=0;i<3;i++)
	     vL[i]*=NL; //moms come in form NV

	for(i=0;i<3;i++)
	      raw_v[i] = vL[i]+vU[i];

 	*raw_N = NL+NU;

	for(j=0;j<3;j++)
		{	if(*raw_N>0)
			raw_v[j] /= *raw_N;
			else
			raw_v[j]=0; }


	for(i=0;i<3;i++)
		for(j=0;j<3;j++)
			raw_P[i][j] = PL[i][j] + PU[i][j];

	return;

}

//
//as above, but just for B and overlap
//
void tmc_data::populate_low_energy_moments(SENSOR_MODE m, double *tmc_EL, double *tmc_EU, double *tmc_N, double tmc_v[3], double tmc_P[3][3], double leea_gain_correction_factor, double heea_gain_correction_factor, double phi_offset, int nIndex, int heea_or_leea, double *tcf, double *bcf, double *ocf)
{
	int i, j;
	double cf;

	double tf = 1.;
  double bf = 1.;
  double of = 1.;

  determine_limits_and_factors(m, phi_offset, heea_or_leea, tmc_EL, tmc_EU, &of, &tf, &bf, nIndex, leea_gain_correction_factor, heea_gain_correction_factor);

	*ocf = of;
	*bcf = bf;
	*tcf = tf;

	int bottom_source = LEEA;

	if(bottom)
	{
	switch((int)j_bottom_source[nIndex])
		{case 0: bottom_source=HEEA; break;
		case 1:  bottom_source=LEEA;  break;
		default: break;}
	}


	if(bottom==TRUE && m==SINGLE_SENSOR_MODE_BOTTOM)
	{
	  //get the leea overlap and add them to the bottom moments
		*tmc_N = ON_BOARD_FACTOR*(double)j_bottom_N[nIndex]/bf;

		for(i=0;i<3;i++)
		 for(j=0;j<3;j++)
		   tmc_P[i][j] = get_j_bottom_Pij(i,j,nIndex)*M_ELECTRON*ON_BOARD_FACTOR/bf;

			if(*tmc_N>0)
			{
			tmc_v[ 0 ] = (double)j_bottom_Vx[nIndex]*ON_BOARD_FACTOR/(*tmc_N*bf);
			tmc_v[ 1 ] = (double)j_bottom_Vy[nIndex]*ON_BOARD_FACTOR/(*tmc_N*bf);
			tmc_v[ 2 ] = (double)j_bottom_Vz[nIndex]*ON_BOARD_FACTOR/(*tmc_N*bf);
			}
			else {
			  for(i=0;i<3;i++)
			  	tmc_v[i] = 0.; }
	}
	else if(bottom==TRUE && m==DOUBLE_SENSOR_MODE)
	{
			//test if there are and if we can use them
	        if(use_B_moments(1-(double)j_efw_U_probe_sc[nIndex], lower_energy_lookup((int)j_bottom_EL[nIndex],(int)j_bottom_energymode[nIndex],spacecraft, bottom_source), phi_offset,(double)j_bottom_N[nIndex]))
		{

		        //get the leea overlap and add them to the bottom moments
			*tmc_N = ON_BOARD_FACTOR*( ((double)j_bottom_N[nIndex]/bf) + ((double)j_ss_overlap_N[nIndex]/of));

			for(i=0;i<3;i++)
		    		for(j=0;j<3;j++)
					tmc_P[i][j] = ((get_j_ss_overlap_Pij(i,j,nIndex)/of) + (get_j_bottom_Pij(i,j,nIndex)/bf))*M_ELECTRON*ON_BOARD_FACTOR;
                        if(*tmc_N>0)
			{
			tmc_v[ 0 ] = ((((double)j_ss_overlap_Vx[ nIndex ]/of) + ((double)j_bottom_Vx[nIndex]/bf))*ON_BOARD_FACTOR)/(*tmc_N);
			tmc_v[ 1 ] = ((((double)j_ss_overlap_Vy[ nIndex ]/of) + ((double)j_bottom_Vy[nIndex]/bf))*ON_BOARD_FACTOR)/(*tmc_N);
			tmc_v[ 2 ] = ((((double)j_ss_overlap_Vz[ nIndex ]/of) + ((double)j_bottom_Vz[nIndex]/bf))*ON_BOARD_FACTOR)/(*tmc_N);
			}
			else {
			  for(i=0;i<3;i++)
			  	tmc_v[i] = 0.; }

		}
		else
		{
      //get the moments
			*tmc_N = (double)j_ss_overlap_N[ nIndex ]*ON_BOARD_FACTOR/of;

			for(i=0;i<3;i++)
		    		for(j=0;j<3;j++)
				  tmc_P[i][j] = get_j_ss_overlap_Pij(i,j,nIndex)*M_ELECTRON*ON_BOARD_FACTOR/of;

			if(*tmc_N>0)
			{
			tmc_v[ 0 ] = ((double)j_ss_overlap_Vx[ nIndex ]*ON_BOARD_FACTOR)/(*tmc_N*of);
			tmc_v[ 1 ] = ((double)j_ss_overlap_Vy[ nIndex ]*ON_BOARD_FACTOR)/(*tmc_N*of);
			tmc_v[ 2 ] = ((double)j_ss_overlap_Vz[ nIndex ]*ON_BOARD_FACTOR)/(*tmc_N*of);
			}
			else {
			  for(i=0;i<3;i++)
			  	tmc_v[i] = 0.; }
		}



	}
	else if(bottom==FALSE && m==DOUBLE_SENSOR_MODE)
	{
      //get the moments
			*tmc_N = (double)j_ss_overlap_N[ nIndex ]*ON_BOARD_FACTOR/of;

			for(i=0;i<3;i++)
		    		for(j=0;j<3;j++)
				  tmc_P[i][j] = get_j_ss_overlap_Pij(i,j,nIndex)*M_ELECTRON*ON_BOARD_FACTOR/of;

			if(*tmc_N>0)
			{
			tmc_v[ 0 ] = ((double)j_ss_overlap_Vx[ nIndex ]*ON_BOARD_FACTOR)/(*tmc_N*of);
			tmc_v[ 1 ] = ((double)j_ss_overlap_Vy[ nIndex ]*ON_BOARD_FACTOR)/(*tmc_N*of);
			tmc_v[ 2 ] = ((double)j_ss_overlap_Vz[ nIndex ]*ON_BOARD_FACTOR)/(*tmc_N*of);
			}
			else {
			  for(i=0;i<3;i++)
			  	tmc_v[i] = 0.; }
	}
	else if(m==SINGLE_SENSOR_MODE_TOP)
	{

		populate_high_energy_moments(tmc_N, tmc_v, tmc_P, tf, nIndex);

	if(*tmc_N>0)
	{
		for(i=0;i<3;i++)
			tmc_v[i] /= *tmc_N;
	}

	}

//cout << *tmc_N << '\t' << ON_BOARD_FACTOR << endl;
return;

}

//
//as above, but just T
//
void tmc_data::populate_high_energy_moments(double *tmc_top_N, double tmc_top_v[3], double tmc_top_P[3][3], double tcf, int nIndex)
{
        int i,j;

	//we think there are T moments
	if(top)
	{
		//...so test if there are any
		if(use_T_moments((double)j_top_N[nIndex]))
		{
		    //get N, NV and P
		    *tmc_top_N = ON_BOARD_FACTOR*((double)j_top_N[nIndex])/tcf;

        //populate P
		    for(i=0;i<3;i++)
		    	for(j=0;j<3;j++)
			      tmc_top_P[i][j] = get_j_top_Pij(i,j,nIndex)*M_ELECTRON*ON_BOARD_FACTOR/tcf;

		    tmc_top_v[ 0 ] = ((double)j_top_Vx[nIndex]*ON_BOARD_FACTOR)/(tcf);
		    tmc_top_v[ 1 ] = ((double)j_top_Vy[nIndex]*ON_BOARD_FACTOR)/(tcf);
		    tmc_top_v[ 2 ] = ((double)j_top_Vz[nIndex]*ON_BOARD_FACTOR)/(tcf);
		}
       }

return;

}

//
//check whether we are in SINGLE or DOUBLE sensor mode
//(i.e. whether all the moments are in LEEA or HEEA)
//
SENSOR_MODE tmc_data::determine_sensor_mode(int n)
{

//what is the energy mode of the overlap region?
int overlap_energy_mode = (int)j_ss_overlap_energymode[n];
int bottom_energy_mode = 0;
int top_energy_mode = 0;

//4 == no overlap
if(overlap_energy_mode == 4)
{
  //are there bottom or top moms?
	if(bottom)
	{
		bottom_energy_mode = (int)j_bottom_energymode[n];
		if(bottom_energy_mode < 4)
			return SINGLE_SENSOR_MODE_BOTTOM;
	}

	if(top)
	{
		top_energy_mode = (int)j_top_energymode[n];
		if(top_energy_mode < 4)
			return SINGLE_SENSOR_MODE_TOP;
	}

}
else //otherwise normal
	return DOUBLE_SENSOR_MODE;

}

//
//determine whether HEEA or LEEA is in operation for SINGLE sensor mode
//
void tmc_data::determine_limits_and_factors(SENSOR_MODE m, double phi_offset, int heea_or_leea, double *EL, double *EU, double *ocf, double *tcf, double *bcf, int n, double leea_gain_correction_factor, double heea_gain_correction_factor)
{

*tcf = 1.;
*bcf = 1.;

int top_source = HEEA, bottom_source = LEEA;

if(heea_or_leea==LEEA)
	*ocf = leea_gain_correction_factor;
else
	*ocf = heea_gain_correction_factor;

	if(bottom)
	{
	switch((int)j_bottom_source[n])
		{case 0: *bcf = heea_gain_correction_factor; bottom_source=HEEA; break;
		case 1: *bcf = leea_gain_correction_factor; bottom_source=LEEA;  break;
		default: *bcf = 1.0; break;}


	}

	if(top)
	{
	switch((int)j_top_source[n])
		{case 0: *tcf = heea_gain_correction_factor; top_source=HEEA; break;
		case 1: *tcf = leea_gain_correction_factor; top_source=LEEA; break;
		default: *tcf = 1.0; break;}
	}


if(m==SINGLE_SENSOR_MODE_TOP)
{

        *EL = lower_energy_lookup((int)j_top_EL[n],
		    		  (int)j_top_energymode[n],
				  spacecraft, top_source);

        *EU = upper_energy_lookup((int)j_top_EU[n],
		    		  (int)j_top_energymode[n],
				  spacecraft, top_source);
}
else if(m==SINGLE_SENSOR_MODE_BOTTOM)
{
	*EL = lower_energy_lookup((int)j_bottom_EL[n],
		    		  (int)j_bottom_energymode[n],
				  spacecraft, bottom_source);

        *EU = upper_energy_lookup((int)j_bottom_EU[n],
		    		  (int)j_bottom_energymode[n],
				  spacecraft, bottom_source);
}
else
{

	if(bottom)
	{
		*EL = lower_energy_lookup((int)j_bottom_EL[n],(int)j_bottom_energymode[n],spacecraft, bottom_source);
       	 if(!use_B_moments(1-(double)j_efw_U_probe_sc[n], *EL, phi_offset, (double)j_bottom_N[n]))
	 {
		//cout << "No B" << endl;
			*EL = lower_energy_lookup((int)j_ss_overlap_EL[n],
		    		  (int)j_ss_overlap_energymode[n],
				  spacecraft, heea_or_leea);
	}

	}
	else
	*EL = lower_energy_lookup((int)j_ss_overlap_EL[n],(int)j_ss_overlap_energymode[n],spacecraft, heea_or_leea);


        *EU = upper_energy_lookup((int)j_ss_overlap_EU[n],
		    		  (int)j_ss_overlap_energymode[n],
				  spacecraft, heea_or_leea);

}




return;

}

//
//print basic command history from start to end (i.e. energy scan in telemetred units
//
void tmc_data::get_command_history()
{

 int SEQSIZE = j_ss_overlap_N->sequence_size();
 int i = 0;

 cout << "Time" << '\t' << '\t' << '\t' << '\t'
      << "BOT" << '\t'                   
      << "OVERLAP" << '\t' 
      << "TOP" << '\t'
      << "H-SCAN (t)" << '\t' 
      << "L-SCAN (t)" << endl;
   
int HEEA_UPPER, HEEA_LOWER, LEEA_UPPER, LEEA_LOWER;
int HEEA_UPPER_OLD, HEEA_LOWER_OLD, LEEA_UPPER_OLD, LEEA_LOWER_OLD;


   cout << current_time_from_seq(0, j_ss_overlap_N) << '\t' << '\t';

   if((int)j_bottom_EL[0]==89)
   cout << "0" << '\t';
   else
   cout << "1" << '\t';

   if((int)j_ss_overlap_EL[0]==89)
   cout << "0" << '\t';
   else
   cout << "1" << '\t';

   if((int)j_top_EL[0]==89)
   cout << "0" << '\t';
   else
   cout << "1" << '\t';

   if((int)j_top_EU[0]==89)
   {
      HEEA_UPPER = j_ss_overlap_EU[0];
      HEEA_LOWER = j_ss_overlap_EL[0];
   }
   else
   {
      HEEA_UPPER = j_top_EU[0];
      HEEA_LOWER = j_ss_overlap_EL[0];
   }


   if((int)j_bottom_EU[0]==89)
   {
      LEEA_UPPER = j_ss_overlap_EU[0];
      LEEA_LOWER = j_ss_overlap_EL[0];
   }
   else
   {
      LEEA_UPPER = j_ss_overlap_EU[0];
      LEEA_LOWER = j_bottom_EL[0];
   }
     
      
   cout << HEEA_UPPER << " - ";
   cout << HEEA_LOWER << '\t';
   cout << LEEA_UPPER << " - ";
   cout << LEEA_LOWER << '\t';

   cout << endl;
   
 
   HEEA_UPPER_OLD = HEEA_UPPER;
   LEEA_UPPER_OLD = LEEA_UPPER;
   HEEA_LOWER_OLD = HEEA_LOWER;
   LEEA_LOWER_OLD = LEEA_LOWER;
   
for(i=1; i<SEQSIZE; i++)
{

   if((int)j_top_EU[i]==89)
   {
      HEEA_UPPER = j_ss_overlap_EU[i];
      HEEA_LOWER = j_ss_overlap_EL[i];
   }
   else
   {
      HEEA_UPPER = j_top_EU[i];
      HEEA_LOWER = j_ss_overlap_EL[i];
   }


   if((int)j_bottom_EU[i]==89)
   {
      LEEA_UPPER = j_ss_overlap_EU[i];
      LEEA_LOWER = j_ss_overlap_EL[i];
   }
   else
   {
      LEEA_UPPER = j_ss_overlap_EU[i];
      LEEA_LOWER = j_bottom_EL[i];
   }
     

if(i==(SEQSIZE-1) || HEEA_UPPER!=HEEA_UPPER_OLD || LEEA_UPPER!=LEEA_UPPER_OLD || HEEA_LOWER!=HEEA_LOWER_OLD || LEEA_LOWER!=LEEA_LOWER_OLD  )   
{
  cout << current_time_from_seq(i, j_ss_overlap_N) << '\t' << '\t';

  if((int)j_bottom_EL[i]==89)
  cout << "0" << '\t';
  else
  cout << "1" << '\t';

  if((int)j_ss_overlap_EL[i]==89)
  cout << "0" <<  '\t' ;
  else
  cout << "1" <<  '\t' ;

  if((int)j_top_EL[i]==89)
  cout << "0" << '\t';
  else
  cout << "1" << '\t';

  cout << HEEA_UPPER << " to ";
  cout << HEEA_LOWER << '\t';
  cout << LEEA_UPPER << " to ";
  cout << LEEA_LOWER << '\t';

  cout << endl;
}

   HEEA_UPPER_OLD = HEEA_UPPER;
   LEEA_UPPER_OLD = LEEA_UPPER;
   HEEA_LOWER_OLD = HEEA_LOWER;
   LEEA_LOWER_OLD = LEEA_LOWER;
   

}

return;

}

//
//populate trunc_struct with corrected values
//
void tmc_data::handover_params(TRUNC_STRUCT *s, double v0[3], double auxlat, double auxlong, double sunaz, int N)
{

	    Ne[N] = s->n0;

      v0[ 0 ] = s->v0*s->vxsc/s->Vsc;
	    v0[ 1 ] = s->v0*s->vysc/s->Vsc;
	    v0[ 2 ] = s->v0*s->vzsc/s->Vsc;

	    // get Maxwellian components of velocity vector in GSE
	    transform_to_gse_vector( auxlat, auxlong , sunaz, v0 );

	    // handover
	    Ve[N][ 0 ]  = v0[ 0 ];
	    Ve[N][ 1 ]  = v0[ 1 ];
	    Ve[N][ 2 ]  = v0[ 2 ];

	    // Thermal speed
	    VT[N] = s->vT;

	    // Temperature
	    Te[N] = s->T0;

return;
}


void tmc_data::fiddle_gain_factors(const char *ctime, double *leea_gf, double *heea_gf)
{ 
/* modify gain factor to take account of errors and inaccurate onboard gains during early part of mission
Here's the relevant bits from an email from Rob Wilson 26 Apr 2005:
----------------------- snip -----------------------------
between 2000-01-01 00:00:00 and 2001-07-10 11:52:00

LEEA
    0.06339849   CL-1
    0.088160754  CL-2
    0.084556454  CL-3
    0.084640281  CL-4

HEEA
    0.089321082
    0.080960963
    0.076035312
    0.092000578

between 2001-07-10 11:52:00 and 2001-09-12 08:15:00

LEEA
    0.831498558
    0.914651205
    0.831498431
    0.831500655

HEEA
    0.913853715
    0.873669464
    0.725269071
    0.831500619

G_factor = one_of_the_above / G_factor_from_normal_RJW_daily_file

moments_corrected = moments_original * 2 * G_factor
---------------------------- snip --------------------------------------
In QTMC, we calculate moment = raw * ON_BOARD_FACTOR / ocf where ocf is leea_gf or 
heea_gf and is from RJW daily files and ON_BOARD_FACTOR = 2
so for above intervals we need to divide xxxx_gf by the above. Here goes:
*/

QdTime_var lookup_time = new QdTime(ctime);
QdTime_var july = new QdTime("2001-07-10T11:52:00Z");
QdTime_var september = new QdTime("2001-09-12T08:15:00");


if ( *lookup_time > *september) {
   return; // no correction needed
}
if ( *lookup_time > *july ){            // use july->sept
   switch (spacecraft){
     case 1: 
        *leea_gf /= 0.831498558;
	*heea_gf /= 0.913853715; break;
     case 2:
        *leea_gf /= 0.914651205;
	*heea_gf /= 0.873669464; break;
     case 3:
        *leea_gf /= 0.831498431;
	*heea_gf /= 0.725269071; break;
     case 4:
        *leea_gf /= 0.831500655;
	*heea_gf /= 0.831500619; break;
     default:                    break;
   }
   return;
} else {                   // use launch -> july
   switch (spacecraft){
     case 1:
       *leea_gf /= 0.06339849;
       *heea_gf /= 0.089321082;  break;
     case 2:
       *leea_gf /= 0.088160754;
       *heea_gf /= 0.080960963;  break;
     case 3:
       *leea_gf /= 0.084556454;
       *heea_gf /= 0.076035312;  break;
     case 4:
       *leea_gf /= 0.084640281;
       *heea_gf /= 0.092000578;  break;
     default:                    break;
   }
   return;
}

return;  // this should be unreachable

}



