

#include "tmc_routines.h"
#include "tmc_data.h"
#include <string>


#define QFT 0
#define CEF 1
#define CDF 2

TMC_ERROR TMC( int start_year, int start_doy, int start_hour, int start_min, int start_sec, double start_nano,
      	       int end_year, int end_doy, int end_hour, int end_min, int end_sec, double end_nano,
	             int spacecraft, double lower_guess_limit, double upper_guess_limit, double guess_loop_res,
	             double delta_phi, double sc_potential, bool override, bool B, bool L, bool H, bool T,
	             int output_format, string *Lcalname, string *Hcalname, string *dest, bool logging, bool clobber)
{

     string LC, HC, D;
     int output_format_copy = output_format;
     LC = *Lcalname;
     HC = *Hcalname;
     D = *dest;

     if(output_format==CDF)
      output_format=CEF;

     //check to see if the destination directory string has a trailing '/'. if not, add one
     if(D[D.size()-1]!='/')
     D+="/";

    //create ptr to tmc_data object called moments -- construct it with the SC number and the moment sum
    tmc_data *moments = new tmc_data(spacecraft, B, L, H, T);

    //output information about the data set in question: useful for long batchmodes
    StringRep Time;

    QdTime *start_t = new QdTime;
    QdTime *end_t = new QdTime;

    start_t->set_ydoy(start_year, start_doy, start_hour, start_min, start_sec);
    end_t->set_ydoy(end_year, end_doy, end_hour, end_min, end_sec);

   // if((end_year < start_year) || ()
    //  {print_error_message(TMC_INPUT_ERR); return TMC_INPUT_ERR;}
      
    IDFS_TIME start, end;

    cout << "\n*******************************************************************\n";
    cout <<   "QTMC: truncated moments corrector v1.2.0" << endl;
    cout <<   "Spacecraft:    " << spacecraft << endl;
    cout << "Data interval: " << Time(start_t).c_str() << "  to  " <<  Time(end_t).c_str() << endl;
    cout << "Sum:           ";

    if(B==TRUE)
    cout << "B";
    if(H==TRUE)
    cout << "H1H2";
    else
    cout << "L1L2";
    if(T==TRUE)
    cout << "T";
    cout << endl;
    cout << endl;
    cout << "Delta Phi:               " << delta_phi << " Volts" << endl;
    cout << "Manual SC potential:     " << sc_potential << " Volts" << endl;
    cout << "Override EFW U_probe_sc? ";
    if(override==0)
    cout << "NO" << endl;
    else
    cout << "YES" << endl;
    cout << endl;

    if( L==H )
	{print_error_message(TMC_INPUT_ERR); return TMC_INPUT_ERR;} //error --> exit (should never happen)
    /*
    if( (start_year>=end_year) && (start_doy==end_doy) &&
(start_hour==end_hour) && ((end_min - start_min) <= 0))      	
{print_error_message(TMC_INPUT_ERR); return TMC_INPUT_ERR;} //create idfs imes based on arguments yyyy, ddd, etc.
     */
     
to_idfs_time(start_year, start_doy, start_hour, start_min, start_sec, (long int)start_nano, &start);
to_idfs_time(end_year, end_doy, end_hour, end_min, end_sec, (long int)end_nano, &end);    //null all the ptrs in moments for safety (may notneed to use all of them after all),
//e.g. no B or T moments. By definition we always have one
//of the overlap moments: heea or leea

   try{moments->null_all();}
   catch(Exception &e){print_error_message(TMC_ERR); return TMC_ERR;}

   //set the lineages ready for data retrieval using the SC number
   try{moments->set_lineages();}
   catch(Exception &e){print_error_message(TMC_SET_LIN_ERR); return TMC_SET_LIN_ERR;}

   //load the calibration file into a QdRScalarSeq_var for leea...

   try{moments->get_leea_calibration_info(LC);}
   catch(Exception &e){print_error_message(TMC_GET_CAL_ERR); return TMC_GET_CAL_ERR;}
   //...and heea

   try{moments->get_heea_calibration_info(HC);}
   catch(Exception &e){print_error_message(TMC_GET_CAL_ERR); return TMC_GET_CAL_ERR;}

   //get the data between start and end... if can't find efw u_probe_sc
   //from efw then use the value sc_potential
   //if override==true then use sc_potential without getting efw
   

   try
   {
   	if(moments->get_data(start, end, override, sc_potential)!=TMC_OK)
   	{print_error_message(TMC_IDFS_RET_ERR);
   	return TMC_IDFS_RET_ERR;}
   }
   catch(Exception &e)
   {

   print_error_message(TMC_IDFS_RET_ERR);
   return TMC_IDFS_RET_ERR;

   }

   //for the overlap moments -- separate the first and second half of the spin,
   //shift first half forward and add together. join results in spin centred seqs
   try
   {
        if(moments->separate_and_sum_overlap()!=TMC_OK)
   	{print_error_message(TMC_SEP_ERR);
	 return TMC_SEP_ERR;}
   }
   catch(Exception &e)
   {
  	print_error_message(TMC_SEP_ERR);
   	return TMC_SEP_ERR;

   }
   //now join everything to the overlap (density, for argument's sake) tags
   //i.e. all are now 4 sec spin centred
   try
   {
   	if(moments->join_all()!=TMC_OK)
   	{print_error_message(TMC_JOIN_ERR);
	      return TMC_JOIN_ERR;}
   }
   catch(Exception &e)
   {
   	print_error_message(TMC_JOIN_ERR);
        return TMC_JOIN_ERR;

   }

   //create output objects to hold data calculated in main solver function
   try
   {

	if(moments->set_up_outputs()!=TMC_OK)
   	{print_error_message(TMC_CREATE_OUTPUT_ERR);
	return TMC_CREATE_OUTPUT_ERR;}

   }
   catch(Exception &e)
   {
  	print_error_message(TMC_CREATE_OUTPUT_ERR);
  	return TMC_CREATE_OUTPUT_ERR;

   }


   //construct filename for log file

   string logname = logfilename(spacecraft,start_year, start_doy, start_hour, start_min,end_year, end_doy, end_hour, end_min, B, L, H, T);

   //correct the moments with the main solver

   try
   {
   	if(moments->correct_moments(lower_guess_limit, upper_guess_limit,
	     guess_loop_res, delta_phi, &D, &logname, logging, clobber, sc_potential) != TMC_OK)
   	{print_error_message(TMC_MAIN_CALC_ERR);
	return TMC_MAIN_CALC_ERR;}

   }
   catch(Exception &e)
   {
      	print_error_message(TMC_MAIN_CALC_ERR);
   	return TMC_MAIN_CALC_ERR;

   }

   //mask any failures, leaving a nice clean set of objects --> only error codes 0,6 or 9 should remain

   try
   {
   	if(moments->clean_outputs()!=TMC_OK)
   	{print_error_message(TMC_CLEAN_ERR);
	return TMC_CLEAN_ERR;}

   }
   catch(Exception &e)
   {
 	print_error_message(TMC_CLEAN_ERR);
	return TMC_CLEAN_ERR;

   }

   //create some more filenames based on whether output is CEF or QFT and the location of the output dir

    const char * name = new char[100]; // (const char*)malloc(30*sizeof(char));

    if(output_format==1)
    name = filename(spacecraft, start_year, start_doy, start_hour, start_min,end_year, end_doy, end_hour, end_min, B, L, H, T);   //make sure enough mem
    else
    name = qftfilename(spacecraft, start_year, start_doy, start_hour, start_min, end_year, end_doy, end_hour, end_min, B, L, H, T);   //make sure enough mem

    D += name;

    int rename = 0;

    if(output_format_copy==CDF)
    D.replace(D.size()-3, D.size(), "cdf");

    if(!clobber)
    {
    ifstream test_exists(D.c_str());

    //check file doesn't exist -- if it does, append 'A' to the end of the string so we don't overwrite

    while(1!=2) //forever until break
    {
    	ifstream test_exists(D.c_str());

	  if(!test_exists.is_open())
	    {test_exists.close(); break;}
  	else
	  {
       	D.insert(D.length()-4,"A");
	      rename++;
	  }

	  test_exists.close();
    }
    
    }
    ofstream outfile;

    D.replace(D.size()-3, D.size(), "cef");
    
    //open the output stream

    outfile.open( D.c_str(), ofstream::out );
    if(!outfile.is_open())
    {cout << "\nError writing data file!\n"; return TMC_ERR;}

   //write header for cef

   if(output_format==1)
   write_cef_header(outfile, name, moments->get_size(), spacecraft, B, L, H, T);


        int k;

	//add records one by one to file
	for(k=0; k<moments->get_size(); k++)
        {


	   if(output_format==CEF)
           {
		      append_cef_record(outfile,
		        moments->current_time(k),
			      moments->get_Vx(k),
			      moments->get_Vy(k),
			      moments->get_Vz(k),
			      moments->get_VT(k),
			      moments->get_Ne(k),
			      moments->get_Te(k),
			      moments->get_Te_para(k),
			      moments->get_Te_perp(k),
			      moments->get_rT(k),
			      moments->get_rN(k),
			      moments->get_rV(k),
	          moments->get_Ie(k),
			      moments->get_err(k),
			      moments->get_p_off_on(k),
			      moments->get_p_perp_ratio(k),
            moments->get_phi_sc(k)); }
	   else
	   {
		append_qft_record(outfile,
		        moments->current_time(k),
			      moments->get_Vx(k),
			      moments->get_Vy(k),
			      moments->get_Vz(k),
			      moments->get_VT(k),
			      moments->get_Ne(k),
			      moments->get_Te(k),
			      moments->get_Te_para(k),
			      moments->get_Te_perp(k),
			      moments->get_rT(k),
			      moments->get_rN(k),
			      moments->get_rV(k),
	          moments->get_Ie(k),
			      moments->get_err(k),
			      moments->get_p_off_on(k),
			      moments->get_p_perp_ratio(k),
            moments->get_phi_sc(k));

            }


	    }

	//tell the user we're done and what the file is called
  //convert to cdf if that was the option
  string qtran_command = "Qtran -ie -oc ";

  if(clobber)
    qtran_command+="-f ";
  
  string remove_command = "rm -f ";
  if(output_format_copy==CDF)
  {
    qtran_command+=D;
    remove_command+=D;

    cout << "Converting to cdf format...\n";
    if(system(qtran_command.c_str())==-1)
      DebugCout("Couldn't convert to cdf, writing to cef format...");
    else
    {
        system(remove_command.c_str());
        D.replace(D.size()-3, D.size(), "cdf");
    }

  }
  
  cout << "\nWrote file " << D << endl;
	cout << "\n*******************************************************************\n";


	  //done! return OK

    return  TMC_OK;
}


//just a routine to cout the correct message given one of the TMC_ERROR codes

void print_error_message(TMC_ERROR e)
{

switch( e )
{

     case TMC_JOIN_ERR: cout << "\nError correcting moments: join stage" << endl;
          		break;

     case TMC_IDFS_RET_ERR: cout << "\nError correcting moments: data retrieval stage" << endl;
          		break;

     case TMC_IS_NIL_ERR: cout << "\nError correcting moments: sequence is nil" << endl;
          		break;

     case TMC_INPUT_ERR: cout << "\nError correcting moments: check inputs" << endl;
          		break;

     case TMC_SET_LIN_ERR: cout << "\nError correcting moments: failed to set lineages" << endl;
          		break;

     case TMC_GET_CAL_ERR: cout << "\nError correcting moments: failed to get calibration data" << endl;
          		break;

     case TMC_MAIN_CALC_ERR: cout << "\nError correcting moments: main calculation failure" << endl;
          		break;

     case TMC_CREATE_OUTPUT_ERR: cout << "\nError correcting moments: failed to create output objects" << endl;
          		break;

     case TMC_CLEAN_ERR: cout << "\nError correcting moments: failed to clean outputs" << endl;
          		break;

     case TMC_SEP_ERR: cout << "\nError correcting moments: failed separate overlap moments" << endl;
         		break;


     case TMC_ERR: cout << "\nError correcting moments: generic" << endl;
          		break;

     default: cout << "\nError correcting moments: unknown error" << endl;
          		break;

}

return;

}


//handy little function to cout msg when checking code, debugging, etc.

void DebugCout (const char* msg)
{

    cout << msg << endl;

}

double vector_mag(double Q[3])
{

return sqrt(Q[0]*Q[0] + Q[1]*Q[1] + Q[2]*Q[2]);

}

/////////////////////////////////////////////////////////////////////////

TMC_ERROR TMC_HISTORY( int start_year, int start_doy, int start_hour, int start_min, int start_sec, double start_nano,
              	       int end_year, int end_doy, int end_hour, int end_min, int end_sec, double end_nano,
	                     int spacecraft)
{



    //create ptr to tmc_data object called moments -- construct it with the SC number and the moment sum
    tmc_data *moments = new tmc_data(spacecraft, 1, 1, 0, 1);

    //output information about the data set in question: useful for long batchmodes
    StringRep Time;

    QdTime *start_t = new QdTime;
    QdTime *end_t = new QdTime;

    start_t->set_ydoy(start_year, start_doy, start_hour, start_min, start_sec);
    end_t->set_ydoy(end_year, end_doy, end_hour, end_min, end_sec);

    IDFS_TIME start, end;

    
    to_idfs_time(start_year, start_doy, start_hour, start_min, start_sec, (long int)start_nano, &start);
    to_idfs_time(end_year, end_doy, end_hour, end_min, end_sec, (long int)end_nano, &end);    //null all the ptrs in moments for safety (may notneed to use all of them after all),

    try{moments->null_all();}
    catch(Exception &e){print_error_message(TMC_ERR); return TMC_ERR;}

     //set the lineages ready for data retrieval using the SC number
     try{moments->set_lineages();}
     catch(Exception &e){print_error_message(TMC_SET_LIN_ERR); return TMC_SET_LIN_ERR;}

   //load the calibration file into a QdRScalarSeq_var for leea...

   //get the data between start and end... if can't find efw u_probe_sc
   //from efw then use the value sc_potential
   //if override==true then use sc_potential without getting efw


   try
   {
   	if(moments->get_data(start, end, false, 7.0)!=TMC_OK)
   	{print_error_message(TMC_IDFS_RET_ERR);
   	return TMC_IDFS_RET_ERR;}
   }
   catch(Exception &e)
   {

   print_error_message(TMC_IDFS_RET_ERR);
   return TMC_IDFS_RET_ERR;

   }

   //for the overlap moments -- separate the first and second half of the spin,
   //shift first half forward and add together. join results in spin centred seqs
   try
   {
    if(moments->separate_and_sum_overlap()!=TMC_OK)
   	{print_error_message(TMC_SEP_ERR);
	 return TMC_SEP_ERR;}
   }
   catch(Exception &e)
   {
  	print_error_message(TMC_SEP_ERR);
   	return TMC_SEP_ERR;

   }
   //now join everything to the overlap (density, for argument's sake) tags
   //i.e. all are now 4 sec spin centred
   try
   {
   	if(moments->join_all()!=TMC_OK)
   	{print_error_message(TMC_JOIN_ERR);
	      return TMC_JOIN_ERR;}
   }
   catch(Exception &e)
   {
   	print_error_message(TMC_JOIN_ERR);
        return TMC_JOIN_ERR;

   }

   cout << "COMMAND HISTORY FOR CLUSTER-" << spacecraft << endl;
   moments->get_command_history();



   

	  //done! return OK

    return  TMC_OK;
}








